/********************************************************************/
/********************************************************************/
/**                                                                **/
/**                        ALGEBRE LINEAIRE                        **/
/**                        (deuxieme partie)                       **/
/**                                                                **/
/**                       copyright Babe Cool                      **/
/**                                                                **/
/********************************************************************/
/********************************************************************/
/* $Id: alglin2.c,v 1.914.7.2 1997/01/20 11:33:21 karim Exp karim $ */

# include "genpari.h"

/*******************************************************************/
/*                                                                 */
/*                      POLYNOME CARACTERISTIQUE                   */
/*                                                                 */
/*******************************************************************/

GEN
char0(GEN x, int v, long flag)
{
  switch(flag)
  {
    case 0: return caradj(x,v,0);
    case 1: return caract(x,v);
    case 2: return carhess(x,v);
  }
  err(flagerr);
  return NULL; /* not reached */
}

/* characteristic pol. Easy cases. Return NULL in case it's not so easy. */
static GEN
easychar(GEN x, int v, GEN *py)
{
  long l,tetpil,lx;
  GEN p1,p2,p3;

  switch(typ(x))
  {
    case t_INT: case t_REAL: case t_INTMOD:
    case t_FRAC: case t_FRACN: case t_PADIC:
      p1=cgetg(4,t_POL);
      p1[1]=evalsigne(1) | evallgef(4) | evalvarn(v);
      p1[2]=lneg(x); p1[3]=un;
      if (py)
      { 
	p2=cgetg(2,t_MAT);
	p2[1]=lgetg(2,t_COL);
	coeff(p2,1,1)=lcopy(x);
	*py=p2;
      }
      return p1;

    case t_COMPLEX: case t_QUAD:
      if (py) err(talker,"incorrect type in matadjoint");
      p1=cgetg(5,t_POL);
      p1[1]=evalsigne(1) | evallgef(5) | evalvarn(v);
      p1[2]=lnorm(x); l=avma; p2=trace(x); tetpil=avma;
      p1[3]=lpile(l,tetpil,gneg(p2));
      p1[4]=un; return p1;

    case t_POLMOD:
      if (py) err(talker,"incorrect type in matadjoint");
      l=avma; p1=gsub(polx[MAXVARN],(GEN) x[2]); tetpil=avma;
      p2 = (GEN)x[1]; p3 = (GEN)p2[lgef(p2)-1];
      p1=subres(p2,p1);

      if (varn(p1)==MAXVARN) setvarn(p1,v);
      else { tetpil=avma; p1=gsubst(p1,MAXVARN,polx[v]); }

      if (gcmp1(p3)) return gerepile(l,tetpil,p1);
      p2=gpuigs(p3,max(lgef(x[2])-3,0)); tetpil=avma;
      return gerepile(l,tetpil,gdiv(p1,p2));

    case t_MAT:
      lx=lg(x);
      if (lx==1)
      { 
	if (py)
	{ 
	  p2=cgetg(2,t_MAT);
	  p2[1]=lgetg(2,t_COL);
	  coeff(p2,1,1)=lcopy(x);
	  *py=p2;
	}
	return polun[v];
      }
      if (lg(x[1]) != lx) err(mattype1);
      return NULL;
  }
  err(mattype1);
  return NULL; /* not reached */
}

GEN
caract(GEN x, int v)
{
  long n,k,l=avma,tetpil;
  GEN  p1,p2,p3,p4,p5,p6;

  if (( p1=easychar(x,v,NULL) )) return p1;

  p1=gzero; p2=gun; p4=cgetg(3,t_RFRACN); p5=cgetg(4,t_POL); p6=cgeti(3);
  n=lg(x)-1; if (n&1) p2=gneg(p2);
  p5[1]=evalsigne(1)+evallgef(4)+evalvarn(v); p5[2]=zero; p5[3]=un;
  p4[2]=(long)p5;
  p6[1]=evalsigne(-1)+evallgef(3);
  for (k=0; k<=n; k++)
  {
    p3=det(gsub(gscalsmat(k,n),x));
    p4[1]=lmul(p3,p2); p6[2]=k;
    p1=gadd(p4,p1); p5[2]=(long)p6;
    if (k!=n) p2=gdivgs(gmulsg(k-n,p2),k+1);
  }
  p2=mpfact(n); tetpil=avma;
  return gerepile(l,tetpil,gdiv((GEN) p1[1],p2));
}

GEN
caradj0(GEN x, long v)
{
  return caradj(x,v,0);
}

/*
 * Methode des traces : retourne le polynome caracteristique. Si un pointeur
 * non nul est fourni, il pointe sur la matrice adjointe a la sortie.
 */
GEN
caradj(GEN x, long v, GEN *py)
{
  long i,j,k,l,av,tetpil;
  GEN p,y,t,*gptr[2];

  if (( p=easychar(x,v,py) )) return p;

  l=lg(x);
  if (l==1) { p=polun[v]; if (py) *py=gcopy(x); return p; }
  if (l==2) { p=gsub(polx[v],trace(x)); if (py) *py=gcopy(x); return p; }

  p=cgetg(l+2,t_POL); p[1]=evalsigne(1) | evallgef(l+2) | evalvarn(v);
  av=avma; t=trace(x); tetpil=avma;
  t=gerepile(av,tetpil,gneg(t));
  p[l]=(long)t; p[l+1]=un;

  av=avma; y=cgetg(l,t_MAT);
  for (j=1; j<l; j++)
  { 
    y[j]=lgetg(l,t_COL);
    for (i=1; i<l; i++)
      coeff(y,i,j) = (i==j) ? ladd(gcoeff(x,i,j),t) : coeff(x,i,j);
  }

  for (k=2; k<l-1; k++)
  {
    GEN z=gmul(x,y);
    
    t=trace(z); tetpil=avma;
    t=gdivgs(t,-k); y=cgetg(l,t_MAT);
    for (j=1; j<l; j++)
    {
      y[j]=lgetg(l,t_COL);
      for (i=1;i<l;i++)
	coeff(y,i,j) = (i==j)? ladd(gcoeff(z,i,i),t): lcopy(gcoeff(z,i,j));
    }
    gptr[0]=&y; gptr[1]=&t;
    gerepilemanysp(av,tetpil,gptr,2);
    p[l-k+1]=(long)t; av=avma;
  }

  t=gzero;
  for (i=1; i<l; i++)
    t=gadd(t,gmul(gcoeff(x,1,i),gcoeff(y,i,1)));
  tetpil=avma; t=gneg(t);

  if (py)
  {
    *py = (l&1) ? gneg(y) : gcopy(y);
    gptr[0] = py; gptr[1] = &t;
    gerepilemanysp(av,tetpil,gptr,2);
    p[2]=(long)t;
  }
  else p[2]=lpile(av,tetpil,t);

  return p;
}

GEN
adj(GEN x)
{
  GEN y;
  caradj(x,MAXVARN,&y); return y;
}

/*******************************************************************/
/*                                                                 */
/*                      INVERSION D'UNE MATRICE                    */
/*                                                                 */
/*******************************************************************/

/* a^(-1)*b, b etant une matrice. Il faut : nblig(a)=nbcol(a)=nblig(b) */
GEN
invmulmat(GEN a, GEN b)
{
  long nbli,nbco,i,j,k,av,av1,av2,av3,av4,tp;
  GEN aa,x,p,m,u;

  if (isinexactreal(a)) return invmulmatreel(a,b);
  nbco=lg(b)-1; if (!nbco) return cgetg(1,t_MAT);
  nbli=(lg(a)==1)?0:lg(a[1])-1;
  if (lg(a)-1 != nbli) err(invmuler1);
  if (nbli!=lg(b[1])-1) err(invmuler1);
  av=avma; x=cgetg(nbco+1,t_MAT);
  for (j=1; j<=nbco; j++)
  {
    x[j]=lgetg(nbli+1,t_COL);
    for (i=1; i<=nbli; i++)
      coeff(x,i,j)=coeff(b,i,j);
  }
  av1=avma; aa=cgetg(nbli+1,t_MAT);
  for (j=1; j<=nbli; j++)
  {
    aa[j]=lgetg(nbli+1,t_COL);
    for (i=1; i<=nbli; i++) coeff(aa,i,j)=coeff(a,i,j);
  }
  for (i=1; i<nbli; i++)
  {
    p=gcoeff(aa,i,i); k=i; tp=typ(p);
    if (gcmp0(p) || ( (tp==t_REAL || tp==t_COMPLEX) && 
	    (gexpo(p) < -((lg(p)-2)<<TWOPOTBITS_IN_LONG)) )
       )
    {
      k=i+1; while (k<=nbli && gcmp0(gcoeff(aa,k,i))) k++;
      if (k>nbli) err(matinv2);
      for (j=i; j<=nbli; j++)
      {
	u=gcoeff(aa,i,j); coeff(aa,i,j)=coeff(aa,k,j);
	coeff(aa,k,j)=(long)u;
      }
      for (j=1; j<=nbco; j++)
      {
	u=gcoeff(x,i,j); coeff(x,i,j)=coeff(x,k,j);
	coeff(x,k,j)=(long)u;
      }
      p=gcoeff(aa,i,i);
    }
    for (k=i+1; k<=nbli; k++)
    {
      m=gcoeff(aa,k,i);
      if (!gcmp0(m))
      {
	m=gdiv(m,p);
	for (j=i+1; j<=nbli; j++)
	  coeff(aa,k,j)=lsub(gcoeff(aa,k,j),gmul(m,gcoeff(aa,i,j)));
	for (j=1; j<=nbco; j++)
	  coeff(x,k,j)=lsub(gcoeff(x,k,j),gmul(m,gcoeff(x,i,j)));
      }
    }
  }
  av2=avma; p=gcoeff(aa,nbli,nbli);
  if (gcmp0(p)) err(matinv2);
  for (j=1; j<=nbco; j++)
  {
    coeff(x,nbli,j)=ldiv(gcoeff(x,nbli,j),p);
    for (i=nbli-1; i>0; i--)
    {
      av3=avma; m=gcoeff(x,i,j);
      for (k=i+1; k<=nbli; k++)
	m= gsub(m,gmul(gcoeff(aa,i,k),gcoeff(x,k,j)));
      av4=avma; coeff(x,i,j)=lpile(av3,av4,gdiv(m,gcoeff(aa,i,i)));
    }
  }
  av=lpile(av1,av2,0);
  for (i=1; i<=nbli; i++)
    for (j=1; j<=nbco; j++)
      if (gcoeff(x,i,j)<=(GEN)av1) coeff(x,i,j)+=av;
/* ATTENTION: VALABLE SI COEFF EST LONG; SINON DIVISER av PAR 4 */
  return x;
}

GEN
invmat(GEN a)
{
  long av=avma,tetpil;
  GEN b=gscalmat(gun,lg(a)-1);

  tetpil=avma; return gerepile(av,tetpil,invmulmat(a,b));
}

/* a^(-1)*b, b etant une matrice. Il faut : nblig(a)=nbcol(a)=nblig(b) */
GEN
invmulmatreel(GEN a, GEN b)
{
  long nbli,nbco,i,j,k,av,av1,av2,av3,av4;
  GEN aa,x,p1,p,m,u;

  nbco=lg(b)-1; if (!nbco) return cgetg(1,t_MAT);
  nbli=(lg(a)==1)?0:lg(a[1])-1;
  if (lg(a)-1 != nbli) err(invmuler1);
  if (nbli!=lg(b[1])-1) err(invmuler1);
  av=avma; x=cgetg(nbco+1,t_MAT);
  for (j=1; j<=nbco; j++)
  {
    x[j]=lgetg(nbli+1,t_COL);
    for (i=1; i<=nbli; i++)
      coeff(x,i,j)=coeff(b,i,j);
  }
  av1=avma; aa=cgetg(nbli+1,t_MAT);
  for (j=1; j<=nbli; j++)
  {
    aa[j]=lgetg(nbli+1,t_COL);
    for (i=1; i<=nbli; i++) coeff(aa,i,j)=coeff(a,i,j);
  }
  for (i=1; i<nbli; i++)
  {
    p=gabs(gcoeff(aa,i,i),DEFAULTPREC); k=i;
    for (j=i+1; j<=nbli; j++)
      if (gcmp(p1=gabs(gcoeff(aa,j,i),DEFAULTPREC),p)>0) { p=p1; k=j; }
    if (gcmp0(p)) err(matinv2);
    if (k>i)
    {
      for (j=i; j<=nbli; j++)
      {
	u=gcoeff(aa,i,j); coeff(aa,i,j)=coeff(aa,k,j);
	coeff(aa,k,j)=(long)u;
      }
      for (j=1; j<=nbco; j++)
      {
	u=gcoeff(x,i,j); coeff(x,i,j)=coeff(x,k,j);
	coeff(x,k,j)=(long)u;
      }
    }
    p=gcoeff(aa,i,i);
    for (k=i+1; k<=nbli; k++)
    {
      m=gcoeff(aa,k,i);
      if (!gcmp0(m))
      {
	m=gdiv(m,p);
	for (j=i+1; j<=nbli; j++)
	  coeff(aa,k,j)=lsub(gcoeff(aa,k,j),gmul(m,gcoeff(aa,i,j)));
	for (j=1; j<=nbco; j++)
	  coeff(x,k,j)=lsub(gcoeff(x,k,j),gmul(m,gcoeff(x,i,j)));
      }
    }
  }
  av2=avma; p=gcoeff(aa,nbli,nbli);
  if (gcmp0(p)) err(matinv2);
  for (j=1; j<=nbco; j++)
  {
    coeff(x,nbli,j)=ldiv(gcoeff(x,nbli,j),p);
    for (i=nbli-1; i>0; i--)
    {
      av3=avma; m=gcoeff(x,i,j);
      for (k=i+1; k<=nbli; k++)
	m= gsub(m,gmul(gcoeff(aa,i,k),gcoeff(x,k,j)));
      av4=avma; coeff(x,i,j)=lpile(av3,av4,gdiv(m,gcoeff(aa,i,i)));
    }
  }
  av=lpile(av1,av2,0);
  for (i=1; i<=nbli; i++)
    for (j=1; j<=nbco; j++)
      if (gcoeff(x,i,j)<=(GEN)av1) coeff(x,i,j)+=av;
  return x;
}

GEN
invmatreel(GEN a)
{
  long av=avma,tetpil;
  GEN b=gscalmat(gun,lg(a)-1);

  tetpil=avma; return gerepile(av,tetpil,invmulmatreel(a,b));
}

/*******************************************************************/
/*                                                                 */
/*              FORME DE HESSENBERG D'UNE MATRICE                  */
/*                                                                 */
/*******************************************************************/

GEN
hess(GEN x)
{
  long lx=lg(x),av=avma,tetpil,m,i,j;
  GEN p1,p2,p3,y;

  if (typ(x) != t_MAT || lg(x[1])!=lx) err(mattype1);
  y=gcopy(x);
  for (m=2; m<lx-1; m++)
  {
    p2=gzero; i=m+1;
    while (i<lx)
    { 
      p2=gcoeff(y,i,m-1);
      if (!gcmp0(p2))
      {
	for (j=m-1; j<lx; j++)
	{
	  p1=gcoeff(y,i,j);
	  coeff(y,i,j)=coeff(y,m,j);
	  coeff(y,m,j)=(long)p1;
	}
	p1=(GEN)y[i]; y[i]=y[m]; y[m]=(long)p1;
	for (i=m+1; i<lx; i++)
	  if (!gcmp0(p3=gcoeff(y,i,m-1)))
	  {
	    p3=gdiv(p3,p2); coeff(y,i,m-1)=zero;
	    for (j=m; j<lx; j++)
	      coeff(y,i,j)=lsub(gcoeff(y,i,j),gmul(p3,gcoeff(y,m,j)));
	    for (j=1; j<lx; j++)
	      coeff(y,j,m)=ladd(gcoeff(y,j,m),gmul(p3,gcoeff(y,j,i)));
	  }
	break;
      }
      i++;
    }
  }
  tetpil=avma; return gerepile(av,tetpil,gcopy(y));
}

GEN
carhess(GEN x, long v)
{
  long av,tetpil,lx,r,i;
  GEN *y,p1,p2,p3,p4;

  if (( p1=easychar(x,v,NULL) )) return p1;

  lx=lg(x); av=avma; y = (GEN*) newbloc(4*lx);
  y[0]=polun[v]; p1=hess(x); p2=polx[v];
  tetpil=avma;
  for (r=1; r<lx; r++)
  {
    y[r]=gmul(y[r-1], gsub(p2,gcoeff(p1,r,r)));
    p3=gun; p4=gzero;
    for (i=1; i<r; i++)
    {
      p3=gmul(p3,gcoeff(p1,r-i+1,r-i));
      p4=gadd(p4,gmul(gmul(p3,gcoeff(p1,r-i,r)),y[r-i-1]));
    }
    tetpil=avma; y[r]=gsub(y[r],p4);
  }
  p1=gerepile(av,tetpil,y[lx-1]);
  killbloc((GEN)y); return p1;
}

/*******************************************************************/
/*                                                                 */
/*                          NORME                                  */
/*                                                                 */
/*******************************************************************/

GEN
gnorm(GEN x)
{
  long l,lx,i,tetpil, tx=typ(x);
  GEN p1,p2,y;

  switch(tx)
  {
    case t_INT:
      return sqri(x);

    case t_REAL:
      return mulrr(x,x);

    case t_FRAC: case t_FRACN:
      return gsqr(x);

    case t_COMPLEX:
      l=avma; p1=gsqr((GEN) x[1]); p2=gsqr((GEN) x[2]);
      tetpil=avma; return gerepile(l,tetpil,gadd(p1,p2));

    case t_QUAD:
      l=avma; p1=(GEN)x[1];
      p2=gmul((GEN) p1[2],gmul((GEN) x[3],(GEN) x[3]));
      p1 = (gcmp0((GEN) p1[3])) ? gsqr((GEN) x[2])
                                : gmul((GEN) x[2],gadd((GEN) x[2],(GEN) x[3]));
      tetpil=avma; return gerepile(l,tetpil,gadd(p1,p2));

    case t_POL: case t_SER: case t_RFRAC: case t_RFRACN:
      l=avma; p1=gmul(gconj(x),x); tetpil=avma;
      return gerepile(l,tetpil,greal(p1));

    case t_POLMOD:
      p1=(GEN)x[1]; p2=(GEN)p1[lgef(p1)-1];
      if (gcmp1(p2) || gcmp0((GEN) x[2])) return subres(p1,(GEN) x[2]);
      l=avma; y=subres(p1,(GEN)x[2]); p1=gpuigs(p2,lgef(x[2])-3);
      tetpil=avma; return gerepile(l,tetpil,gdiv(y,p1));

    case t_VEC: case t_COL: case t_MAT:
      lx=lg(x); y=cgetg(lx,tx);
      for (i=1; i<lx; i++) y[i]=lnorm((GEN) x[i]);
      return y;
  }
  err(normer1);
  return NULL; /* not reached */
}

GEN
gnorml2(GEN x)
{
  GEN y,p1;
  long i,tx=typ(x),lx,av,tetpil;

  if (! is_matvec_t(tx)) return gnorm(x);
  lx=lg(x); if (lx==1) return gzero;

  av=tetpil=avma; y=gnorml2((GEN) x[1]);
  for (i=2; i<lx; i++)
    { p1=gnorml2((GEN) x[i]); tetpil=avma; y=gadd(p1,y); }
  return gerepile(av,tetpil,y);
}

/*******************************************************************/
/*                                                                 */
/*                          CONJUGAISON                            */
/*                                                                 */
/*******************************************************************/

GEN
gconj(GEN x)
{
  long lx,i,tx=typ(x);
  GEN z,p1;

  switch(tx)
  {
    case t_INT: case t_REAL: case t_INTMOD:
    case t_FRAC: case t_FRACN: case t_PADIC:
      return gcopy(x);

    case t_COMPLEX:
      z=cgetg(3,t_COMPLEX); z[1]=lcopy((GEN) x[1]);
      z[2]=lneg((GEN) x[2]); break;

    case t_QUAD:
      z=cgetg(4,t_QUAD); z[1]=copyifstack(x[1]);
      p1=(GEN)x[1]; z[3]=lneg((GEN) x[3]);
      z[2]=(gcmp0((GEN) p1[3])) ? lcopy((GEN) x[2])
                                : ladd((GEN) x[2],(GEN) x[3]);
      break;

    case t_POL:
      lx=lgef(x); z=cgetg(lx,tx); z[1]=x[1];
      for (i=2; i<lx; i++) z[i]=lconj((GEN) x[i]);
      break;

    case t_SER:
      lx=lg(x); z=cgetg(lx,tx); z[1]=x[1];
      if (!gcmp0(x))
	for (i=2; i<lx; i++) z[i]=lconj((GEN) x[i]);
      break;

    case t_RFRAC: case t_RFRACN: case t_VEC: case t_COL: case t_MAT:
      lx=lg(x); z=cgetg(lx,tx);
      for (i=1; i<lx; i++) z[i]=lconj((GEN) x[i]);
      break;

    default: err(conjer1); return NULL; /* not reached */
  }
  return z;
}

GEN
conjvec(GEN x,long prec)
{
  long lx,s,av,tetpil,i,tx=typ(x);
  GEN z,y,p1,p2,p;

  switch(tx)
  {
    case t_INT: case t_INTMOD: case t_FRAC: case t_FRACN:
      z=cgetg(2,t_COL); z[1]=lcopy(x); break;

    case t_COMPLEX: case t_QUAD:
      z=cgetg(3,t_COL); z[1]=lcopy(x); z[2]=lconj(x); break;

    case t_VEC: case t_COL:
      lx=lg(x); z=cgetg(lx,t_MAT);
      for (i=1; i<lx; i++) z[i]=(long)conjvec((GEN)x[i],prec);
      s=lg(z[1]);
      for (i=2; i<lx; i++)
	if (lg(z[i])!=s) err(conjvecer1);
      break;

    case t_POLMOD:
      y=(GEN)x[1]; lx=lgef(y);
      if (lx<=3) return cgetg(1,t_COL);
      av=avma; p=gzero;
      for (i=2; i<lx; i++)
      {
	tx=typ(y[i]);
	if ((tx>5)||(tx==2)) err(conjvecer2);
	if (tx==3) p=(GEN)((GEN)y[i])[1];
      }
      if (!signe(p))
      {
	p1=roots(y,prec);
	z=cgetg(lx-2,t_COL);
	for (i=1; i<=lx-3; i++)
	{
	  p2=(GEN)p1[i]; if (gcmp0(gimag(p2))) p2=greal(p2);
	  z[i]=(long)poleval((GEN)x[2],(GEN)p1[i]);
	 }
	tetpil=avma; return gerepile(av,tetpil,gcopy(z));
      }
      z=cgetg(lx-2,t_COL); z[1]=lcopy(x);
      for (i=2; i<=lx-3; i++) z[i]=(long)gpui((GEN) z[i-1],p,prec);
      break;

    default: err(conjer1); return NULL; /* not reached */
  }
  return z;
}

/*******************************************************************/
/*                                                                 */
/*                            TRACES                               */
/*                                                                 */
/*******************************************************************/

GEN
assmat(GEN x)
{
  long lx,i,j;
  GEN y,p1;

  if (typ(x)!=t_POL || gcmp0(x)) err(mattype2);
  lx=lgef(x)-2; y=cgetg(lx,t_MAT);
  for (i=1; i<lx-1; i++)
  {
    p1=cgetg(lx,t_COL); y[i]=(long)p1;
    for (j=1; j<lx; j++) { p1[j] = (j==(i+1)) ? un : zero; }
  }
  p1=cgetg(lx,t_COL); y[i]=(long)p1;
  if (gcmp1((GEN) x[lx+1]))
    for (j=1; j<lx; j++) p1[j]=lneg((GEN)x[j+1]);
  else
  {
    gnegz((GEN)x[lx+1],(GEN)x[lx+1]);
    for (j=1; j<lx; j++)
      p1[j]=ldiv((GEN)x[j+1],(GEN)x[lx+1]);
    gnegz((GEN)x[lx+1],(GEN)x[lx+1]);
  }
  return y;
}

GEN
trace(GEN x)
{
  long i,l,n,tx=typ(x),lx,tetpil;
  GEN y,p1,p2;

  switch(tx)
  {
    case t_INT: case t_REAL: case t_FRAC: case t_FRACN:
      return gmul2n(x,1);

    case t_COMPLEX:
      return gmul2n((GEN)x[1],1);

    case t_QUAD:
      p1=(GEN)x[1];
      if (!gcmp0((GEN) p1[3]))
      {
	l=avma; p2=gmul2n((GEN)x[2],1);
	tetpil=avma; return gerepile(l,tetpil,gadd((GEN)x[3],p2));
      }
      return gmul2n((GEN)x[2],1);

    case t_POL:
      lx=lg(x); y=cgetg(lx,tx);
      y[1]=x[1];
      for (i=2; i<lgef(x); i++)
        y[i]=ltrace((GEN)x[i]);
      return y;

    case t_SER:
      lx=lg(x); y=cgetg(lx,tx);
      y[1]=x[1];
      if (!gcmp0(x))
      {
	for (i=2; i<lx; i++)
	  y[i]=ltrace((GEN)x[i]);
      }
      return y;

    case t_POLMOD:
      l=avma; p1=polsym((GEN)x[1],n=(lgef(x[1])-4)); p2=gzero;
      for (i=0; i<=n; i++)
	p2=gadd(p2,gmul(truecoeff((GEN)x[2],i),(GEN)p1[i+1]));
      tetpil=avma; return gerepile(l,tetpil,gcopy(p2));

    case t_RFRAC: case t_RFRACN:
      return gadd(x,gconj(x));

    case t_VEC: case t_COL:
      lx=lg(x); y=cgetg(lx,tx);
      for (i=1; i<lx; i++) y[i]=ltrace((GEN)x[i]);
      return y;

    case t_MAT:
      lx=lg(x); if (lx!=lg(x[1])) err(mattype3);
      l=avma; p1=gcopy(gcoeff(x,1,1)); if (lx==2) return p1;
      for (i=2; i<lx-1; i++)
	p1=gadd(p1,gcoeff(x,i,i));
      tetpil=avma; return gerepile(l,tetpil,gadd(p1,gcoeff(x,i,i)));

  }
  err(mattype3);
  return NULL; /* not reached */
}

/* =======================================================
 * Reduction de Gauss ( Matrice definie >0 )
 * ========================================================*/
GEN
sqred1(GEN a)
{
  GEN b,p;
  long av,tetpil,i,j,k, lim=(avma+bot)>>1, n=lg(a);

  if (typ(a)!=t_MAT) err(kerer1);
  if (lg(a[1])!=n) err(mattype1);
  av=avma; b=cgetg(n,t_MAT);
  for (j=1; j<n; j++)
  {
    GEN p1=cgetg(n,t_COL), p2=(GEN)a[j];

    b[j]=(long)p1;
    for (i=1; i<=j; i++) p1[i] = p2[i];
    for (   ; i<n ; i++) p1[i] = zero;
  }
  for (k=1; k<n; k++)
  {
    p=gcoeff(b,k,k);
    if (gsigne(p)<=0) err(sqreder1);
    for (i=k+1; i<n; i++)
      for (j=i; j<n; j++)
	coeff(b,i,j)=lsub(gcoeff(b,i,j),
	                  gdiv(gmul(gcoeff(b,k,i),gcoeff(b,k,j)), p));
    for (j=k+1; j<n; j++)
      coeff(b,k,j)=ldiv(gcoeff(b,k,j),p);
    if (avma<lim) { tetpil=avma; b=gerepile(av,tetpil,gcopy(b)); }
  }
  tetpil=avma;
  return gerepile(av,tetpil,gcopy(b));
}

GEN
sqred3(GEN a)
{
  long av=avma,tetpil,i,j,k,l, lim=(avma+bot)>>1, n=lg(a);
  GEN p1,z;

  if (typ(a)!=t_MAT) err(kerer1);
  if (lg(a[1])!=n) err(mattype1);
  av=avma; z=cgetg(n,t_MAT);
  for (j=1; j<n; j++)
  {
    p1=cgetg(n,t_COL); z[j]=(long)p1;
    for (i=1; i<n; i++) p1[i]=zero;
  }
  for (i=1; i<n; i++)
  {
    for (k=1; k<i; k++)
    {
      p1=gzero;
      for (l=1; l<k; l++)
	p1=gadd(p1, gmul(gmul(gcoeff(z,l,l),gcoeff(z,k,l)), gcoeff(z,i,l)));
      coeff(z,i,k)=ldiv(gsub(gcoeff(a,i,k),p1),gcoeff(z,k,k));
    }
    p1=gzero;
    for (l=1; l<i; l++)
      p1=gadd(p1, gmul(gmul(gcoeff(z,l,l),gcoeff(z,i,l)), gcoeff(z,i,l)));
    coeff(z,i,k)=lsub(gcoeff(a,i,i),p1);
    if (avma<lim) { tetpil=avma; z=gerepile(av,tetpil,gcopy(z)); }
  }
  tetpil=avma; return gerepile(av,tetpil,gcopy(z));
}

/*=======================================================
  Reduction de Gauss (matrice symetrique quelconque)
  Signature d'une matrice symetrique
  ( seule la partie superieure est consideree )
  ========================================================*/
static GEN
sqred2(GEN a, long do_signat)
{
  GEN b,r,p,u;
  long av,av1,av2,lim,n,i,j,k,l,sp,sn,t;
  if (typ(a)!=t_MAT) err(kerer1);
  if (lg(a[1])!=(n=lg(a))) err(mattype1);
  av=avma; lim=(avma+bot)>>1;
  r=cgeti(n); for (i=1; i<n; i++) r[i]=1;
  av2=avma; b=gcopy(a);
  t=(--n); sp=sn=0;

  while (t)
  {
    k=1; while (k<=n && (gcmp0(gcoeff(b,k,k)) || !r[k])) k++;
    if (k<=n)
    {
      p=gcoeff(b,k,k); if (gsigne(p)>0) sp++; else sn++;
      r[k]=0; t--;
      for (j=1; j<=n; j++)
	coeff(b,k,j)=r[j] ? ldiv(gcoeff(b,k,j),p) : zero;
	
      for (i=1; i<=n; i++) if (r[i])
      {
	for (j=1; j<=n; j++)
	  coeff(b,i,j) = r[j] ? lsub(gcoeff(b,i,j),
	                             gmul(gmul(gcoeff(b,k,i),gcoeff(b,k,j)),p))
			      : zero;
      }
      coeff(b,k,k)=(long)p;
    }
    else
    {
      for (k=1; k<=n; k++) if (r[k])
      {
	l=k+1; while (l<=n && (gcmp0(gcoeff(b,k,l)) || !r[l])) l++;
	if (l<=n)
	{
	  p=gcoeff(b,k,l); r[k]=r[l]=0; sp++; sn++; t-=2;
	  for (i=1; i<=n; i++) if (r[i])
	  {
	    for (j=1; j<=n; j++)
	      coeff(b,i,j) =
		r[j] ? lsub(gcoeff(b,i,j),
			    gdiv(gadd(gmul(gcoeff(b,k,i),gcoeff(b,l,j)),
				      gmul(gcoeff(b,k,j),gcoeff(b,l,i))),
				 p))
		    : zero;
	  }
	  for (j=1; j<=n; j++) if (r[j])
	  {
	    u=gcoeff(b,k,j);
	    coeff(b,k,j)=ldiv(gadd(u,gcoeff(b,l,j)),p);
	    coeff(b,l,j)=ldiv(gsub(u,gcoeff(b,l,j)),p);
	  }
	  coeff(b,k,l)=un; coeff(b,l,k)=lneg(gun);
	  coeff(b,k,k)=lmul2n(p,-1); coeff(b,l,l)=lneg(gcoeff(b,k,k));
	  break;
	 }
	if (avma<lim) { av1=avma; b=gerepile(av2,av1,gcopy(b)); }
      }
      if (k>n) break;
    }
  }
  if (do_signat) { av1=avma; return gerepile(av,av1,gcopy(b)); }
  avma=av;
  b=cgetg(3,t_VEC); b[1]=lstoi(sp); b[2]=lstoi(sn); return b;
}

GEN
sqred(GEN a) { return sqred2(a,1); }

GEN
signat(GEN a) { return sqred2(a,0); }

/*===========================================================================
  Diagonalisation d'une matrice symetrique REELLE;
  matrice de passage orthogonale R
  (  Renvoie un vecteur a 2 comp :
  1-re comp = vect des valeurs propres
  2-me comp = matr des vecteurs propres   ).
  ============================================================================*/
GEN
jacobi(GEN a, long prec)
{
  long de,e,e1,e2,l,n,i,j,p,q,av1,av2,iter=0;
  GEN c,s,t,u,ja,lambda,r,unr,x,y,x1,y1;

  if (typ(a)!=t_MAT) err(mattype1);
  ja=cgetg(3,t_VEC);
  n=(l=lg(a))-1;
  e1=HIGHEXPOBIT-1;
  lambda=cgetg(l,t_COL); ja[1]=(long)lambda;
  for (j=1; j<=n; j++)
  {
    gaffect(gcoeff(a,j,j),x = (GEN)(lambda[j]=lgetr(prec)));
/*      if (((e=expo(x))<e1)&&(gsigne(x))) e1=e; */
    if ((e=expo(x))<e1) e1=e;

/* e1 = min des expo des coeff diagonaux
   e2 = max des expo des coeff extra-diagonaux
   Test d'arret: e2 < e1-precision  */
  }
  r=cgetg(l,t_MAT); ja[2]=(long)r;
  for (j=1; j<=n; j++)
  {
    r[j]=lgetg(l,t_COL);
    for (i=1; i<l; i++)
      affsr(i==j,(GEN)(coeff(r,i,j)=lgetr(prec)));
  }
  av1=avma;
  e2=-HIGHEXPOBIT;
  c=cgetg(l,t_MAT);
  for (j=1; j<=n; j++)
  {
    c[j]=lgetg(j,t_COL);
    for (i=1; i<j; i++)
    {
      gaffect(gcoeff(a,i,j),x=(GEN)(coeff(c,i,j)=lgetr(prec)));
      if ((e=expo(x))>e2) { e2=e; p=i; q=j; }
    }
  }
  a=c; affsr(1,unr=cgetr(prec));
  de=((prec-2)<<TWOPOTBITS_IN_LONG);

  while (e1-e2<de)
  {
    iter++;
	/*calcul de la rotation associee dans le plan
	  des p et q-iemes vecteurs de base   */
    av2=avma;
    x=divrr(subrr((GEN)lambda[q],(GEN)lambda[p]),shiftr(gcoeff(a,p,q),1));
    y=mpsqrt(addrr(unr,mulrr(x,x)));
    t=(gsigne(x)>0)? divrr(unr,addrr(x,y)) : divrr(unr,subrr(x,y));
    c=divrr(unr,mpsqrt(addrr(unr,mulrr(t,t))));
    s=mulrr(t,c); u=divrr(s,addrr(unr,c));

	/* Recalcul des transformees successives de a et de la matrice
	   cumulee (r) des rotations :  */

    for (i=1; i<p; i++)
    {
      x=gcoeff(a,i,p); y=gcoeff(a,i,q);
      x1=subrr(x,mulrr(s,addrr(y,mulrr(u,x))));
      y1=addrr(y,mulrr(s,subrr(x,mulrr(u,y))));
      affrr(x1,gcoeff(a,i,p)); affrr(y1,gcoeff(a,i,q));
    }
    for (i=p+1; i<q; i++)
    {
      x=gcoeff(a,p,i); y=gcoeff(a,i,q);
      x1=subrr(x,mulrr(s,addrr(y,mulrr(u,x))));
      y1=addrr(y,mulrr(s,subrr(x,mulrr(u,y))));
      affrr(x1,gcoeff(a,p,i)); affrr(y1,gcoeff(a,i,q));
    }
    for (i=q+1; i<=n; i++)
    {
      x=gcoeff(a,p,i); y=gcoeff(a,q,i);
      x1=subrr(x,mulrr(s,addrr(y,mulrr(u,x))));
      y1=addrr(y,mulrr(s,subrr(x,mulrr(u,y))));
      affrr(x1,gcoeff(a,p,i)); affrr(y1,gcoeff(a,q,i));
    }
    x=(GEN)lambda[p]; y=gcoeff(a,p,q); subrrz(x,mulrr(t,y),(GEN)lambda[p]);
    x=y; y=(GEN)lambda[q]; addrrz(y,mulrr(t,x),y);
    setexpo(x,expo(x)-de-1);

    for (i=1; i<=n; i++)
    {
      x=gcoeff(r,i,p); y=gcoeff(r,i,q);
      x1=subrr(x,mulrr(s,addrr(y,mulrr(u,x))));
      y1=addrr(y,mulrr(s,subrr(x,mulrr(u,y))));
      affrr(x1,gcoeff(r,i,p)); affrr(y1,gcoeff(r,i,q));
    }

    e2=expo(gcoeff(a,1,2)); p=1; q=2;
    for (j=1; j<=n; j++)
    {
      for (i=1; i<j; i++) if ((e=expo(gcoeff(a,i,j)))>e2) { e2=e; p=i; q=j; }
      for (i=j+1; i<=n; i++) if ((e=expo(gcoeff(a,j,i)))>e2) { e2=e; p=j; q=i; }
    }
    avma=av2;
  }
  avma=av1; return ja;
}

/*************************************************************************/
/**									**/
/**		    MATRICE RATIONNELLE-->ENTIERE			**/
/**									**/
/*************************************************************************/

GEN
matrixqz0(GEN x,GEN pp)
{
  if (typ(pp)!=t_INT) err(talker,"incorrect type in matrixqz");

  if (signe(pp)>=0) return matrixqz(x,pp);
  if (gegal(pp,stoi(-1))) return matrixqz2(x);
  if (gegal(pp,stoi(-2))) return matrixqz3(x);
  err(flagerr);
  return NULL; /* not reached */
}

GEN
matrixqz(GEN x, GEN pp)
{
  long av=avma,av1,tetpil,i,j,j1,m,n,t,fl,lim,nfact;
  GEN p,p1,p2,p3,p4,p5,unmodp,pk,pt;

  if (typ(x)!=t_MAT) err(matqzer1);
  lim=(avma+bot)>>1;
  n=lg(x)-1; if (!n) return gcopy(x);
  m=lg(x[1])-1; if (n>m) err(matqzer2);
  if (n==m)
  {
    p1=det(x); if (gcmp0(p1)) err(matqzer4);
    return idmat(n);
  }
  p1=cgetg(n+1,t_MAT);
  for (j=1; j<=n; j++)
  {
    p2=gun; p3=(GEN)x[j];
    for (i=1; i<=m; i++)
    {
      t=typ(p3[i]); if ((t>5)||(t==3)) err(matqzer3);
      p2=ggcd(p2,(GEN)p3[i]);
    }
    p1[j]=ldiv(p3,p2);
  }
  x=p1;
  if (gcmp0(pp))
  {
    pt=gtrans(x);
    p1=cgetg(n+1,t_MAT);
    for (j=1; j<=n; j++) p1[j]=pt[j];
    p2=det(p1); p1[n]=pt[n+1]; p3=det(p1);
    p4=ggcd(p2,p3); if (!signe(p4)) err(impl,"matrixqz when the first 2 dets are zero");
    if (gcmp1(p4)) { tetpil=avma; return gerepile(av,tetpil,gcopy(x)); }
    p3=factor(p4); p1=(GEN)p3[1]; p2=(GEN)p3[2]; nfact=lg(p1)-1;
    av1=avma; p3=cgetg(n+1,t_VEC);
    for (i=1; i<=nfact; i++)
    {
      p=(GEN)p1[i]; unmodp=gmodulcp(gun,p); fl=1;
      while (fl)
      {
	pk=ker(gmul(unmodp,x)); if (lg(pk)==1) fl=0;
	else
	{
	  p5=centerlift(pk); p4=gdiv(gmul(x,p5),p);
	  for (i=1; i<=n; i++) p3[i]=0;
	  for (j=1; j<lg(p5); j++)
	  {
	    j1=n; while (gcmp0(gcoeff(p5,j1,j))) j1--;
	    p3[j1]=j;
	  }
	  for (j=1; j<=n; j++) if (p3[j]) x[j]=p4[p3[j]];
	 }
      }
      if (avma<lim) { tetpil=avma; x=gerepile(av1,tetpil,gcopy(x)); }
    }
  }
  else
  {
    unmodp=gmodulcp(gun,pp); fl=1; av1=avma;
    while (fl)
    {
      pk=ker(gmul(unmodp,x)); if (lg(pk)==1) fl=0;
      else
      {
	p5=centerlift(pk); p4=gdiv(gmul(x,p5),pp);
	for (i=1; i<=n; i++) p3[i]=0;
	for (j=1; j<lg(p5); j++)
	{
	  j1=n; while (gcmp0(gcoeff(p5,j1,j))) j1--;
	  p3[j1]=j;
	 }
	for (j=1; j<=n; j++)	if (p3[j]) x[j]=p4[p3[j]];
	if (avma<lim) { tetpil=avma; x=gerepile(av1,tetpil,gcopy(x)); }
      }
    }
  }
  tetpil=avma; return gerepile(av,tetpil,gcopy(x));
}

GEN
matrixqz2(GEN x)
{
  long av=avma,tetpil,i,j,k,m,n,fl,lim,in[2];
  GEN p1;

  if (typ(x)!=t_MAT) err(matqzer1);
  lim=(avma+bot)>>1;
  n=lg(x)-1; if (!n) return gcopy(x);
  x=gcopy(x);
  m=lg(x[1])-1;
  for (i=1; i<=m; i++)
  {
    do
    {
      for (fl=0,j=1; (j<=n)&&(fl<2); j++)
	if (!gcmp0(gcoeff(x,i,j))) in[fl++]=j;
      if (fl==2)
      {
	j=(gcmp(gabs(gcoeff(x,i,in[0]),DEFAULTPREC),gabs(gcoeff(x,i,in[1]),DEFAULTPREC))>0)?in[1]:in[0];
	p1=(GEN)coeff(x,i,j);
	for (k=1; k<=n; k++)
	  if (k!=j) x[k]=lsub((GEN)x[k],gmul(ground(gdiv(gcoeff(x,i,k),p1)),(GEN)x[j]));
      }
    }
    while (fl==2);
    j=1; while ((j<=n)&&gcmp0(gcoeff(x,i,j))) j++;
    if (j<=n) x[j]=lmul(denom(gcoeff(x,i,j)),(GEN)x[j]);
    if (avma<lim) { tetpil=avma; x=gerepile(av,tetpil,gcopy(x)); }
  }
  tetpil=avma; return gerepile(av,tetpil,hnf(x));
}

GEN
matrixqz3(GEN x)
{
  long av=avma,av1,tetpil,i,j,j1,k,m,n,fl,lim,in[2];
  GEN p1,c,d;

  if (typ(x)!=t_MAT) err(matqzer1);
  lim=(avma+bot)>>1;
  n=lg(x)-1; if (!n) return gcopy(x);
  x=gcopy(x); m=lg(x[1])-1;
  c=cgeti(n+1); for (i=1; i<=n; i++) c[i]=0;
  d=cgeti(m+1); for (i=1; i<=m; i++) d[i]=0;
  av1=avma;
  for (k=1; k<=m; k++)
  {
    j=1; while ((j<=n)&&(c[j]||gcmp0(gcoeff(x,k,j)))) j++;
    if (j<=n)
    {
      x[j]=ldiv((GEN)x[j],gcoeff(x,k,j));
      for (j1=1; j1<=n; j1++)
	if (j1!=j) x[j1]=lsub((GEN)x[j1],gmul(gcoeff(x,k,j1),(GEN)x[j]));
      c[j]=k; d[k]=j;
    }
    if (avma<lim) { tetpil=avma; x=gerepile(av1,tetpil,gcopy(x)); }
  }
  for (i=1; i<=m; i++)
  {
    do
    {
      for (fl=0,j=1; (j<=n)&&(fl<2); j++)
	if (!gcmp0(gcoeff(x,i,j))) in[fl++]=j;
      if (fl==2)
      {
	j=(gcmp(gabs(gcoeff(x,i,in[0]),DEFAULTPREC),gabs(gcoeff(x,i,in[1]),DEFAULTPREC))>0)?in[1]:in[0];
	p1=(GEN)coeff(x,i,j);
	for (k=1; k<=n; k++)
	  if (k!=j) x[k]=lsub((GEN)x[k],gmul(ground(gdiv(gcoeff(x,i,k),p1)),(GEN)x[j]));
      }
    }
    while (fl==2);
    j=1; while ((j<=n)&&gcmp0(gcoeff(x,i,j))) j++;
    if (j<=n) { p1=denom(gcoeff(x,i,j)); if (!gcmp1(p1)) x[j]=lmul(p1,(GEN)x[j]); }
    if (avma<lim) { tetpil=avma; x=gerepile(av1,tetpil,gcopy(x)); }
  }
  tetpil=avma; return gerepile(av,tetpil,hnf(x));
}

GEN
intersect(GEN x, GEN y)
{
  long av=avma,tetpil,i,j,k,p;
  GEN z,p1,r;

  if (typ(x)!=t_MAT || typ(y)!=t_MAT) err(interer1);
  if (lg(x)==1 || lg(y)==1) return cgetg(1,t_MAT);
  z=ker(concat(x,y)); k=lg(x); p=lg(z); r=cgetg(p,t_MAT);
  for (j=1; j<p; j++)
  {
    p1=cgetg(k,t_COL); r[j]=(long)p1;
    for (i=1; i<k; i++) p1[i]=(long)coeff(z,i,j);
  }
  tetpil=avma; return gerepile(av,tetpil,gmul(x,r));
}
