/*********************************************************************/
/*********************************************************************/
/**                                                                 **/
/**                    FONCTIONS ARITHMETIQUES                      **/
/**                       (premiere partie)                         **/
/**                                                                 **/
/**                      copyright Babe Cool                        **/
/**                                                                 **/
/*********************************************************************/
/*********************************************************************/
/* $Id: arith1.c,v 1.914.7.2 1997/01/20 11:33:21 karim Exp karim $ */

#include  "genpari.h"


/*********************************************************************/
/**                                                                 **/
/**                  ARITHMETIC FUNCTION PROTOTYPES                 **/
/**                                                                 **/
/*********************************************************************/
GEN
garith_proto(GEN f(GEN), GEN x)
{
  long tx=typ(x),lx,i;
  GEN y;

  if (is_matvec_t(tx))
  {
    lx=lg(x); y=cgetg(lx,tx);
    for (i=1; i<lx; i++) y[i] = (long) garith_proto(f,(GEN) x[i]);
    return y;
  }
  if (tx != t_INT) err(arither1);
  return f(x);
}

GEN
arith_proto(long f(GEN), GEN x, int do_error)
{
  long tx=typ(x),lx,i;
  GEN y;

  if (is_matvec_t(tx))
  {
    lx=lg(x); y=cgetg(lx,tx);
    for (i=1; i<lx; i++) y[i] = (long) arith_proto(f,(GEN) x[i], do_error);
    return y;
  }
  if (tx != t_INT && do_error) err(arither1);
  return stoi(f(x));
}

static GEN
arith_proto2(long f(GEN,GEN), GEN x, GEN n)
{
  long l,i,tx = typ(x);
  GEN y;

  if (is_matvec_t(tx))
  {
    l=lg(x); y=cgetg(l,tx);
    for (i=1; i<l; i++) y[i] = (long) arith_proto2(f,(GEN) x[i],n);
    return y;
  }
  if (tx != t_INT) err(arither1);
  tx=typ(n);
  if (is_matvec_t(tx))
  {
    l=lg(n); y=cgetg(l,tx);
    for (i=1; i<l; i++) y[i] = (long) arith_proto2(f,x,(GEN) n[i]);
    return y;
  }
  if (tx != t_INT) err(arither1);
  return stoi(f(x,n));
}

static GEN
arith_proto2gs(long f(GEN,long), GEN x, long y)
{
  long l,i,tx = typ(x);
  GEN t;

  if (is_matvec_t(tx))
  {
    l=lg(x); t=cgetg(l,tx);
    for (i=1; i<l; i++) t[i]= (long) arith_proto2gs(f,(GEN) x[i],y);
    return t;
  }
  if (tx != t_INT) err(arither1);
  return stoi(f(x,y));
}

GEN
garith_proto2gs(GEN f(GEN,long), GEN x, long y)
{
  long l,i,tx = typ(x);
  GEN t;

  if (is_matvec_t(tx))
  {
    l=lg(x); t=cgetg(l,tx);
    for (i=1; i<l; i++) t[i]= (long) garith_proto2gs(f,(GEN) x[i],y);
    return t;
  }
  if (tx != t_INT) err(arither1);
  return f(x,y);
}

/*********************************************************************/
/**                                                                 **/
/**                       BINARY DECOMPOSITION                      **/
/**                                                                 **/
/*********************************************************************/

GEN
binaire(GEN x)
{
  uLong m,u;
  long i,lx,ex,ly,tx=typ(x);
  GEN y,p1,p2;

  switch(tx)
  {
    case t_INT:
      lx=lgefint(x);
      if (lx==2) { y=cgetg(2,t_VEC); y[1]=zero; return y; }
      ly = BITS_IN_LONG+1; m=HIGHBIT; u=x[2];
      while (!(m & u)) { m>>=1; ly--; }
      y = cgetg(ly+((lx-3)<<TWOPOTBITS_IN_LONG),t_VEC); ly=1;
      do { y[ly] = m & u ? un : zero; ly++; } while (m>>=1);
      for (i=3; i<lx; i++)
      {
	m=HIGHBIT; u=x[i];
	do { y[ly] = m & u ? un : zero; ly++; } while (m>>=1);
      }
      break;

    case t_REAL:
      ex=expo(x);
      if (!signe(x))
      {
	lx=1+max(-ex,0); y=cgetg(lx,t_VEC);
	for (i=1; i<lx; i++) y[i]=zero;
	return y;
      }

      lx=lg(x); y=cgetg(3,t_VEC);
      if (ex>((lx-2)<<TWOPOTBITS_IN_LONG)) err(biner1);
      p1 = cgetg(max(ex,0)+2,t_VEC);
      p2 = cgetg(((lx-2)<<TWOPOTBITS_IN_LONG)-ex,t_VEC);
      y[1] = (long) p1; y[2] = (long) p2;
      ly = -ex; ex++;
      if (ex<=0)
      {
	p1[1]=zero; for (i=1; i <= -ex; i++) p2[i]=zero;
	i=2; m=HIGHBIT;
      }
      else
      {
	ly=1;
	for (i=2; i<lx && ly<=ex; i++)
	{
	  m=HIGHBIT; u=x[i];
	  do
	    { p1[ly] = (m & u) ? un : zero; ly++; }
	  while ((m>>=1) && ly<=ex);
	}
	ly=1;
	if (m) i--; else m=HIGHBIT;
      }

      for (; i<lx; i++)
      {
	u=x[i];
	do { p2[ly] = m & u ? un : zero; ly++; } while (m>>=1);
	m=HIGHBIT;
      }
      break;

    case t_VEC: case t_COL: case t_MAT:
      lx=lg(x); y=cgetg(lx,tx);
      for (i=1; i<lx; i++) y[i]=(long)binaire((GEN)x[i]);
      break;
    default: err(biner2); return NULL; /* not reached */
  }
  return y;
}

/* Renvoie 0 ou 1 selon que le bit numero n de x est a 0 ou 1 */
long
bittest(GEN x, long n)
{
  long l;

  if (!signe(x) || n<0) return 0;
  l = lgefint(x)-1-(n>>TWOPOTBITS_IN_LONG);
  if (l<=1) return 0;
  n = (1L << (n & (BITS_IN_LONG-1))) & x[l];
  return n ? 1 : 0;
}

static long
bittestg(GEN x, GEN n)
{
  return bittest(x,itos(n));
}

GEN
gbittest(GEN x, GEN n)
{
  return arith_proto2(bittestg,x,n);
}

/*********************************************************************/
/**                                                                 **/
/**               ORDER of INTEGERMOD x  in  (Z/nZ)*                **/
/**                                                                 **/
/*********************************************************************/

GEN
order(GEN x)
{
  long av=avma,av1,i,e;
  GEN o,m,p;

  if (typ(x) != t_INTMOD) err(orderer);
  m = (GEN) x[1];
  if (!gcmp1(mppgcd(m,(GEN) x[2]))) err(orderer);
  o=phi(m); m=decomp(o);
  for (i = lg(m[1])-1; i; i--)
  {
    p=gcoeff(m,i,1); e=itos(gcoeff(m,i,2));
    do
    {
      GEN o1=divii(o,p), y=gpui(x,o1,0);

      e--; if (!gcmp1((GEN)y[2])) break;
      o=o1;
    }
    while (e);
  }
  av1=avma; return gerepile(av,av1,gcopy(o));
}

/******************************************************************/
/*                                                                */
/*                 GENERATOR of (Z/mZ)*                           */
/*                                                                */
/******************************************************************/

GEN
ggener(GEN m)
{
  return garith_proto(gener,m);
}

GEN
gener(GEN m)
{
  long av=avma,av1,k,i,e;
  GEN x,xi,y,t,q,p;

  if (typ(m) != t_INT) err(arither1);
  m=absi(m); if (gcmp1(m)) return gmodulcp(gzero,gun);
  e = m[lgefint(m)-1];
  if (!(e&3)) /* e = 0 mod 4 */
  {
    if (cmpis(m,4)) err(generer);
    return gmodulcp(stoi(3),m);
  }

  if (!(e&1)) /* e = 0 mod 2 */
  {
    q=gshift(m,-1); t = (GEN) gener(q)[2];
    if (!mpodd(t)) t=addii(t,q);
    av1=avma; return gerepile(av,av1,gmodulcp(t,m));
  }

  t=decomp(m); if (lg(t[1]) != 2) err(generer);
  p=gcoeff(t,1,1); e=itos(gcoeff(t,1,2)); q=subis(p,1);
  if (e>=2)
  {
    t = (GEN) gener(p)[2]; x=gmodulcp(t,gsqr(p));
    if (gcmp1((GEN)gpui(x,q,0)[2])) t=addii(t,p);
    av1=avma; return gerepile(av,av1,gmodulcp(t,m));
  }

  t=decomp(q); k=lg(t[1])-1; xi=stoi(1);
  for(;;)
  {
    xi[2]++;
    if (gcmp1(mppgcd(xi,m)))
    {
      x=gmodulcp(xi,m);
      for (i=k; i; i--)
      {
	y=gpui(x,divii(q,gcoeff(t,i,1)),0);
	if (gcmp1((GEN)y[2])) break;
      }
      if (!i) break;
    }
  }
  av1=avma; return gerepile(av,av1,gcopy(x));
}

GEN
znstar(GEN n)
{
  GEN p1,z,q,u,v,d,fa,list,ep,gen2,h,generators,moduli,p,a;
  long i,j,c,nbp,jlist,sizeh,epp,av=avma,tetpil;

  if (typ(n) != t_INT) err(arither1);
  if (!signe(n))
  {
    z=cgetg(4,t_VEC); z[1]=lcopy(gdeux); p1=cgetg(2,t_VEC); z[2]=(long)p1;
    p1[1]=lcopy(gdeux); p1=cgetg(2,t_VEC); z[3]=(long)p1; p1[1]=lneg(gun);
    return z;
  }
  n=absi(n);
  if (cmpis(n,2)<=0)
  {
    z=cgetg(4,t_VEC); z[1]=un; z[2]=lgetg(1,t_VEC); z[3]=lgetg(1,t_VEC);
    return z;
  }
  fa=factor(n); list=(GEN)fa[1]; ep=(GEN)fa[2]; nbp=lg(list)-1;
  switch(mod8(n))
  {
    case 0:
      z=cgetg(3,t_VEC); z[1]=(long)gmul2n(gun,itos((GEN)ep[1])-2);
      z[2]=deux; gen2=cgetg(3,t_VEC); gen2[1]=lstoi(5);
      gen2[2]=(long)addis(gmul2n((GEN)z[1],1),-1);
      jlist=2; break;
    case 4:
      z=cgetg(2,t_VEC); z[1]=deux; gen2=cgetg(2,t_VEC); gen2[1]=lstoi(3);
      jlist=2; break;
    case 2: case 6:
      z=cgetg(1,t_VEC); gen2=cgetg(1,t_VEC); jlist=2; break;
    case 1: case 3: case 5: case 7:
      z=cgetg(1,t_VEC); gen2=cgetg(1,t_VEC); jlist=1; break;
  }
  sizeh=lg(z)+nbp-jlist;
  h=cgetg(sizeh+1,t_VEC);
  generators=cgetg(sizeh+1,t_VEC);
  moduli=cgetg(sizeh+1,t_VEC);
  for (i=1; i<lg(z); i++)
  {
    h[i]=z[i]; generators[i]=gen2[i];
    moduli[i]=(long)gmul2n(gun,itos((GEN)ep[1]));
  }
  for (j=jlist; j<=nbp; i++,j++)
  {
    p=(GEN)list[j]; epp=itos((GEN)ep[j]);
    h[i]=(long)mulii(addis(p,-1),q=gpuigs(p,epp-1));
    moduli[i]=(long)mulii(p,q);
    generators[i]=gener((GEN)moduli[i])[2];
  }
  for (i=1; i<=sizeh; i++)
  {
    q=(GEN)moduli[i]; a=(GEN)generators[i];
    if (!gcmp1(bezout(q,divii(n,q),&u,&v))) err(talker,"bug in znstar");
    generators[i]=lmodulcp(addii(a,mulii(mulii(subii(gun,a),u),q)),n);
  }
  for (i=sizeh; i>=2; i--)
  {
    for (j=i-1; j>=1; j--)
    {
      if (!divise((GEN)h[j],(GEN)h[i]))
      {
	d=bezout((GEN)h[i],(GEN)h[j],&u,&v);
	h[j]=(long)mulii((GEN)h[i],q=divii((GEN)h[j],d)); h[i]=(long)d;
	generators[j]=(long)gdiv((GEN)generators[j],(GEN)generators[i]);
	generators[i]=lmul((GEN)generators[i],
			   gpui((GEN)generators[j],mulii(v,q),0));
      }
    }
  }
  q=gun; for (i=1; (i<=sizeh)&&(!gcmp1((GEN)h[i])); i++) q=mulii(q,(GEN)h[i]);
  c=i-1; tetpil=avma;
  z=cgetg(4,t_VEC); z[1]=lcopy(q);
  p1=cgetg(c+1,t_VEC); z[2]=(long)p1;
  for (i=1; i<=c; i++) p1[i]=lcopy((GEN)h[i]);
  p1=cgetg(c+1,t_VEC); z[3]=(long)p1;
  for (i=1; i<=c; i++) p1[i]=lcopy((GEN)generators[i]);
  return gerepile(av,tetpil,z);
}

/*********************************************************************/
/**                                                                 **/
/**                     INTEGRAL SQUARE ROOT                        **/
/**                                                                 **/
/*********************************************************************/

GEN
gracine(GEN a)
{
  return garith_proto(racine,a);
}

GEN
racine(GEN a)
{
  GEN x,y,z;
  long av,av2,k, count = 0;

  if (typ(a) != t_INT) err(arither1);
  switch (signe(a))
  {
    case 0: return gzero;
    case -1:
      x=cgetg(3,t_COMPLEX); x[1]=zero;
      setsigne(a,1); x[2]= (long) racine(a); setsigne(a,-1);
      return x;

    case 1:
      k = (long) sqrt((double)(uLong)a[2]);
      x=gshifts(k+1,(lgefint(a)-3)*(BITS_IN_LONG/2));
      av=avma; y=cgeti(lgefint(x)); av2=avma;
      do
      {
	gops2gsz(shifti,addii(x,divii(a,x)),-1,y);
	z=y; y=x; x=z;
	avma=av2; count++;
      }
      while (cmpii(x,y)<0);
      if (odd(count)) x=y; else affii(y,x);
      avma=av; return x;
  }
  return NULL; /* not reached */
}

/* prototype of positive small ints */
static long pos_s[] = {
  evaltyp(t_INT) | evalpere(1) | evallg(3),
  evalsigne(1) | evallgefint(3), 0 };

/* prototype of negative small ints */
static long neg_s[] = {
  evaltyp(t_INT) | evalpere(1) | evallg(3),
  evalsigne(-1) | evallgefint(3), 0 };

GEN
gshifts(long x, long y)
{
  if (!x) return gzero;
  if (x>0) { pos_s[2]=x; return shifti(pos_s,y); }
  neg_s[2] = -x; return shifti(neg_s,y);
}

/*********************************************************************/
/**                                                                 **/
/**                      PERFECT SQUARE                             **/
/**                                                                 **/
/*********************************************************************/

GEN
gcarrecomplet(GEN x, GEN *pt)
{
  long tx = typ(x),l,i;
  GEN z,y,p,t;

  if (is_matvec_t(tx))
  {
    l=lg(x); y=cgetg(l,tx); z=cgetg(l,tx);
    for (i=1; i<l; i++)
    {
      t=gcarrecomplet((GEN)x[i],&p); y[i]=(long)t;
      z[i]=gcmp0(t) ? zero: (long)p;
    }
    *pt=z; return y;
  }
  if (tx != t_INT) err(arither1);
  return stoi(carrecomplet(x,pt));
}

long
carrecomplet(GEN x, GEN *pt)
{

  static carresmod64[]={ 1,1,0,0,1,0,0,0,0,1,0,0,0,0,0,0,
			1,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,
			0,1,0,0,1,0,0,0,0,1,0,0,0,0,0,0,
			0,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0 };

  static carresmod63[]={ 1,1,0,0,1,0,0,1,0,1,0,0,0,0,0,0,
                        1,0,1,0,0,0,1,0,0,1,0,0,1,0,0,0,
			0,0,0,0,1,1,0,0,0,0,0,1,0,0,1,0,
			0,1,0,0,0,0,0,0,0,0,1,0,0,0,0 };

  static carresmod65[]={ 1,1,0,0,1,0,0,0,0,1,1,0,0,0,1,0,
			1,0,0,0,0,0,0,0,0,1,1,0,0,1,1,0,
			0,0,0,1,1,0,0,1,1,0,0,0,0,0,0,0,
			0,1,0,1,0,0,0,1,1,0,0,0,0,1,0,0,1 };

  static carresmod11[]={ 1,1,0,1,1,1,0,0,0,1,0 };

  GEN y;
  long av,tetpil;

  switch(signe(x))
  {
    case -1: return 0;
    case 0: *pt=gzero; return 1;
    case 1: if (!carresmod64[mod64(x)]) return 0;
  }
  av=avma;
  if (!carresmod63[smodis(x,63)]) return 0;
  if (!carresmod65[smodis(x,65)]) return 0;
  if (!carresmod11[smodis(x,11)]) return 0;
  y=racine(x);

  if (cmpii(sqri(y),x)) { avma=av; return 0; }
  tetpil=avma; *pt=gerepile(av,tetpil,gcopy(y)); return 1;
}

#define eps(t) (((signe(t)*(t[lgefint(t)-1]))&3)==3)
#define ome(t) ( mod8(t) == 3 || mod8(t) == 5 )

GEN
gcarreparfait(GEN x)
{
  GEN p1,p2,p3,p4;
  long tx=typ(x),l,i,av,v;

  switch(tx)
  {
    case t_INT:
      return stoi(carreparfait(x));

    case t_REAL:
      return (signe(x)>=0)? gun: gzero;

    case t_INTMOD:
      if (!signe(x[2])) return gun;
      av=avma; p1=factor(absi((GEN)x[1]));
      p2=(GEN)p1[1]; l=lg(p2); p3=(GEN)p1[2];
      p4=cgeti(lgefint(x[2]));
      for (i=1; i<l; i++)
      {
	v=pvaluation((GEN)x[2],(GEN)p2[i],&p4);
	if (v<itos((GEN)p3[i]))
	{
	  if (v&1) { avma=av; return gzero; }
	  if (gegal((GEN)p2[i],gdeux))
	  {
	    v=itos((GEN)p3[i])-v;
	    if ((v>=3 && mod8(p4) != 1 ) ||
	        (v==2 && mod4(p4) != 1)) { avma=av; return gzero; }
	  }
	  else if (kronecker(p4,(GEN)p2[i]) == -1) { avma=av; return gzero; }
	 }
      }
      avma=av; return gun;

    case t_FRAC: case t_FRACN:
      av=avma; l=carreparfait(mulii((GEN)x[1],(GEN)x[2]));
      avma=av; return stoi(l);

    case t_COMPLEX:
      return gun;

    case t_PADIC:
      p4=(GEN)x[4]; if (!signe(p4)) return gun;
      if (valp(x)&1) return gzero;
      if (gegal((GEN)x[2],gdeux))
      {
	v=precp(x);
	if ((v>=3 && mod8(p4) != 1 ) ||
	    (v==2 && mod4(p4) != 1)) return gzero;
	return gun;
      }
      return (kronecker((GEN)x[4],(GEN)x[2])== -1)?gzero:gun;

    case t_POL:
      if (!signe(x)) return gun;
      l=lgef(x)-3; if (l&1) return gzero;
      av=avma; p2=(GEN)x[2];
      if (!signe(gcarreparfait(p2))) return gzero;
      p1=gtrunc(gsqrt(gdiv(gadd(x,ggrando(polx[varn(x)],l+1)),p2),DEFAULTPREC));
      v=gegal(gmul(p2,gmul(p1,p1)),x); avma=av; return v?gun:gzero;

    case t_SER:
      if (!signe(x)) return gun;
      if (valp(x)&1) return gzero;
      return gcarreparfait((GEN)x[2]);

    case t_RFRAC: case t_RFRACN:
      av=avma;
      l=itos(gcarreparfait(gmul((GEN)x[1],(GEN)x[2])));
      avma=av; return stoi(l);

    case t_QFR: case t_QFI:
      return gcarreparfait((GEN)x[1]);

    case t_VEC: case t_COL: case t_MAT:
      l=lg(x); p1=cgetg(l,tx);
      for (i=1; i<l; i++) p1[i]=(long)gcarreparfait((GEN)x[i]);
      return p1;
  }
  err(impl,"issquare for this type");
  return NULL; /* not reached */
}

long
carreparfait(GEN x)
{
  long av=avma,f;
  GEN p1;

  f=carrecomplet(x,&p1); avma=av; return f;
}

/*********************************************************************/
/**                                                                 **/
/**                        KRONECKER SYMBOL                         **/
/**                                                                 **/
/*********************************************************************/

GEN
gkronecker(GEN x, GEN y)
{
  return arith_proto2(kronecker,x,y);
}

long
kronecker(GEN x, GEN y)
{
  GEN z;
  long av,r,s=1;

  av=avma;
  switch (signe(y))
  {
    case -1: y=negi(y); if (signe(x)<0) s= -1; break;
    case 0: return is_pm1(x);
  }
  r=vali(y);
  if (r)
    if (mpodd(x))
    {
      if (odd(r) && ome(x)) s= -s;
      y=shifti(y,-r);
    }
    else { avma=av; return 0; }
  x=modii(x,y);
  while (signe(x))
  {
    r=vali(x);
    if (r)
    {
      if (odd(r) && ome(y)) s= -s;
      x=shifti(x,-r);
    }
    if (y[lgefint(y)-1]&2 && x[lgefint(x)-1]&2)
      s= -s; /* x=3 mod 4 && y=3 mod 4 since both are odd here */
    z=resii(y,x); y=x; x=z;
  }
  avma=av; return cmpsi(1,y) ? 0 : s;
}

GEN
gkrogs(GEN x, long y)
{
  return arith_proto2gs(krogs,x,y);
}

long
krogs(GEN x, long y)
{
  long av,r,s=1,x1,z;

  av=avma;
  if (y<=0)
  {
    if (y) { y = -y; if (signe(x)<0) s = -1; }
    else  return is_pm1(x);
  }
  r=vals(y);
  if (r)
    if (mpodd(x))
    {
      if (odd(r) && ome(x)) s= -s;
      y>>=r;
    }
    else return 0;
  x1=smodis(x,y);
  while (x1)
  {
    r=vals(x1);
    if (r)
    {
      if (odd(r) && (labs((y&7)-4)==1)) s= -s;
      x1>>=r;
    }
    if (y&2 && x1&2) s= -s;
    z = y % x1; y=x1; x1=z;
  }
  avma=av; return (y==1) ? s : 0;
}

long
krosg(long s, GEN x)
{
  long av,y;

  av=avma; y=kronecker(stoi(s),x); avma=av; return y;
}

long
kross(long x, long y)
{
  long r,s=1,x1,z;

  if (y<=0)
  {
    if (y) { y= -y; if (x<0) s = -1; }
    else  return (labs(x)==1);
  }
  r=vals(y);
  if (r)
    if (odd(x))
    {
      if (odd(r) && labs((x&7)-4) == 1) s = -s;
      y>>=r;
    }
    else return 0;
  x1=x%y; if (x1<0) x1+=y;
  while (x1)
  {
    r=vals(x1);
    if (r)
    {
      if (odd(r) && labs((y&7)-4) == 1) s= -s;
      x1>>=r;
    }
    if (y&2 && x1&2) s= -s;
    z=y%x1; y=x1; x1=z;
  }
  return (y==1) ? s : 0;
}

long
hil0(GEN x, GEN y, GEN p)
{
  return (gcmp1(p)) ? hil(x,y,gzero) : hil(x,y,p);
}

long
hil(GEN x, GEN y, GEN p)
{
  long a,b,av,tx,ty,z;
  GEN p1,p2,u,v;

  if (gcmp0(x) || gcmp0(y)) return 0;
  tx=typ(x); ty=typ(y);
  if (tx>ty) { p1=x; x=y; y=p1; av=tx,tx=ty; ty=av; }
  av=avma;
  switch(tx)
  {
    case t_INT:
      switch(ty)
      {
	case t_INT:
	  if (signe(p)<=0)
	    return (signe(x) < 0 && signe(y)<0) ? -1 : 1;
	  a=pvaluation(x,p,&u); b=pvaluation(y,p,&v);
	  if (cmpsi(2,p))
	  {
	    z = (odd(a) && odd(b) && eps(p))? -1:1;
	    if (odd(a) && kronecker(v,p)<0) z= -z;
	    if (odd(b) && kronecker(u,p)<0) z= -z;
	  }
	  else
	  {
	    z = (eps(u) && eps(v)) ? -1:1;
	    if (odd(a) && ome(v)) z= -z;
	    if (odd(b) && ome(u)) z= -z;
	  }
	  avma=av; return z;
	case t_REAL:
	  z=(signe(x)<0 && signe(y)<0) ? -1 : 1; return z;
	case t_INTMOD:
	  if (!cmpsi(2,(GEN)y[1])) err(hiler1);
	  return hil(x,(GEN)y[2],(GEN)y[1]);
	case t_FRAC: case t_FRACN:
	  p1=mulii((GEN)y[1],(GEN)y[2]); z=hil(x,p1,p);
	  avma=av; return z;
	case t_PADIC:
	  if (!cmpsi(2,(GEN)y[2]) && precp(y) <= 2) err(hiler1);
	  p1 = (odd(valp(y)))? mulii((GEN)y[2],(GEN)y[4]): (GEN)y[4];
	  z=hil(x,p1,(GEN)y[2]); avma=av; return z;

	default:
	  err(hiler2);
      }

    case t_REAL:
      if (! is_frac_t(ty)) err(hiler2);
      if (signe(x)>0) return 1;
      return signe(y[1])*signe(y[2]);

    case t_INTMOD:
      if (!cmpsi(2,(GEN)y[1])) err(hiler1);
      switch(ty)
      {
        case t_INTMOD:
          if (cmpii((GEN)x[1],(GEN)y[1])) err(hiler2);
          return hil((GEN)x[2],(GEN)y[2],(GEN)x[1]);
        case t_FRAC: case t_FRACN:
	  return hil((GEN)x[2],y,(GEN)x[1]);
        case t_PADIC:
          if (cmpii((GEN)x[1],(GEN)y[2])) err(hiler2);
          return hil((GEN)x[2],y,(GEN)x[1]);
        default:
	  err(hiler2);
      }

    case t_FRAC: case t_FRACN:
      p1=mulii((GEN)x[1],(GEN)x[2]);
      switch(ty)
      {
	case t_FRAC: case t_FRACN:
	  p2=mulii((GEN)y[1],(GEN)y[2]);
	  z=hil(p1,p2,p); avma=av; return z;
	case t_PADIC:
	  z=hil(p1,y,0); avma=av; return z;
	default:
	  err(hiler2);
      }

    case t_PADIC:
      if (ty != t_PADIC || cmpii((GEN)x[2],(GEN)y[2])) err(hiler2);
      p1 = (odd(valp(x))) ? mulii((GEN)x[2],(GEN)x[4]) : (GEN)x[4];
      p2 = (odd(valp(y))) ? mulii((GEN)y[2],(GEN)y[4]): (GEN)y[4];
      z=hil(p1,p2,(GEN)x[2]); avma=av; return z;

  }
  err(hiler2);
  return 0; /* not reached */
}

/*******************************************************************/
/*                                                                 */
/*                       SQUARE ROOT MODULO p                      */
/*                                                                 */
/*******************************************************************/

#define sqrmod(x,p) (modii(sqri(x),p))

int
mpsqrtmod(GEN a, GEN p, GEN *pr)
{
  long  l,av0,av1,av3,f,i,k,e,r;
  GEN   p1,p2,p3,m,m1,v,y,w,n;

  if (typ(a) != t_INT || typ(p) != t_INT) err(arither1);
  l=lg(p); av0=avma; v=cgeti(l);
  av1=avma; w=cgeti(l); y=cgeti(l);
  p1=addsi(-1,p); e=vali(p1);
  p2=shifti(p1,-e); n=stoi(2); av3=avma;
  if (e==1) affii(p1,y);
  else
    do
    {
      m=puissmodulo(n,p2,p); m1=m;
      for (i=1,f=0; i<e && !f; i++)
	f=gcmp1(m=sqrmod(m,p));
      if (f) addsiz(1,n,n); else  affii(m1,y);
      avma=av3;
    }
    while (f);
/*  y contient un generateur de (Z/pZ)* eleve a la puis (p-1)/2^e  */

  p1=shifti(p2,-1); p2=puissmodulo(a,p1,p);
  if (!signe(p2)) { avma=av0; *pr=gzero; return 1; }

  p3=mulii(p2,a); modiiz(p3,p,v);
  p1=mulii(mulii(p2,p2),a);
  modiiz(p1,p,w); avma=av3; r=e;
  while (!gcmp1(w))
  {
    /* if p is not prime, next loop will not end !! */
    k=1 ; p1=w; while (!gcmp1(p1=sqrmod(p1,p))) k++;
    if (k==r) { avma=av0; return 0; }
    avma=av3;
    for (i=1,p1=y; i<r-k; i++)  p1=sqrmod(p1,p);
    modiiz(mulii(p1,v),p,v);
    p1=sqrmod(p1,p); modiiz(mulii(p1,w),p, w);
    affii(p1,y); avma=av3; r=k;
  }
  p1=subii(p,v); if (cmpii(v,p1)>0) affii(p1,v);
  *pr=v; avma=av1; return 1;
}

/*********************************************************************/
/**                                                                 **/
/**                        GCD & BEZOUT                             **/
/**                                                                 **/
/*********************************************************************/

GEN
mppgcd(GEN a, GEN b)
{
  GEN t;
  long av=avma,tetpil,lim,st,k,va,vb;

  if (typ(a) != t_INT || typ(b) != t_INT) err(arither1);
  if (lgefint(a) < lgefint(b)) { t=b; b=a; a=t; }
  b=absi(b); tetpil=avma; a=absi(a);
  if (signe(b)) { t=resii(a,b); a=b; b=t; } else return gerepile(av,tetpil,a);
  if (!signe(b)) { avma=tetpil; return a; }

  lim=(avma+3*bot)>>2;
  va=vali(a); vb=vali(b); k=min(va,vb);
  a=shifti(a,-va); b=shifti(b,-vb);
  t=subii(a,b); st=signe(t); if (st) setsigne(t,1);

  while (st)
  {
    t=shifti(t,-vali(t));
    if (st>0) a=t; else b=t;
    t=subii(a,b); st=signe(t); if (st) setsigne(t,1);
    if (avma<lim)
    {
      GEN *gptr[3];

      gptr[0]=&t; gptr[1]=&a; gptr[2]=&b;
      gerepilemany(av,gptr,3);
    }
  }
  tetpil=avma; return gerepile(av,tetpil,shifti(a,k));
}

GEN
bezout(GEN a, GEN b, GEN *u, GEN *v)
{
  GEN p,v1,d1,d,q,uu,vv;
  long av,av2,count=0;

  if (typ(a) != t_INT || typ(b) != t_INT) err(arither1);
  if (lgefint(b) > lgefint(a))
  {
    GEN u1,*bof;
    u1=b; b=a; a=u1; bof=u; u=v; v=bof;
  }
  d=gcopy(a); if (!signe(b)) { *u=gun; *v=gzero; return d; }

  *u = uu = cgeti(lgefint(b));
  *v = vv = cgeti(lgefint(a));
  affsi(0,vv); av=avma;
  v1=cgeti(lgefint(a)); affsi(1,v1);
  d1=gcopy(b); q=cgeti(lgefint(a));
  av2=avma;

  while (signe(d1))
  {
    dvmdiiz(d,d1,q,d);
    subiiz(vv,mulii(q,v1),vv);
    p=vv; vv=v1; v1=p;
    p=d; d=d1; d1=p;
    avma=av2; count++;
  }
  if (odd(count))
  {
    affii(vv,v1); vv=v1; affii(d,d1); d=d1;
  }

  diviiz(subii(d,mulii(b,vv)),a,uu);
  avma=av;
  if (signe(d)== -1)
  {
    setsigne(d,1); setsigne(uu,-signe(uu)); setsigne(vv,-signe(vv));
  }
  return d;
}

/*********************************************************************/
/**                                                                 **/
/**                      CHINESE REMAINDERS                         **/
/**                                                                 **/
/*********************************************************************/

GEN
chinois(GEN x, GEN y)
/*  P.M. & M.H.
 *
 *      Chinese Remainder Theorem.  x and y must have the same type
 *      (integermod, polymod, or polynomial/vector/matrix recursively
 *      constructed with these as coefficients).  Creates (with the same
 *      type) a z in the same residue class as x and the same residue class
 *      as y, if it is possible.
 *
 *      We also allow (during recursion) two identical objects
 *      even if they are not integermod or polymod.  For example, if
 *
 *      	x = [1. mod(5, 11), mod(X + mod(2, 7), X^2 + 1)]
 *      	y = [1, mod(7, 17), mod(X + mod(0, 3), X^2 + 1)],
 *
 *      then chinois(x, y) returns
 *
 *      	    [1, mod(16, 187), mod(X + mod(9, 21), X^2 + 1)]
 *
 *      Someone else may want to allow power series,
 *      complex numbers, and quadratic numbers.
 */
{
  const long tx=typ(x);
  long i,lx,vx,av,tetpil;
  GEN z,p1,p2,p3,p4,d,u,v;

  if (tx != typ(y))  err(chiner);

  if (gegal(x,y)) return gcopy(x); /* Equal objects? */
  switch(tx)
  {
    default: err(chiner);

    case t_POLMOD:
      if (gegal((GEN)x[1],(GEN)y[1]))  /* Polymods with same modulus */
      {
	z=cgetg(3,tx);
	z[1]=lcopy((GEN)x[1]);
	z[2]=(long)chinois((GEN)x[2],(GEN)y[2]);
	break;
      }  /* Otherwise fall through to integermod code */
    case t_INTMOD:
      {
        z=cgetg(3,tx); av=avma;
        d=gbezout((GEN)x[1],(GEN)y[1],&u,&v);
        if (!gegal(gmod((GEN)x[2],d),gmod((GEN)y[2],d))) err(chiner);
        p1=gdiv((GEN)x[1],d);
        p2=gdiv((GEN)y[1],d);
        p3=gmul((GEN)y[2],gmul(u,p1));
        p4=gmul((GEN)x[2],gmul(v,p2));
        p3=gadd(p3,p4);

        tetpil=avma;
	z[1]=lmul((GEN)x[1],p2); z[2]=lmod(p3,(GEN)z[1]);
	gerepilemanyvec(av,tetpil,z+1,2);
      }
      break;

    case t_POL:
      lx=lgef(x); vx=varn(x);
      if ((lx!=lgef(y))||(vx!=varn(y))) err(chiner);
      z=cgetg(lx,tx);
      setsigne(z,1); setlgef(z,lx); setvarn(z,vx);
      for (i=2; i<lx; i++) z[i]=(long)chinois((GEN)x[i],(GEN)y[i]);
      break;

    case t_VEC: case t_COL: case t_MAT:
      lx=lg(x); if (lx!=lg(y)) err(chiner);
      z=cgetg(lx,tx);
      for (i=1; i<lx; i++) z[i]=(long)chinois((GEN)x[i],(GEN)y[i]);
      break;
  }
  return z;
}

/*********************************************************************/
/**                                                                 **/
/**                      INVERSE MODULO b                           **/
/**    If a is invertible, return 1, and set res  = a^{ -1 }        **/
/**    Otherwise, return 0, and set res = gcd(a,b)                  **/
/**                                                                 **/
/*********************************************************************/

int
inversemodulo(GEN a, GEN b, GEN *res)
{
  GEN u,u1,d1,q;
  long av,av2,count=0;

  if (typ(a) != t_INT || typ(b) != t_INT) err(arither1);
  if (!signe(b)) { *res=absi(a); return 0; }
  *res=absi(b); av=avma;
  u=cgeti(lgefint(b)); affsi(0,u);
  u1=cgeti(lgefint(b)); affsi(1,u1);
  d1=cgeti(lgefint(b)); modiiz(a,*res,d1);
  q=cgeti(lgefint(b));
  av2=avma;

  while (signe(d1))
  {
    dvmdiiz(*res,d1,q,*res);
    subiiz(u,mulii(q,u1),u);
    a=u; u=u1; u1=a;
    a= *res; *res=d1; d1=a;
    avma=av2; count++;
  }
  if (odd(count))
  {
    affii(*res,d1); *res=d1; affii(u,u1); u=u1;
  }

  if (cmpis(*res,1)) { avma=av; return 0; }
  modiiz(u,b,*res); avma=av; return 1;
}

GEN
mpinvmod(GEN a, GEN m)
{
  GEN res;

  if (inversemodulo(a,m,&res)) return res;
  err(invmoder,"",gmodulcp(res,m));
  return NULL; /* not reached */
}


/*********************************************************************/
/**                                                                 **/
/**                      POWERING MODULO                            **/
/**     a ^ n mod m (results takes as much place as modulus         **/
/**                                                                 **/
/*********************************************************************/

GEN
puissmodulo(GEN a, GEN n, GEN m)
{
  GEN res,aux;
  long av,*p,man,k,nb;

  if (typ(a) != t_INT || typ(n) != t_INT || typ(m) != t_INT) err(arither1);
  res=cgeti(lgefint(m)); av=avma;
  switch (signe(n))
  {
    case 0: if (divise(a,m)) affsi(0,res); else affsi(1,res); break;
    case -1:
      if (inversemodulo(a,m,&aux))
      {
	affii(puissmodulo(aux,absi(n),m),res);
	avma=av;
      }
      else err(puissmoder);
      break;

    case 1: modiiz(a,m,res); p=n+2; k = BITS_IN_LONG-1;
      for (man = *p; man>0; man<<=1) k--;
      man<<=1; /* le premier bit est implicite */
      for (nb=lgefint(n)-2; nb; man= *++p,k=BITS_IN_LONG,nb--)
        for (; k; man<<=1,k--)
	{
	  modiiz(sqri(res),m,res);
	  if (man<0) modiiz(mulii(res,a),m,res);
	  avma=av;
	 }
  }
  return res;
}

/*********************************************************************/
/**                                                                 **/
/**                      PSEUDO PRIMALITY                           **/
/**               n strong pseudo-prime for the base a ?            **/
/**                                                                 **/
/*********************************************************************/

GEN
gpseudopremier(GEN n, GEN a)
{
  return arith_proto2(pseudopremier,n,a);
}

long
pseudopremier(GEN n, GEN a)
{
  long r, av = avma;
  GEN t,z,c,b;

  t=subis(absi(n),1); r=vali(t);
  if (!r) { avma = av; return !cmpsi(1,t); } /* n even, test n=2 or -2 */

  b=puissmodulo(a,shifti(t,-r),n);
  if (cmpsi(1,b))
  {
    c=cgeti(lgefint(n));
    for (; r; r--)
    {
      modiiz(sqri(b),n,c);
      z=b; b=c; c=z;
      if (!cmpsi(1,b)) break;
    }
    if (!r || cmpii(c,t)) { avma=av; return 0; }
  }
  avma=av; return 1;
}

/* test k times that n is pseudo-prime */
GEN
gmillerrabin(GEN n, long k)
{
  return arith_proto2gs(millerrabin,n,k);
}

long
millerrabin(GEN n, long k)
{
  long r,r1,av2, av = avma;
  GEN t,t1,z,c,b;

  t=subis(absi(n),1); r1=vali(t);
  if (!r1) { avma = av; return !cmpsi(1,t); } /* n even, test n=2 or -2 */

  b = cgeti(lgefint(n)); c = cgeti(lgefint(n));
  t1 = shifti(t,-r1); av2=avma;
  for (; k; k--)
  {
    do z = modii(stoi(mymyrand()),n); while (!signe(z));
    affii(puissmodulo(z,t1,n),b);
    if (cmpsi(1,b))
    {
      for (r=r1; r; r--)
      {
	modiiz(sqri(b),n,c);
	z=b; b=c; c=z;
	if (!cmpsi(1,b)) break;
      }
      if (!r || cmpii(c,t)) { avma=av; return 0; }
    }
    avma=av2;
  }
  avma=av; return 1;
}

/* Miller-Rabin test for k bases taken in lispr (i.e not random).
 * We must have k<11 (unchecked).
 */
static int
miller(GEN n, long k)
{
  long r,r1,av2, av = avma;
  GEN t,t1,z,c,b;
  static long lispr[10] = { 2,3,5,7,11,13,17,19,23,29 };

  t=subis(absi(n),1); r1=vali(t);
  if (!r1) { avma = av; return !cmpsi(1,t); } /* n even, test n=2 or -2 */

  b = cgeti(lgefint(n)); c = cgeti(lgefint(n));
  t1 = shifti(t,-r1); av2=avma;
  for (; k; k--)
  {
    z = modsi(lispr[k],n);
    if (signe(z))
    {
      affii(puissmodulo(z,t1,n),b);
      if (cmpsi(1,b))
      {
	for (r=r1; r; r--)
	{
	  modiiz(sqri(b),n,c);
	  z=b; b=c; c=z;
	  if (!cmpsi(1,b)) break;
	 }
	if (!r || cmpii(c,t)) { avma=av; return 0; }
      }
      avma=av2;
    }
  }
  avma=av; return 1;
}

GEN
gnextprime(GEN n)
{
  return garith_proto(nextprime,n);
}

GEN
nextprime(GEN n)
{
  long av1,av2,av3, toadd, do_copy;

  if (typ(n) != t_INT) err(arither1);
  if (cmpsi(2,n) >= 0) return gdeux;
  if (lgefint(n) == 3 && n[2] == 3) return stoi(3); /* if n=3 */

  av1=avma; divis(n,6); /* here n>3 */
  switch(hiremainder)
  {
    case 0: do_copy=0; n=addsi(1,n); toadd=4; break;
    case 1: do_copy=1; toadd=4; break;
    case 2: do_copy=0; n=addsi(3,n); toadd=2; break;
    case 3: do_copy=0; n=addsi(2,n); toadd=2; break;
    case 4: do_copy=0; n=addsi(1,n); toadd=2; break;
    case 5: do_copy=1; toadd=2; break;
  }
  av3 = av2 = avma;
  for(;;)
  {
    if (miller(n,10)) break;
    av2=avma; n=addsi(toadd,n); toadd = 6-toadd;
  }
  if (av2!=av3) return gerepile(av1,av2,n);
  if (do_copy) return gcopy(n);
  return n;
}


GEN
gprecprime(GEN n)
{
  return garith_proto(precprime,n);
}

GEN
precprime(GEN n)
{
  long av1,av2,av3, tosub, do_copy;

  if (typ(n) != t_INT) err(arither1);
  if (cmpsi(1,n) >= 0) return gzero;
  if (cmpsi(3,n) >= 0) return gcopy(n);
  if (lgefint(n) == 3 && n[2] == 4) return stoi(3); /* if n=4 */

  av1=avma; divis(n,6); /* here n>4 */
  switch(hiremainder)
  {
    case 0: do_copy=0; n=addsi(-1,n); tosub=4; break;
    case 1: do_copy=1; tosub=4; break;
    case 2: do_copy=0; n=addsi(-1,n); tosub=2; break;
    case 3: do_copy=0; n=addsi(-2,n); tosub=2; break;
    case 4: do_copy=0; n=addsi(-3,n); tosub=2; break;
    case 5: do_copy=1; tosub=4; break;
  }
  av3 = av2 = avma;
  for(;;)
  {
    if (miller(n,10)) break;
    av2=avma; n=addsi(-tosub,n); tosub = 6-tosub;
  }
  if (av2!=av3) return gerepile(av1,av2,n);
  if (do_copy) return gcopy(n);
  return n;
}

static long
isprime_proto(GEN x,long k)
{
  long av=avma;

  x=absi(x); avma=av;
  if (gcmpgs(x,3)<=0) return !gcmp1(x);
  if (!mpodd(x)) return 0;
  return millerrabin(x,k);
}

GEN
gisprime(GEN x)
{
  return arith_proto(isprime,x,1);
}

long
isprime(GEN x)
{
  return isprime_proto(x,10);
}

GEN
gispsp(GEN x)
{
  return arith_proto(ispsp,x,1);
}

long
ispsp(GEN x)
{
  return isprime_proto(x,1);
}

/*********************************************************************/
/**                                                                 **/
/**                    FUNDAMENTAL DISCRIMINANTS                    **/
/**                                                                 **/
/*********************************************************************/

GEN
gissquarefree(GEN x)
{
  return arith_proto(issquarefree,x,0);
}

long
issquarefree(GEN x)
{
  long av=avma,tx,lx,i;
  GEN p1;

  if (gcmp0(x)) return 0;
  tx=typ(x);
  if (tx == t_INT)
  {
    p1= (GEN) (auxdecomp(x,1)[2]); lx=lg(p1);
    i=1; while (i<lx && gcmp1((GEN)p1[i])) i++;
    avma=av; return (i==lx);
  }
  if (tx != t_POL) err(issquer1);
  p1=ggcd(x,deriv(x,varn(x)));
  avma=av; return (lgef(p1)==3);
}

GEN
gisfundamental(GEN x)
{
  return arith_proto(isfundamental,x,1);
}

long
isfundamental(GEN x)
{
  long av,r;
  GEN p1;

  if (gcmp0(x)) return 0;
  r=mod4(x);
  if (!r)
  {
    av=avma; p1=shifti(x,-2); avma=av;
    r=mod4(p1); if (!r) return 0;
    if (signe(x)<0) r=4-r;
    return  (r==1)? 0: issquarefree(p1);
  }
  if (signe(x)<0) r=4-r;
  return (r==1) ? issquarefree(x) : 0;
}

GEN
quaddisc(GEN x)
{
  long av=avma,tetpil,i,r,tx=typ(x);
  GEN p1,p2,f,s;

  if (tx != t_INT && ! is_frac_t(tx)) err(arither1);
  f=factor(x); p1=(GEN)f[1]; p2=(GEN)f[2];
  s=gun;
  for (i=1; i<lg(p1); i++)
    if ( odd(mael(p2,i,2)) ) { tetpil=avma; s=gmul(s,(GEN)p1[i]); }
  r=mod4(s); if (signe(x)<0) r=4-r;
  if (r>1) { tetpil=avma; s=shifti(s,2); }
  return gerepile(av,tetpil,s);
}

/*********************************************************************/
/**                                                                 **/
/**                              FACTORIAL                          **/
/**                                                                 **/
/*********************************************************************/

GEN
mpfact(long n)
{
  long av,tetpil,lim,k;
  GEN f;

  if (n<2)
    if (n<0) err(facter); else return gun;
  lim = (bot+avma) >> 1; av = avma; f = gun;
  for (k=2; k<n; k++)
    if (avma<lim)
    {
      tetpil = avma;
      f = gerepile(av,tetpil,mulsi(k,f));
    }
    else f = mulsi(k,f);
  tetpil = avma; return gerepile(av,tetpil,mulsi(k,f));
}

GEN
mpfactr(long n, long prec)
{
  long av,tetpil,lim,k;
  GEN f = cgetr(prec);

  affsr(1,f);
  if (n<2)
    if (n<0) err(facter); else return f;
  lim = (bot+avma) >> 1; av = avma;
  for (k=2; k<n; k++)
    if (avma<lim)
    {
      tetpil = avma;
      f = gerepile(av,tetpil,mulsr(k,f));
    }
    else f = mulsr(k,f);
  tetpil = avma; return gerepile(av,tetpil,mulsr(k,f));
}

/*******************************************************************/
/**                                                               **/
/**                      LUCAS & FIBONACCI                        **/
/**                                                               **/
/*******************************************************************/

void
lucas(long n, GEN *ln, GEN *ln1)
{
  long taille,av;
  GEN z,t;

  if (!n) { *ln=stoi(2); *ln1=stoi(1); return; }

  taille=(long)(C3*(1+labs(n))+3);
  *ln=cgeti(taille); *ln1=cgeti(taille);
  av=avma; lucas(n/2,&z,&t);
  switch(n % 4)
  {
    case -3:
      addsiz(2,sqri(z),*ln1);
      subiiz(addsi(1,mulii(z,t)),*ln1,*ln); break;
    case -2:
      addsiz(2,sqri(z),*ln); addsiz(1,mulii(z,t),*ln1); break;
    case -1:
      subisz(sqri(z),2,*ln1);
      subiiz(subis(mulii(z,t),1),*ln1,*ln); break;
    case  0: subisz(sqri(z),2,*ln); subisz(mulii(z,t),1,*ln1); break;
    case  1: subisz(mulii(z,t),1,*ln); addsiz(2,sqri(t),*ln1); break;
    case  2: addsiz(2,sqri(z),*ln); addsiz(1,mulii(z,t),*ln1); break;
    case  3: addsiz(1,mulii(z,t),*ln); subisz(sqri(t),2,*ln1);
  }
  avma=av;
}

GEN
fibo(long n)
{
  long av;
  GEN ln,ln1,f;

  f=cgeti((long) (C3*labs(n)+3));
  av=avma; lucas(n-1,&ln,&ln1);
  divisz(addii(mulsi(2,ln),ln1),5,f);
  avma=av; return f;
}

/*******************************************************************/
/*                                                                 */
/*                      CONTINUED FRACTIONS                        */
/*                                                                 */
/*******************************************************************/
static GEN sfcont2(GEN b, GEN x, long k);

GEN
gcf(GEN x)
{
  return sfcont(x,x,0);
}

GEN
gcf2(GEN b, GEN x)
{
  return sfcont0(x,b,0);
}

GEN
gboundcf(GEN x, long k)
{
  return sfcont(x,x,k);
}

GEN
sfcont0(GEN x, GEN b, long flag)
{
  long lb,tb=typ(b),i,av,tetpil;
  GEN y;

  if (gcmp0(b)) return sfcont(x,x,flag);

  if (tb == t_INT) return sfcont(x,x,itos(b));

  if (! is_matvec_t(tb)) err(sfconter1);
  lb=lg(b); if (lb==1) return cgetg(1,t_VEC);
  if (tb == t_MAT)
  {
    if (lg(b[1])==1) return sfcont(x,x,flag);
    av=avma; y=cgetg(lb,t_VEC);
    for (i=1; i<lb; i++) y[i]=coeff(b,1,i);
    tetpil=avma; return gerepile(av,tetpil,sfcont2(y,x,flag));
  }
  return sfcont2(b,x,flag);
}

GEN
sfcont(GEN x, GEN x1, long k)
{
  long  av,lx=lg(x),tx=typ(x),e,i,j,l,l1,l2,lx1,tetpil,f,f1;
  GEN   y,p1,p2,p3,p4,yp;

  if (is_scalar_t(tx))
  {
    if (gcmp0(x)) { y=cgetg(2,t_VEC); y[1]=zero; return y; }
    switch(tx)
    {
      case t_INT:
        y=cgetg(2,t_VEC); y[1]=lcopy(x); return y;
      case t_REAL:
	if (gcmp0(x)) { y=cgetg(2,t_VEC); y[1]=zero; return y; }
        l=avma; p1=cgetg(3,t_FRACN);
	p2=gcopy(x); settyp(p2,t_INT); setlgefint(p2,lx);
	p1[1] = (long) p2; e = ((lx-2)<<TWOPOTBITS_IN_LONG)-1-expo(x);
	if (e<0) err(sfconter2);
	l1 = (e>>TWOPOTBITS_IN_LONG)+3; p2=cgeti(l1);
	p2[1]= evalsigne(1)+evallgefint(l1);
	p2[2]= (1L<<(e&(BITS_IN_LONG-1)));
	for (i=3; i<l1; i++) p2[i]=0;
	p1[2]=(long) p2; p3=cgetg(3,t_FRACN);
	p3[2]=lcopy(p2);
	p3[1]=laddsi(signe(x),(GEN)p1[1]);
	p1=sfcont(p1,p1,k); tetpil=avma;
	p3=sfcont(p3,p1,k);
	return gerepile(l,tetpil,p3);

      case t_FRAC: case t_FRACN:
        l=avma; lx1=lgefint(x[2]);
	l1 = (long) ((double) BYTES_IN_LONG/4.0*46.093443*(lx1-2)+3);
	if (k>0 && l1>k+1) l1=k+1;
	if (l1>LGBITS) l1=LGBITS;
	if (lgefint(x[1]) >= lx1) p1=gcopy((GEN)x[1]);
	else affii((GEN)x[1],p1=cgeti(lx1));
	p2=gcopy((GEN)x[2]); l2=avma; lx1=lg(x1);
	y=cgetg(l1,t_VEC); f1=1; f=(x!=x1); i=0;
	while (f1 && !gcmp0(p2) && i<=l1-2)
	{
	  i++; y[i]=ldvmdii(p1,p2,&p3);
	  if (signe(p3)<0)
	  {
	    p4=addii(p3,p2); affii(p4,p1);
	    cgiv(p4); cgiv(p3); cgiv((GEN)y[1]);
	    y[1]=laddsi(-1,(GEN)y[1]);
	  }
	  else { affii(p3,p1); cgiv(p3); }
	  p4=p1; p1=p2; p2=p4;
	  if (f)
	    f1=(i<lx1 && !cmpii((GEN)y[i],(GEN)x1[i]));
	}

	if (!f1)
	{
	  if (i>=lx1) --i;
	  else
	  {
	    av=avma;
	    if (gcmp1(absi(subii((GEN)x1[i],(GEN)y[i]))))
	    {
	      if (i>=lx1-1 || !gcmp1((GEN)x1[i+1]))
	        affii((GEN)x1[i],(GEN)y[i]);
	    }
	    else --i;
	    avma=av;
	  }
	}
	if (i>=2 && gcmp1((GEN)y[i]))
	{
	  cgiv((GEN)y[i]); --i; cgiv((GEN)y[i]);
	  y[i]=laddsi(1,(GEN)y[i]);
	}
	setlg(y,i+1); l2=l2-((l1-i-1)<<TWOPOTBYTES_IN_LONG);
	return gerepile(l,l2,y);
    }
    err(sfconter1);
  }

  switch(tx)
  {
    case t_POL:
      y=cgetg(2,t_VEC); y[1]=lcopy(x); break;
    case t_SER:
      av=avma; p1=gtrunc(x); tetpil=avma;
      y=gerepile(av,tetpil,sfcont(p1,p1,k)); break;
    case t_RFRAC:
    case t_RFRACN:
      av=avma; l1=lgef(x[1]); if (lgef(x[2])>l1) l1=lgef(x[2]);
      if (k>0 && l1>k+1) l1=k+1;
      yp=cgetg(l1,t_VEC); p1=(GEN)x[1]; i=0; p2=(GEN)x[2];
      while (!gcmp0(p2) && i<=l1-2)
      {
	i++; yp[i]=ldivres(p1,p2,&p3);
	p1=p2; p2=p3;
      }
      tetpil=avma; y=cgetg(i+1,t_VEC);
      for (j=1; j<=i; j++) y[j]=lcopy((GEN)yp[j]);
      y=gerepile(av,tetpil,y); break;
    default: err(sfconter1);
  }
  return y;
}

static GEN
sfcont2(GEN b, GEN x, long k)
{
  long av,tx,l1=lg(b),i,j,tetpil;
  GEN y,z,p1;

  if (k)
  {
    if (k>=l1) err(sfconter1);
    l1=k+1;
  }
  y=cgetg(l1,t_VEC);
  if (l1==1) return y;
  av=avma; tx=typ(x);
  if (is_scalar_t(tx))
  {
    if (tx >= t_COMPLEX || tx == t_INTMOD) err(sfconter1);
  }
  else if (tx == t_SER) x=gtrunc(x);
  if (!gcmp1((GEN)b[1])) x=gmul((GEN)b[1],x);
  y[1]=lfloor(x); p1=gsub(x,(GEN)y[1]);
  if (!gcmp0(p1))
    for (i=2; i<l1; i++)
    {
      x=gdiv((GEN)b[i],p1); y[i]=lfloor(x);
      p1=gsub(x,(GEN)y[i]); if (gcmp0(p1)) break;
    }
  tetpil=avma; z=cgetg(i,t_VEC);
  for (j=1; j<i; j++) z[j]=lcopy((GEN)y[j]);
  return gerepile(av,tetpil,z);
}

GEN
pnqn(GEN x)
{
  long av=avma,tetpil,lx,ly,tx=typ(x),i;
  GEN y,p0,p1,q0,q1,a,b,p2,q2;

  if (! is_matvec_t(tx)) err(pnqner1);
  lx=lg(x); if (lx==1) return idmat(2);
  if (tx != t_MAT)
  {
    p0=q1=gun; q0=gzero; p1=(GEN)x[1];
    for (i=2; i<lx; i++)
    {
      a=(GEN)x[i];
      p2=gadd(gmul(a,p1),p0); p0=p1; p1=p2;
      q2=gadd(gmul(a,q1),q0); q0=q1; q1=q2;
    }
    tetpil=avma; y=cgetg(3,t_MAT);
    p2=cgetg(3,t_COL); y[1]=(long)p2; p2[1]=lcopy(p1); p2[2]=lcopy(q1);
    p2=cgetg(3,t_COL); y[2]=(long)p2; p2[1]=lcopy(p0); p2[2]=lcopy(q0);
    return gerepile(av,tetpil,y);
  }

  ly=lg(x[1]); if (ly==1 || ly>3) err(pnqner2);
  if (ly==2)
  {
    p1=cgetg(lx,t_VEC); for (i=1; i<lx; i++) p1[i]=(long)(((GEN)x[i])[1]);
    tetpil=avma; return gerepile(av,tetpil,pnqn(p1));
  }
  p0=gun; q0=gzero; p1=gcoeff(x,2,1); q1=gcoeff(x,1,1);

  for (i=2; i<lx; i++)
  {
    a=gcoeff(x,2,i); b=gcoeff(x,1,i);
    p2=gadd(gmul(a,p1),gmul(b,p0)); p0=p1; p1=p2;
    q2=gadd(gmul(a,q1),gmul(b,q0)); q0=q1; q1=q2;
  }
  tetpil=avma; y=cgetg(3,t_MAT);
  p2=cgetg(3,t_COL); y[1]=(long)p2; p2[1]=lcopy(p1); p2[2]=lcopy(q1);
  p2=cgetg(3,t_COL); y[2]=(long)p2; p2[1]=lcopy(p0); p2[2]=lcopy(q0);
  return gerepile(av,tetpil,y);
}

GEN
bestappr(GEN x, GEN k)
{
  long av=avma,tetpil,tx,tk=typ(k),lx,i,e;
  GEN p_0,p_1,p,q_0,q_1,q,a,y;

  if (tk==t_INTMOD || tk >= t_COMPLEX) err(bester1);
  if (tk != t_INT) k=gcvtoi(k,&e);
  if (cmpii(k,gun)<0) k=gun;
  tx=typ(x); if (tx == t_FRACN) x=gdiv((GEN)x[1],(GEN)x[2]);
  switch(tx)
  {
    case t_INT:
      tetpil=avma;
      return (av!=avma) ? gerepile(av,tetpil,gcopy(x)) : gcopy(x);

    case t_FRAC:
      if (gcmp((GEN)x[2],k)<=0)
      {
	tetpil=avma; return (av!=avma)?gerepile(av,tetpil,gcopy(x)):gcopy(x);
      }

    case t_REAL:
      p_1=gun; p_0=gfloor(x); q_1=gzero; q_0=gun; a=p_0;
      while (gcmp(q_0,k)<=0)
      {
	x=gsub(x,a);
	if (gcmp0(x))
	{
	  tetpil=avma; return gerepile(av,tetpil,gdiv(p_0,q_0));
	 }
	x=gdiv(gun,x); if (gcmp(x,k)<0) a=gfloor(x); else a=k;
	p=gadd(gmul(a,p_0),p_1); q=gadd(gmul(a,q_0),q_1);
	p_1=p_0; q_1=q_0; p_0=p; q_0=q;
      }
      tetpil=avma; return gerepile(av,tetpil,gdiv(p_1,q_1));

    case t_POL:
      lx=lgef(x); y=cgetg(lx,tx);
      for (i=1; i<lontyp[tx]; i++) y[i]=x[i];
      for (i=lontyp[tx]; i<lx; i++) y[i]=(long)bestappr((GEN)x[i],k);
      return y;

    case t_COMPLEX: case t_SER: case t_RFRAC: case t_RFRACN:
    case t_VEC: case t_COL: case t_MAT:
      lx=lg(x); y=cgetg(lx,tx);
      for (i=1; i<lontyp[tx]; i++) y[i]=x[i];
      for (i=lontyp[tx]; i<lx; i++) y[i]=(long)bestappr((GEN)x[i],k);
      return y;
  }
  err(bester2);
  return NULL; /* not reached */
}

/***********************************************************************/
/**                                                                   **/
/**            FUNDAMENTAL UNIT AND REULATOR (QUAD. FIELDS)           **/
/**                                                                   **/
/***********************************************************************/

GEN
gfundunit(GEN x)
{
  return garith_proto(fundunit,x);
}

GEN
fundunit(GEN x)
{
  GEN  p1,q1,y,u,v,a,u1,v1,sqd,f,m,c;
  long av,tetpil,r,p4,lx,flp,flq,av2;

  if (typ(x) != t_INT) err(arither1);
  if (signe(x)<=0) err(funder1);
  r=mod4(x); if (r==2 || r==3) err(funder2);
  p4=lclone(quadpoly(x));
  av=avma; sqd=racine(x); lx=lgefint(sqd)+1;
  affsi(2,v=cgeti(lx)); affsi(r,u=cgeti(lx));
  affii(shifti(addsi(r,sqd),-1),a=cgeti(lx));
  m=cgetg(3,t_MAT);
  c=cgetg(3,t_COL); c[1]=(long)a; c[2]=un; m[1]=(long)c;
  c=cgetg(3,t_COL); c[1]=un; c[2]=zero; m[2]=(long)c;

  av2=avma;
  f=cgetg(3,t_MAT);
  c=cgetg(3,t_COL); c[1]=lcopy(a); c[2]=un; f[1]=(long)c;
  c=cgetg(3,t_COL); c[1]=un; c[2]=zero; f[2]=(long)c;
  for(;;)
  {
    u1=subii(mulii(a,v),u); flp=cmpii(u,u1); affii(u1,u);
    v1=divii(subii(x,sqri(u)),v); flq=cmpii(v,v1); affii(v1,v);
    diviiz(addii(sqd,u),v,a);
    tetpil=avma;
    if (flp && flq) f=gerepile(av2,tetpil,gmul(f,m));
    else break;
  }
  c=(GEN)(f[2]); p1=(GEN)(c[1]); q1=(GEN)(c[2]);
  y=cgetg(4,t_QUAD); y[1]=p4;
  y[2]= (r) ? lsubii(p1,q1) : lcopy(p1);
  y[3]=lcopy(q1);

  if (!flq)
  {
    f=gmul(f,m);
    c=(GEN)(f[2]); p1=(GEN)(c[1]); q1=(GEN)(c[2]);
    v1=cgetg(4,t_QUAD); v1[1]=p4;
    if (r) { v1[2]=lsubii(p1,q1); v1[3]=lcopy(q1); }
    else { v1[2]=lcopy(p1); v1[3]=lcopy(q1); }
    u1=gconj(y); tetpil=avma; y=gdiv(v1,u1);
  }
  else
  {
    u1=gconj(y); tetpil=avma; y=gdiv(y,u1);
  }
  if (signe(y[3]) < 0) { tetpil=avma; y=gneg(y); }
  return gerepile(av,tetpil,y);
}

GEN
gregula(GEN x, long prec)
{
  return garith_proto2gs(regula,x,prec);
}

GEN
regula(GEN x, long prec)
{
  GEN  reg,reg1,rsqd;
  GEN  rexp,y,u,v,a,u1,v1,sqd;
  long av,tetpil,r,lx,flp,flq,av2;

  if (typ(x) != t_INT) err(arither1);
  if (signe(x)<=0) err(funder1);
  r=mod4(x); if (r==2 || r==3) err(funder2);
  av=avma; sqd=racine(x); if (gegal(sqri(sqd),x)) err(reguler1);
  rsqd=gsqrt(x,prec); affsi(0,rexp=cgeti(4));
  lx=lgefint(sqd)+1; affsr(2,reg=cgetr(prec));
  affsi(2,v=cgeti(lx)); affsi(r,u=cgeti(lx));
  affii(shifti(addsi(r,sqd),-1),a=cgeti(lx));
  av2=avma;
  for(;;)
  {
    u1=subii(mulii(a,v),u); flp=cmpii(u,u1); affii(u1,u);
    reg1=divri(addir(u,rsqd),v);
    v1=divii(subii(x,sqri(u)),v); flq=cmpii(v,v1);
    if (flp && flq)
    {
      affii(v1,v); diviiz(addii(sqd,u),v,a);
      tetpil=avma; reg=gerepile(av2,tetpil,mulrr(reg,reg1));
      r=expo(reg); addsiz(r,rexp,rexp); setexpo(reg,0);
    }
    else { tetpil=avma; break; }
  }
  reg=shiftr(mulrr(reg,reg),-1);
  if (!flq)
  {
    u1 = subii(mulii(a,v),u);
    reg = mulrr(reg, divri(addir(u,rsqd),v));
  }
  y = glog(divri(reg,v),prec);
  u1 = mpshift(mulir(rexp,glog(gdeux,prec)),1);
  tetpil=avma; return gerepile(av,tetpil,gadd(y,u1));
}


/*************************************************************************/
/**                                                                     **/
/**                            CLASS NUMBER                             **/
/**                                                                     **/
/*************************************************************************/
static GEN
gclassno(GEN x)
{
  return garith_proto(classno,x);
}

static GEN
gclassno2(GEN x)
{
  return garith_proto(classno2,x);
}

/* calcul de h(x) pour x<0 par la methode de Shanks */
GEN
classno(GEN x)
{
  static long court[3]= { evaltyp(t_INT) | evalpere(1) | evallg(3),
                          evalsigne(1) | evallgefint(3),0 };
  static long prem, ptforme;
  long av,av1,av2,av3,tetpil,k,k2,i,j,j1,lim,limite,com,j2,s;
  GEN tabla, tablb, count, index, hash;
  GEN p1,p2,p3,hin,q,formes[100],l,h,hp,f,fh,fg,ftest,pm2;
  static byteptr p;

  if (typ(x) != t_INT) err(arither1);
  s=signe(x);
  if (!s) err(arither2);
  if (s>0) return classno2(x);

  k=mod4(x);
  if (k==1 || k==2) err(classer2);
  if (gcmpgs(x,-12) >= 0) return gun;

  tabla = newbloc(10000); tablb = newbloc(10000);
  count = newbloc(256); index = newbloc(257);
  hash = newbloc(10000);
  prem=ptforme=0; p=diffptr; av=avma; limite=(av+bot)>>1;
  p1=divrr(gsqrt(absi(x),DEFAULTPREC),mppi(DEFAULTPREC));
  l = gtrunc(shiftr(gsqrt(gsqrt(absi(x),DEFAULTPREC),DEFAULTPREC),1));
  if (gcmpgs(l,1000)<0) affsi(1000,l);
  while (gcmpgs(l,prem)>=0 && *p)
  {
    prem += *p++;
    k=krogs(x,prem);
    if (k)
    {
      av2=avma;
      if (k>0)
      {
	divrsz(mulsr(prem,p1),prem-1,p1); avma=av2;
	if (ptforme<100 && prem>2)
	{
	  court[2]=prem;
	  formes[ptforme++]=redimag(primeform(x,court,0));
	 }
      }
      else { divrsz(mulsr(prem,p1),prem+1,p1); avma=av2; }
    }
  }
  hin=ground(p1); h=gcopy(hin); f=formes[0]; fh=gpui(f,h,0);
  s = 2*itos(gceil(gpui(p1,gdivgs(gun,4),DEFAULTPREC)));
  if (s>=10000) s=10000;
  p1=fh; av2=avma;
  for (i=0; i<=255; i++) count[i]=0;
  for (i=0; i<=s-1; i++)
  {
    p2=(GEN)p1[1]; tabla[i]=p2[lgefint(p2)-1]; j=tabla[i]&255; count[j]++;
    p2=(GEN)p1[2]; tablb[i]=p2[lgefint(p2)-1];
    p1=compimag(p1,f);
  }
  fg=gpuigs(f,s); ftest=gpuigs(p1,0);
  index[0]=0; for (i=0; i<=254; i++) index[i+1]=index[i]+count[i];
  for (i=0; i<=s-1; i++) hash[index[tabla[i]&255]++]=i;
  index[0]=0; for (i=0; i<=255; i++) index[i+1]=index[i]+count[i];
  com=0; av3=avma;
  for(;;)
  {
    p1=(GEN)ftest[1]; k=p1[lgefint(p1)-1]; j=k&255;
    pm2=negi((GEN)ftest[2]);
    for (j1=index[j]; j1 < index[j+1]; j1++)
    {
      j2=hash[j1];
      if (tabla[j2]==k)
      {
	p2=(GEN)ftest[2]; k2=p2[lgefint(p2)-1];
	if (tablb[j2]==k2 || tablb[j2]== -k2)
	{
	  p1=gmul(gpuigs(f,j2),fh);
	  if (!cmpii((GEN) p1[1],(GEN) ftest[1]) &&
	      ( !cmpii((GEN) p1[2],(GEN) ftest[2]) ||
	        !cmpii((GEN) p1[2],pm2) )
	     )
	          goto go_on;
	}
      }
    }
    com++;
    if (avma>=limite) ftest=gmul(ftest,fg);
    else { tetpil=avma; ftest=gerepile(av3,tetpil,gmul(ftest,fg)); }
    if (gcmp1((GEN)ftest[1]))
      err(impl, "classno with too small order");
  }

go_on:

  h=addis(h,j2); p2=mulsi(s,stoi(com)); tetpil=avma;
  h = (!cmpii((GEN)p1[2],(GEN)ftest[2])) ? subii(h,p2)  :addii(h,p2);
  p2=factor(h); p1=(GEN)p2[1]; p2=(GEN)p2[2];
  for (i=1; i<lg(p1); i++)
  {
    p3=divii(h,(GEN)p1[i]); fh=gpui(f,p3,0); lim=itos((GEN)p2[i]);
    for (j=1; j<=lim && !cmpii((GEN)fh[1],gun); j++)
    {
      h=p3;
      if (j<lim) { p3=divii(h,(GEN)p1[i]); fh=gpui(f,p3,0); }
    }
  }
  q=ground(gdiv(hin,h)); tetpil=avma;
  hp=mulii(q,h); av1=avma;
  for (i=1; i<=10 && i<ptforme; i++)
  {
    f=formes[i]; com=0;
    fg=gpui(f,h,0); fh=gpui(fg,q,0); ftest=fg;
    if (cmpis((GEN)fh[1],1))
    {
      for(;;)
      {
	com++;
	if (!cmpii((GEN) fh[1], (GEN) ftest[1]) &&
	    ( !cmpii((GEN) fh[2], (GEN) ftest[2]) ||
	      !cmpii((GEN) fh[2], negi((GEN) ftest[2]))))
		break;
	ftest=gmul(ftest,fg);
      }
      if (!cmpii((GEN)fh[2],(GEN)ftest[2])) com= -com;
      p2=mulsi(com,h); q=addsi(com,q); tetpil=avma;
      hp=addii(hp,p2); av1=avma;
    }
  }
  avma=av1;
  killbloc(tabla); killbloc(tablb); killbloc(count);
  killbloc(index); killbloc(hash);
  return gerepile(av,tetpil,hp);
}

GEN
qfbclassno0(GEN x,long flag)
{
  switch(flag)
  {
    case 0: return gclassno(x);
    case 1: return gclassno2(x);
    default: err(flagerr);
  }
  return NULL; /* not reached */
}

GEN
classno2(GEN x)
{
  long av=avma,tetpil,n,i,k,s=signe(x),fl2,ex;
  GEN p1,p2,p3,p4,p5,p6,p7,p9,pi4,d,reg,logd,fd;

  if (typ(x) != t_INT) err(arither1);
  if (!s) err(arither2);

  p1=auxdecomp(absi(x),1); p2=(GEN)p1[2]; p1=(GEN)p1[1];
  n=lg(p1); p3=gun; fl2=0;
  for (i=1; i<n; i++)
  {
    ex=itos((GEN)p2[i]);
    if (ex>=2)
    {
      p4=(GEN)p1[i];
      if (ex & 1) p3=mulii(p3,p4);
      if (gegal(p4,gdeux)) fl2=1;
    }
    else p3=mulii(p3,(GEN)p1[i]);
  }

  if (mod4(p3) != (2-s))
  {
    if (!fl2) err(classer2);
    p3=shifti(p3,2);
  }
  else fl2=0;
  p9=gun; fd = (s<0) ? gneg(p3) : p3;
  for (i=1; i<n; i++)
  {
    ex=itos((GEN)p2[i]); p4=(GEN)p1[i];
    if (gegal(p4,gdeux)&&fl2) ex-=2;
    if (ex>=2)
    {
      p9=mulii(p9,subis(p4,kronecker(fd,p4)));
      if (ex>=4) p9=mulii(p9,gpuigs(p4,(ex>>1)-1));
     }
  }

/****************************/
  if (s<0)
    if (!gcmpgs(p3,4))
    {
      if (gcmpgs(x,-4)) p9=gdivgs(p9,2);
    }
    else if (!gcmpgs(p3,3))
      if (gcmpgs(x,-3)) p9=gdivgs(p9,3);
/****************************/

  pi4=mppi(DEFAULTPREC);
  if (s>0)
  {
    reg=regula(fd,DEFAULTPREC); logd=glog(fd,DEFAULTPREC);
    p1=gsqrt(gdiv(gmul(fd,logd),gmul2n(pi4,1)),DEFAULTPREC);
    p2=gsubsg(1,gmul2n(gdiv(glog(reg,DEFAULTPREC),logd),1));
    p3=gsqrt(gdivsg(2,logd),DEFAULTPREC);
    if (gcmp(p2,p3)>=0) p1=gmul(p2,p1);
    p1=gtrunc(p1);
    if (lgefint(p1)!=3) err(classer1);
    n=p1[2]; if (n<0) err(classer1);
    p1=gsqrt(fd,DEFAULTPREC); p4=divri(pi4,fd); p3=gzero;
    p7=divsr(1,mpsqrt(pi4));
    for (i=1; i<=n; i++)
    {
      k=krogs(fd,i);
      if (k)
      {
	p6=mulir(mulss(i,i),p4);
	p5=subsr(1,mulrr(p7,incgam3(ghalf,p6,DEFAULTPREC)));
	p5=addrr(divrs(mulrr(p1,p5),i),eint1(p6,DEFAULTPREC));
	if (k>0) p3=addrr(p3,p5); else p3=subrr(p3,p5);
      }
    }
    p3=shiftr(divrr(p3,reg),-1);

/**********************/
    if (gcmp(x,fd))
      p9=gdiv(p9,ground(gdiv(regula(x,DEFAULTPREC),reg)));
/**********************/
  }
  else
  {
    d=p3;
    if (gcmpgs(fd,-11)>=0)
    {
      tetpil=avma; return gerepile(av,tetpil,gcopy(p9));
    }

    p1=gtrunc(gsqrt(gdiv(gmul(d,glog(d,DEFAULTPREC)),gmul2n(pi4,1)),
                    DEFAULTPREC));
    if (lgefint(p1)!=3) err(classer1);
    n=p1[2]; if (n<0) err(classer1);
    p1=gdiv(gsqrt(d,DEFAULTPREC),pi4); p4=divri(pi4,d); p3=gzero;
    p7=divsr(1,mpsqrt(pi4));
    for (i=1; i<=n; i++)
    {
      k=krogs(fd,i);
      if (k)
      {
	p6=mulir(mulss(i,i),p4);
	p5=subsr(1,mulrr(p7,incgam3(ghalf,p6,DEFAULTPREC)));
	p5=addrr(p5,divrr(divrs(p1,i),mpexp(mulir(mulss(i,i),p4))));
	if (k>0) p3=addrr(p3,p5); else p3=subrr(p3,p5);
      }
     }
  }
  p3=ground(p3); tetpil=avma; return gerepile(av,tetpil,gmul(p9,p3));
}

GEN
classno3(GEN x)
{
  long d,a,b,h,b2,f,av,tetpil;
  GEN y;

  d= -itos(x); if (d>0 || (d & 3) > 1) return gzero;
  if (!d) return gdivgs(gun,-12);
  h=0; b=d&1; b2=(b-d)>>2; f=0;
  if (!b)
  {
    for (a=1; a*a<b2; a++) { if (!(b2%a)) h++; }
    f = (a*a==b2); b=2; b2=(4-d)>>2;
  }
  while (b2*3+d<0)
  {
    if (!(b2%b)) h++;
    for (a=b+1; a*a<b2; a++) { if (!(b2%a)) h+=2; }
    if (a*a==b2) h++;
    b+=2; b2=(b*b-d)>>2;
  }
  if (b2*3+d==0)
  {
    av=avma; y=gdivgs(gun,3); tetpil=avma;
    return gerepile(av,tetpil,gaddsg(h,y));
  }
  if (f) return gaddsg(h,ghalf); else return stoi(h);
}
