/********************************************************************/
/********************************************************************/
/**                                                                **/
/**                  BIBLIOTHEQUE  MATHEMATIQUE                    **/
/**                      (premiere partie)                         **/
/**                                                                **/
/**                     Copyright Babe Cool                        **/
/**                                                                **/
/********************************************************************/
/********************************************************************/
/* $Id: bibli1.c,v 1.914.7.1 1997/01/19 16:18:09 karim Exp karim $ */
#include "genpari.h"
#include "nf.h"
static GEN lllgramintwithcontent(GEN x, GEN veccon, long all);
static GEN lllintwithcontent(GEN x);
static GEN lllgramall(GEN x, long all);
static GEN lllall0(GEN x, long all);

/********************************************************************/
/**                                                                **/
/**                 DEVELOPPEMENTS  LIMITES                        **/
/**                                                                **/
/********************************************************************/

GEN
tayl(GEN x, long v, long precdl)
{
  long tetpil,i,vx = gvar9(x), av=avma;
  GEN p1, y;

  if (v <= vx)
  {
    p1 = cgetg(3,t_SER);
    p1[2] = zero; p1[1] = HIGHVALPBIT + precdl; setvarn(p1,v);
    tetpil = avma; return gerepile(av, tetpil, gadd(p1,x));
  }
  p1=cgetg(v+2,t_VEC);
  for (i=0; i<vx; i++) p1[i+1]=lpolx[i]; p1[vx+1]=lpolx[v];
  for (i=vx+1; i<v; i++) p1[i+1]=lpolx[i]; p1[v+1]=lpolx[vx];
  y = tayl(changevar(x, p1), vx, precdl); tetpil=avma;
  return gerepile(av, tetpil, changevar(y,p1));
}

GEN
grando0(GEN x, long n, long do_clone)
{
  long m, v, tx=typ(x);
  GEN y;

  if (gcmp0(x)) err(grandoer1);
  if (tx == t_INT)
  {
    if (!gcmp1(x)) /* bug: 3 + O(1). We suppose x is a truc() */
    {
      y=cgetg(5,t_PADIC); y[3]=un; y[4]=zero;
      y[2] = do_clone? lclone(x): lcopy(x);
      setvalp(y,n); setprecp(y,0);
      return y;
    }
    v=0; m=0; /* 1 = x^0 */
  }
  else
  {
    if (tx != t_POL && ! is_rfrac_t(tx)) err(grandoer1);
    v=gvar(x); m=n*gval(x,v);
  }
  y=cgetg(3,t_SER); y[1]=HIGHVALPBIT + m; y[2]=zero;
  setvarn(y,v); return y;
}

/*******************************************************************/
/**                                                               **/
/**                      SPECIAL POLYNOMIALS                      **/
/**                                                               **/
/*******************************************************************/

/* Tchebichev polynomial */
/* T0=1; T1=X; T(n)=2*X*T(n-1)-T(n-2) */
GEN
tchebi(long n, long v)
{
  long av,tetpil,limite,m;
  GEN p0,p1,px,q;

  if (n==0) return polun[v];
  if (n==1) return polx[v];

  p0=polun[v]; p1=px=polx[v];
  av=avma; limite=(avma+3*bot)>>2;
  for (m=1; m<n-1; m++)
  {
    GEN p2;

    q=gmul(px,gmul2n(p1,1));
    tetpil=avma; p2=gsub(q,p0); p0=p1; p1=p2;
    if (avma<limite)
    {
      GEN *gptr[2];

      p0=gcopy(p0); gptr[0]=&p0; gptr[1]=&p1;
      gerepilemanysp(av,tetpil,gptr,2);
    }
  }
  q=gmul(px,gmul2n(p1,1));
  tetpil=avma; return gerepile(av,tetpil,gsub(q,p0));
}

/* Legendre polynomial */
/* L0=1; L1=X; (n+1)*L(n+1)=(2*n+1)*X*L(n)-n*L(n-1) */
GEN
legendre(long n, long v)
{
  long av,tetpil,m,limite;
  GEN p0,p1,px,q;

  if (n==0) return polun[v];
  if (n==1) return polx[v];

  p0=polun[v]; px=polx[v];
  av=avma; limite=(avma+3*bot)>>2;
  p1=gmul2n(px,1);
  for (m=1; m<n; m++)
  {
    GEN p2 = gsub(gmul(px,gmulsg(4*m+2,p1)),gmulsg(4*m,p0));

    p0 = p1; tetpil=avma; p1 = gdivgs(p2,m+1);
    if (avma<limite)
    {
      GEN *gptr[2];

      p0=gcopy(p0); gptr[0]=&p0; gptr[1]=&p1;
      gerepilemanysp(av,tetpil,gptr,2);
    }
  }
  q=shifti(gun,n);
  tetpil=avma; return gerepile(av,tetpil,gdiv(p1,q));
}

/* cyclotomic polynomial */
GEN
cyclo(long n, long v)
{
  long av=avma,tetpil,d,q,m;
  GEN p1,yn,yd;

  if (n<=0) err(arither2);
  d=1; yn=gun; yd=gun;
  while (d*d<=n)
  {
    if (!(n%d))
    {
      q=n/d; m=mu(stoi(q));
      if (m)
      {
	p1=gsub(gpuigs(polx[v],d),gun);
	if (m>0) yn=gmul(yn,p1); else yd=gmul(yd,p1);
      }
      if (q!=d)
      {
	m=mu(stoi(d));
	if (m)
	{
	  p1=gsub(gpuigs(polx[v],q),gun);
	  if (m>0) yn=gmul(yn,p1); else yd=gmul(yd,p1);
	 }
      }
    }
    d++;
  }
  tetpil=avma; return gerepile(av,tetpil,gdiv(yn,yd));
}

/********************************************************************/
/**                                                                **/
/**                  HILBERT & PASCAL MATRICES                     **/
/**                                                                **/
/********************************************************************/

GEN
hilb(long n) /* Hilbert matrix of order n */
{
  long i,j;
  static long sho[]={ evaltyp(t_INT)+evalpere(1)+evallg(3),
                      evalsigne(1)+evallgefint(3), 0 };
  GEN a, p = cgetg(n+1,t_MAT);

  for (j=1; j<=n; j++)
  {
    p[j]=lgetg(n+1,t_COL);
    for (i=1; i<=n; i++)
    {
      a=cgetg(3,t_FRAC); coeff(p,i,j)=(long)a;
      a[1]=un; sho[2]=i+j-1; a[2]=lcopy(sho);
    }
  }
  return p;
}

GEN
pasc(long n) /* matrice triangle de PASCAL d'ordre n */
{
  long i,j;
  GEN p = cgetg(n+2,t_MAT);

  for (j=1; j<=n+1; j++)  p[j]=lgetg(n+2,t_COL);
  for (i=1; i<=n+1; i++)
  {
    coeff(p,i,1)=un; coeff(p,i,i)=un;
    for (j=2; j<i; j++)
      coeff(p,i,j)=laddii(gcoeff(p,i-1,j),gcoeff(p,i-1,j-1));
    for (j=i+1; j<=n+1; j++)
      coeff(p,i,j)=zero;
  }
  return p;
}

/********************************************************************/
/**                                                                **/
/**                  LAPLACE TRANSFORM (OF A SERIES)               **/
/**                                                                **/
/********************************************************************/

GEN
laplace(GEN x)
{
  long i,l,ec,av,tetpil;
  GEN y,p1;

  if (typ(x)!=t_SER) err(laper1);
  if (gcmp0(x)) return gcopy(x);

  av=avma; ec=valp(x); if (ec<0) err(laper2);
  l=lg(x); y=cgetg(l,t_SER);
  p1=mpfact(ec); y[1]=x[1];
  for (i=2; i<l; i++)
  {
    y[i]=lmul(p1,(GEN)x[i]);
    ec++; p1=gmulsg(ec,p1);
  }
  tetpil=avma; return gerepile(av,tetpil,gcopy(y));
}

/********************************************************************/
/**                                                                **/
/**              CONVOLUTION PRODUCT (OF TWO SERIES)               **/
/**                                                                **/
/********************************************************************/

GEN
convol(GEN x, GEN y)
{
  long l,l1,i,j,v,f, lx = lg(x), ly = lg(y), ex = valp(x), ey = valp(y);
  GEN z;

  if (typ(x) != t_SER || typ(y) != t_SER) err(convol1);
  if (gcmp0(x) || gcmp0(y)) err(convol2);
  v=ex; if (ey>v) v=ey;
  l=ex+lx; l1=ey+ly; if (l1<l) l=l1;
  l=l-v; if (l<3) err(convol3); f=1;
  for (i=v+2; (i<(l+v)) && f; i++)
    f=(gcmp0((GEN)x[i-ex]) || gcmp0((GEN)y[i-ey]));
  if (i==(l+v))
  {
    z=cgetg(3,t_SER); z[1]=HIGHVALPBIT-2+l+v; z[2]=zero;
  }
  else
  {
    z=cgetg(l-i+3+v,t_SER); z[1]=evalsigne(1)+HIGHVALPBIT-3+i;
    for (j=i-1; j<l+v; j++) z[j-i+3]=lmul((GEN)x[j-ex],(GEN)y[j-ey]);
  }
  return z;
}

/*******************************************************************/
/**                                                               **/
/**         CONVERT                                               **/
/**            series --> polynomial or rat. frac                 **/
/**            p-adic --> int. or rat.                            **/
/**                                                               **/
/*******************************************************************/

GEN
gconvsp(GEN x)
{
  long e,av,tetpil,i,v;
  GEN p1,y;

  if (typ(x)!=t_SER) err(csper1);
  v=varn(x); e=valp(x); av=avma; y=gcopy(x); settyp(y,t_POL);
  if (gcmp0(x)) y[1]=evallgef(2) | evalvarn(v);
  else
  {
    i=lg(x)-1; while (i>1 && gcmp0((GEN)y[i])) i--;
    y[1] = evalsigne(1) | evallgef(1+i) | evalvarn(v);
    p1=gpuigs(polx[v],e); tetpil=avma;
    y=gerepile(av,tetpil,gmul(p1, y));
  }
  return y;
}

GEN
gconvpe(GEN x)
{
  long av,tetpil;
  GEN p1;

  if (typ(x)!=t_PADIC) err(csper1);
  if (!signe(x[4])) return gzero;
  av=avma; p1=gpuigs((GEN)x[2],valp(x));
  tetpil=avma; return gerepile(av,tetpil,gmul(p1,(GEN)x[4]));
}

/******************************************************************/
/**                                                              **/
/**                       PRECISION CHANGES                      **/
/**                                                              **/
/******************************************************************/

GEN
gprec(GEN x, long l)
{
  long tx=typ(x),lx=lg(x),i,pr;
  GEN y;

  if (l<=0) err(precer1);
  switch(tx)
  {
    case t_REAL:
      pr = (long) (l*K1+3); y=cgetr(pr); affrr(x,y); break;

    case t_PADIC:
      y=cgetg(lx,tx); y[2] = lcopy((GEN) x[2]);
      if (!signe(x[4]))
      {
	l+=precp(x); y[1]=evalvalp(l)+evalprecp(0);
	y[3]=un; y[4]=zero; return y;
      }
      y[1]=x[1]; setprecp(y,l);
      y[3]=lpuigs((GEN)x[2],l);
      y[4]=lmodii((GEN)x[4],(GEN)y[3]);
      break;

    case t_SER:
      if (gcmp0(x))
      {
	y=cgetg(3,t_SER); y[1]=HIGHVALPBIT+l; setvarn(y,varn(x));
	return y;
      }
      y=cgetg(l+2,t_SER); y[1]=x[1]; l++; i=l;
      if (l>=lx)
      {
	for ( ; i>=lx; i--) y[i]=zero;
      }
      for ( ; i>=2; i--) y[i]=lcopy((GEN)x[i]);
      break;

    case t_POL:
      lx=lgef(x); y=cgetg(lx,tx); y[1]=x[1];
      for (i=2; i<lx; i++) y[i]=lprec((GEN)x[i],l);
      break;

    case t_COMPLEX: case t_POLMOD: case t_RFRAC: case t_RFRACN:
    case t_VEC: case t_COL: case t_MAT:
      y=cgetg(lx,tx);
      for (i=1; i<lx; i++) y[i]=lprec((GEN)x[i],l);
      break;
    default: y=gcopy(x);
  }
  return y;
}

/*******************************************************************/
/**                                                               **/
/**                     RECIPROCAL POLYNOMIAL                     **/
/**                                                               **/
/*******************************************************************/

GEN
polrecip(GEN x)
{
  GEN y;
  long lx=lgef(x),i;

  if (typ(x) != t_POL) err(polrecer1);
  y=cgetg(lx,t_POL); y[1]=x[1];
  for (i=2; i<lx; i++) y[i]=lcopy((GEN)x[lx+1-i]);
  normalizepol(&y); return y;
}

/*******************************************************************/
/**                                                               **/
/**                      BINOMIAL COEFFICIENTS                    **/
/**                                                               **/
/*******************************************************************/

GEN
binome(GEN x, long k)
{
  GEN y,p1;
  long av,tetpil,i;

  if (k<0) return gzero;
  if (!k) return gun;
  if (k==1) return gcopy(x);
  av=avma; y=x;
  for (i=k; i>=2; i--)
  {
    p1=gmul(y,gaddgs(x,i-1-k));
    tetpil=avma;
    y=gdivgs(p1,i);
  }
  return gerepile(av,tetpil,y);
}

/********************************************************************/
/**                                                                **/
/**                   INTEGRAL KERNEL (LLL REDUCED)                **/
/**                                                                **/
/********************************************************************/

/* produit scalaire de x par y sans verification de compatibilite */
static GEN
gscal(GEN x,GEN y)
{
  GEN z,p;
  long i,av,tetpil,lx=lg(x);

  z=gzero; if (lx==1) return z;
  av=avma;
  for (i=1; i<lx; i++)
  {
    p=gmul((GEN)x[i],(GEN)y[i]);
    tetpil=avma; z=gadd(z,p);
  }
  return gerepile(av,tetpil,z);
}

GEN
kerint0(GEN x, long flag)
{
  switch(flag)
  {
    case 0: return kerint(x);
    case 1: return kerint1(x);
    case 2: return kerint2(x);
    default: err(flagerr);
  }
  return NULL; /* not reached */
}

GEN
kerint1(GEN x)
{
  long av=avma,tetpil;
  GEN p1,p2;

  p1=matrixqz3(ker(x)); p2=lllint(p1); tetpil=avma;
  return gerepile(av,tetpil,gmul(p1,p2));
}

GEN
kerint2(GEN x)
{
  long lx=lg(x), i,j,av,av1;
  GEN g,p1;

  if (typ(x) != t_MAT) err(lller1);
  av=avma; g=cgetg(lx,t_MAT);
  for (j=1; j<lx; j++) g[j]=lgetg(lx,t_COL);
  for (i=1; i<lx; i++)
    for (j=1; j<=i; j++)
      coeff(g,i,j) = coeff(g,j,i) = (long)gscal((GEN)x[i],(GEN)x[j]);
  g=lllgramall(g,1); p1=lllint(g);
  av1=avma; return gerepile(av,av1,gmul(g,p1));
}

GEN
kerint(GEN x)
{
  long av=avma,av1;
  GEN g,p1;

  g=lllall0(x,1); if (lg(g)==1) return g;
  p1=lllint(g); av1=avma;
  return gerepile(av,av1,gmul(g,p1));
}

/********************************************************************/
/**                                                                **/
/**                          LLL ALGORITHM                         **/
/**                                                                **/
/********************************************************************/

static GEN lllgramallgen(GEN x, long all);

GEN
lll0(GEN x, long flag, long prec)
{
  switch(flag)
  {
    case 0: return lll(x,prec);
    case 1: return lllint(x);
    case 2: return lllintpartial(x);
    case 3: return lllrat(x);
    case 4: return lllkerim(x);
    case 5: return lllkerimgen(x);
    case 7: return lll1(x,prec);
    case 8: return lllgen(x);
    case 9: return lllintwithcontent(x);
    default: err(flagerr);
  }
  return NULL; /* not reached */
}

GEN
lllgram0(GEN x, long flag, long prec)
{
  switch(flag)
  {
    case 0: return lllgram(x,prec);
    case 1: return lllgramint(x);
    case 4: return lllgramkerim(x);
    case 5: return lllgramkerimgen(x);
    case 7: return lllgram1(x,prec);
    case 8: return lllgramgen(x);
    default: err(flagerr);
  }
  return NULL; /* not reached */
}

/*   x est la matrice d'une base b_i; retourne la matrice u (entiere
 * unimodulaire) d'une base LLL-reduite c_i en fonction des b_i (la base
 * reduite est c=b*u).
 */
static GEN
lll_proto(GEN x, GEN f(GEN, long), long prec)
{
  long lx=lg(x),i,j,av,av1;
  GEN g;

  if (typ(x) != t_MAT) err(lller1);
  if (lx==1) return cgetg(1,t_MAT);
  av=avma; g=cgetg(lx,t_MAT);
  for (j=1; j<lx; j++) g[j]=lgetg(lx,t_COL);
  for (i=1; i<lx; i++)
    for (j=1; j<=i; j++)
      coeff(g,i,j)=coeff(g,j,i)=(long)gscal((GEN)x[i],(GEN)x[j]);
  av1=avma;
  return gerepile(av,av1,f(g,prec));
}

static GEN
lllintwithcontent(GEN x)
{
  long lx=lg(x),i,j,av,tetpil;
  GEN veccon,con,xred,g;

  if (typ(x) != t_MAT) err(lller1);
  if (lx==1) return cgetg(1,t_MAT);
  av=avma; veccon=cgetg(lx,t_VEC); g=cgetg(lx,t_MAT); xred=cgetg(lx,t_MAT);
  for (j=1; j<lx; j++)
  {
    g[j]=lgetg(lx,t_COL); con=content((GEN)x[j]);
    xred[j]=(long)gdiv((GEN)x[j],con); veccon[j]=(long)con;
  }
  for (i=1; i<lx; i++)
    for (j=1; j<=i; j++)
      coeff(g,i,j)=coeff(g,j,i)=(long)gscal((GEN)xred[i],(GEN)xred[j]);
  tetpil=avma;
  return gerepile(av,tetpil,lllgramintwithcontent(g,veccon,2));
}

GEN
lll(GEN x, long prec) { return lll_proto(x,lllgram,prec); }

GEN
lll1(GEN x, long prec) { return lll_proto(x,lllgram1,prec); }

GEN
lllrat(GEN x) { return lll_proto(x,lllgram,0); }

GEN
lllint(GEN x) { return lll_proto(x,lllgramall,2); }

GEN
lllgen(GEN x) { return lll_proto(x,lllgramallgen,2); }

GEN
lllgramgen(GEN x) { return lllgramallgen(x,2); }

GEN
lllgramkerimgen(GEN x) { return lllgramallgen(x,0); }

static GEN
lllkerim_proto(GEN x, GEN f(GEN,long))
{
  long lx=lg(x), i,j,av,av1;
  GEN g;

  if (typ(x) != t_MAT) err(lller1);
  if (lx==1)
  {
    g=cgetg(3,t_VEC); g[1]=lgetg(1,t_MAT); g[2]=lgetg(1,t_MAT); return g;
  }
  if (lg((GEN)x[1])==1)
  {
    g=cgetg(3,t_VEC); g[1]=(long)idmat(lx-1); g[2]=lgetg(1,t_MAT); return g;
  }
  av=avma; g=cgetg(lx,t_MAT);
  for (j=1; j<lx; j++) g[j]=lgetg(lx,t_COL);
  for (i=1; i<lx; i++)
    for (j=1; j<=i; j++)
      coeff(g,i,j)=coeff(g,j,i)=(long)gscal((GEN)x[i],(GEN)x[j]);
  av1=avma; return gerepile(av,av1,f(g,0));
}

GEN
lllkerim(GEN x) { return lllkerim_proto(x,lllgramall); }

/* corrected bug : x=[; ] => used to return [; ] instead of [[; ], [; ]] */
GEN
lllkerimgen(GEN x) { return lllkerim_proto(x,lllgramallgen); }

static int
pslg(GEN x)
{
  long tx=typ(x);
  if (gcmp0(x)) return 2;
  return is_scalar_t(tx) ? 3 : lgef(x);
}

static GEN
lllgramallgen(GEN x, long all)
{
  long av=avma,tetpil,lx=lg(x),tu,i,j,k,l,n,lim;
  GEN u,B,lam,q,cq,h,la,bb,p1,p2,p3,p4,y,fl;
  int ps1,ps2,flc;

  if (typ(x) != t_MAT) err(lllger1);
  n=lx-1; if (n<=1) return idmat(n);
  if (lg((GEN)x[1])!=lx) err(lllger2);
  av=avma; lim=(avma+bot)>>1;
  B=cgetg(lx+1,t_COL);
  fl=cgetg(lx,t_VEC);
  B[1]=un; lam=cgetg(lx,t_MAT);
  for (j=1; j<lx; j++) lam[j]=lgetg(lx,t_COL);
  for (i=1; i<lx; i++)
    for (j=1; j<=i; j++)
    {
      if (j<i && !signe(fl[j])) coeff(lam,i,j)=coeff(lam,j,i)=zero;
      else
      {
	u=gcoeff(x,i,j); tu=typ(u);
	if (! is_scalar_t(tu) && tu != t_POL) err(lllger4);
	for (k=1; k<j; k++)
	  if (signe(fl[k]))
	    u=gdiv(
                gsub(
		   gmul((GEN)B[k+1],u),
                   gmul(gcoeff(lam,i,k),gcoeff(lam,j,k))
		),
		(GEN)B[k]
	      );
	if (j<i) { coeff(lam,i,j)=(long)u; coeff(lam,j,i)=zero; }
	else
	{
	  if (!gcmp0(u)) { B[i+1]=(long)u; coeff(lam,i,i)=fl[i]=un; }
	  else { B[i+1]=B[i]; coeff(lam,i,i)=fl[i]=zero; }
	 }
      }
    }
  k=2; h=idmat(n); flc=0;
  for(;;)
  {
    u=gcoeff(lam,k,k-1);
    if (pslg(u) >= pslg((GEN)B[k]))
    {
      q=gdeuc(u,(GEN)B[k]); cq=gdivsg(1,content(q)); q=gmul(q,cq); flc=1;
      h[k]=lsub(gmul(cq,(GEN)h[k]),gmul(q,(GEN)h[k-1]));
      coeff(lam,k,k-1)=lsub(gmul(cq,gcoeff(lam,k,k-1)),gmul(q,(GEN)B[k]));
      for (i=1; i<k-1; i++)
	coeff(lam,k,i)=lsub(gmul(cq,gcoeff(lam,k,i)),gmul(q,gcoeff(lam,k-1,i)));
    }
    ps1 = pslg(gsqr((GEN)B[k]));
    p3 = gmul((GEN)B[k-1],(GEN)B[k+1]);
    la=gcoeff(lam,k,k-1); p4 = gmul(la,gcoeff(lam,k,k-1));
    ps2=pslg(gadd(p3,p4));
    if (signe(fl[k-1]) &&
           (ps1>ps2 || (ps1==ps2 && flc) || !signe(fl[k])) )
    {
      p1=(GEN)h[k-1]; h[k-1]=h[k]; h[k]=(long)p1;
      if ((ps1==ps2)&&flc) flc=0; else flc=1;
      for (j=1; j<=k-2; j++)
      {
	p1=gcoeff(lam,k-1,j); coeff(lam,k-1,j)=coeff(lam,k,j);
	coeff(lam,k,j)=(long)p1;
      }
      if (signe(fl[k]))
      {
	for (i=k+1; i<=n; i++)
	{
	  bb=gcoeff(lam,i,k);
	  coeff(lam,i,k)=ldiv(gsub(gmul((GEN)B[k+1],gcoeff(lam,i,k-1)),gmul(la,bb)),(GEN)B[k]);
	  coeff(lam,i,k-1)=ldiv(gadd(gmul(la,gcoeff(lam,i,k-1)),gmul((GEN)B[k-1],bb)),(GEN)B[k]);
	 }
	B[k]=ldiv(gadd(p3,p4),(GEN)B[k]);
      }
      else
      {
	if (!gcmp0(la))
	{
	  p2=(GEN)B[k]; p1=gdiv(p4,p2);
	  for (i=k+1; i<lx; i++)
	    coeff(lam,i,k-1)=ldiv(gmul(la,gcoeff(lam,i,k-1)),p2);
	  for (j=k+1; j<lx-1; j++)
	    for (i=j+1; i<lx; i++)
	      coeff(lam,i,j)=ldiv(gmul(p1,gcoeff(lam,i,j)),p2);
	  B[k+1]=B[k]=(long)p1;
	  for (i=k+2; i<=lx; i++)
	    B[i]=ldiv(gmul(p1,(GEN)B[i]),p2);
	 }
	else
	{
	  coeff(lam,k,k-1)=zero;
	  for (i=k+1; i<lx; i++)
	  { coeff(lam,i,k)=coeff(lam,i,k-1); coeff(lam,i,k-1)=zero; }
	  B[k]=B[k-1]; fl[k]=un; fl[k-1]=zero;
	 }
      }
      if (k>2) k--;
    }
    else
    {
      for (l=k-2; l>=1; l--)
      {
	u=gcoeff(lam,k,l);
	if (pslg(u)>=pslg((GEN)B[l+1]))
	{
	  q=gdeuc(u,(GEN)B[l+1]); cq=gdivsg(1,content(q)); q=gmul(q,cq); flc=1;
	  h[k]=lsub(gmul(cq,(GEN)h[k]),gmul(q,(GEN)h[l]));
	  coeff(lam,k,l)=lsub(gmul(cq,gcoeff(lam,k,l)),gmul(q,(GEN)B[l+1]));
	  for (i=1; i<l; i++) coeff(lam,k,i)=lsub(gmul(cq,gcoeff(lam,k,i)),gmul(q,gcoeff(lam,l,i)));
	 }
      }
      k++;
      if (k>n)
      {
	k=1; while (k<=n && !signe(fl[k])) k++;
	tetpil=avma;
	if (!all)
	{
	  y=cgetg(3,t_VEC);
	  p2=cgetg(k,t_MAT); for (i=1; i<k; i++) p2[i]=lcopy((GEN)h[i]);
	  y[1]=(long)p2; p2=cgetg(n-k+2,t_MAT); y[2]=(long)p2;
	  for (i=k; i<=n; i++) p2[i-k+1]=lcopy((GEN)h[i]);
	 }
	else
	{
	  if (all==1)
	  {
	    y=cgetg(k,t_MAT); for (i=1; i<k; i++) y[i]=lcopy((GEN)h[i]);
	  }
	  else
	  {
	    y=cgetg(n-k+2,t_MAT);
	    for (i=k; i<=n; i++) y[i-k+1]=lcopy((GEN)h[i]);
	  }
	 }
	return gerepile(av,tetpil,y);
      }
    }
    if (avma<lim)
    {
      GEN *gptr[4];
      gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
      gerepilemany(av,gptr,4);
    }
  }
}

/* x est ici la matrice de GRAM des bi. */
GEN
lllgramold(GEN x, long prec)
{
  GEN mu,u,B,BB,BK,p,q,r,cst,unreel,sv,mu1,mu2;
  long av,tetpil,lim,l,i,j,k,lx=lg(x),n,temp,e;

  if (typ(x) != t_MAT) err(lllger1);
  if (lg((GEN)x[1])!=lx) err(lllger2); n=lx-1;
  if (n<=1) return idmat(n);
  cst=gdivgs(stoi(99),100);
  lim=(avma+bot)>>1;
  if (prec)
  {
    affsr(1,unreel=cgetr(prec+1));
    x=gmul(x,unreel);
    cst=gmul(cst,unreel);
  }
  av=avma;
  mu=sqred3(x); B=cgetg(lx,t_COL);
  for (i=1,l=0; i<=n; i++)
  {
    if (gsigne((GEN)(B[i]=coeff(mu,i,i)))>0) l++;
    coeff(mu,i,i)=un;
  }
  if (l<n)
  {
    if (DEBUGLEVEL) outerr(x);
    err(lllger3);
  }
  u=idmat(n);
  k=2;
  do
  {
    if (!gcmp0(r=grndtoi(gcoeff(mu,k,k-1),&e)))
    {
      u[k]=lsub((GEN)u[k],gmul(r,(GEN)u[k-1]));
      for (j=1; j<k-1; j++)
	coeff(mu,k,j)=lsub(gcoeff(mu,k,j),gmul(r,gcoeff(mu,k-1,j)));
      mu1=(GEN)(coeff(mu,k,k-1)=lsub(gcoeff(mu,k,k-1),r));
    }
    else mu1=gcoeff(mu,k,k-1);
    q=gmul((GEN)B[k-1],gsub(cst,mu2=gsqr(mu1)));
    if (gcmp(q,(GEN)B[k])>0)
    {
      BB=gadd((GEN)B[k],gmul((GEN)B[k-1],mu2));
      coeff(mu,k,k-1)=ldiv(gmul(mu1,(GEN)B[k-1]),BB);
      B[k]=lmul((GEN)B[k-1],BK=gdiv((GEN)B[k],BB));
      B[k-1]=(long)BB;
      temp=u[k]; u[k]=u[k-1]; u[k-1]=temp;
      for (j=1; j<=k-2; j++)
      {
	temp=(long)coeff(mu,k,j); coeff(mu,k,j)=coeff(mu,k-1,j);
	coeff(mu,k-1,j)=temp;
      }
      for (i=k+1; i<=n; i++)
      {
	p=gcoeff(mu,i,k);
	coeff(mu,i,k)=lsub(gcoeff(mu,i,k-1),gmul(mu1,p));
	coeff(mu,i,k-1)=ladd(gmul(BK,p),gmul(gcoeff(mu,k,k-1),gcoeff(mu,i,k-1)));
      }
      if (k>2) k--;
    }
    else
    {
      for (l=k-2; l; l--)
      {
	if (!gcmp0(r=grndtoi(gcoeff(mu,k,l),&e)))
	{
	  u[k]=lsub((GEN)u[k],gmul(r,(GEN)u[l]));
	  for (j=1; j<l; j++)
	    coeff(mu,k,j)=lsub(gcoeff(mu,k,j),gmul(r,gcoeff(mu,l,j)));
	  coeff(mu,k,l)=lsub(gcoeff(mu,k,l),r);
	 }
      }
      k++;
    }
    if (avma<lim)
    {
      tetpil=avma;
      sv=cgetg(4,t_VEC);
      sv[1]=lcopy(B); sv[2]=lcopy(u); sv[3]=lcopy(mu);
      sv=gerepile(av,tetpil,sv);
      B=(GEN)sv[1]; u=(GEN)sv[2]; mu=(GEN)sv[3];
    }
  }
  while (k<=n);
  tetpil=avma; return gerepile(av,tetpil,gcopy(u));
}

/* x est ici la matrice de GRAM des bi.  */
GEN
lllgram(GEN x, long prec)
{
  GEN mu,u,B,BB,BK,p,q,r,cst,unreel,mu1,mu2,p1,A,s,p2,xinit;
  GEN *gptr[6];
  long av0,av,tetpil,lim,l,i,j,k,k1,lx=lg(x),n,temp,e,kmax,l1,l2,cmpt;

  if (typ(x) != t_MAT) err(lllger1);
  if (lg((GEN)x[1])!=lx) err(lllger2); n=lx-1;
  if (n<=1) return idmat(n);
  av0=avma; cmpt=0; xinit=x;

LABLLLGRAM:
  av=avma; cst=gdivgs(stoi(99),100); /* LLL proposent 0.75 */
  lim=(avma+bot)>>1; l1=2;
  for (j=1; j<lx; j++)
  {
    p2=(GEN)x[j];
    for (i=1; i<lx; i++)
    {
    /* A MODIFIER lg <-> expo */
      if (typ(p2[i]) == t_REAL) { l2=lg((GEN)p2[i]); if (l2>l1) l1=l2; }
    }
  }
  if (l1>2)
  {
    if (prec>l1) l1=prec;
    l2 = (long)((l1-2)*K);
    x = gprec(x,l2);
    cst = gprec(cst,l2);
  }
  else
  {
    if (prec)
    { 
      affsr(1,unreel=cgetr(prec+1));
      x=gmul(x,unreel); cst=gmul(cst,unreel);
    }
    else { avma=av0; return lllgramint(x); }
  }

  mu=cgetg(lx,t_MAT); B=cgetg(lx,t_COL); A=cgetg(lx,t_VEC);
  for (j=1; j<lx; j++)
  {
    p1=cgetg(lx,t_COL); mu[j]=(long)p1; for (i=1; i<lx; i++) p1[i]=zero;
    B[j]=zero; A[j]=zero;
  }
  u=idmat(n); k=2; kmax=1; B[1]=coeff(x,1,1);
  if (DEBUGLEVEL>5) { fprintferr("k = %ld",k); flusherr(); }
  do
  {
    if (k>kmax)
    {
      if (DEBUGLEVEL>=4) {fprintferr(" K%ld",k);flusherr();}
      kmax=k;
      for (j=1; j<k; j++)
      {
	s=gzero; for (i=1; i<j; i++) s=gadd(s,gmul(gcoeff(mu,j,i),(GEN)A[i]));
	s=gsub(gcoeff(x,k,j),s);
	A[j]=(long)s;
	coeff(mu,k,j)=ldiv(s,(GEN)B[j]);
      }
      s=gzero; for (i=1; i<k; i++) s=gadd(s,gmul(gcoeff(mu,k,i),(GEN)A[i]));
      s=gsub(gcoeff(x,k,k),s); B[k]=(long)s;
      if (gsigne(s)<=0)
      {
	cmpt++; avma=av0;
	if (cmpt<4)
	{
	  x=xinit; prec=(prec<<1)-2; avma=av;
	  if(DEBUGLEVEL)
	  {
	    fprintferr("*** Warning: doubling precision in lllgram\n");
	    flusherr();
	  }
	  goto LABLLLGRAM;
	}
	else{ if (DEBUGLEVEL) outerr(xinit); err(lllger3); }
      }
    }
    if (DEBUGLEVEL>9)
      fprintferr(" %ld",gexpo(gcoeff(mu,k,k-1))-((lg(gcoeff(mu,k,k-1))-2)<<TWOPOTBITS_IN_LONG));
    mu1=gcoeff(mu,k,k-1);
    if ((!gcmp0(mu1))&&((2*gexpo(mu1))>((lg(mu1)-2)<<TWOPOTBITS_IN_LONG)))
    {
      if (DEBUGLEVEL>9)
      {
	fprintferr("\nRecalcul de Gram Schmidt dans lllgram pour kmax = %ld\n Valeurs de B avant :\n",kmax); outerr(B);
      }
      for (k1=1; k1<=kmax; k1++)
      {
	for (j=1; j<k1; j++)
	{
	  s=gzero; for (i=1; i<j; i++) s=gadd(s,gmul(gcoeff(mu,j,i),(GEN)A[i]));
	  s=gsub(gcoeff(x,k1,j),s); A[j]=(long)s; coeff(mu,k1,j)=ldiv(s,(GEN)B[j]);
	}
	s=gzero; for (i=1; i<k1; i++) s=gadd(s,gmul(gcoeff(mu,k1,i),(GEN)A[i]));
	s=gsub(gcoeff(x,k1,k1),s); B[k1]=(long)s;
	if (gcmp0(s))
	{
	  cmpt++; avma=av0;
	  if (cmpt<3){ x=xinit; prec=(prec<<1)-2; goto LABLLLGRAM; }
	  else{ if (DEBUGLEVEL) outerr(x); err(lllger3); }
	}
      }
      if (avma<lim)
      {
	gptr[0]=&B; gptr[1]=&u; gptr[2]=&mu; gptr[3]=&A;
        gerepilemany(av,gptr,4);
      }
      mu1=gcoeff(mu,k,k-1);
      if (DEBUGLEVEL>9){ fprintferr("Valeurs de B apres :\n"); outerr(B); }
    }
    if (!gcmp0(r=grndtoi(mu1,&e)))
    {
      r=negi(r);
      for (j=1; j<=kmax; j++) coeff(u,j,k)=laddii(gcoeff(u,j,k),mulii(r,gcoeff(u,j,k-1)));
      for (j=1; j<=n; j++) coeff(x,j,k)=ladd(gcoeff(x,j,k),gmul(r,gcoeff(x,j,k-1)));
      for (j=1; j<=n; j++) coeff(x,k,j)=ladd(gcoeff(x,k,j),gmul(r,gcoeff(x,k-1,j)));
      for (j=1; j<k-1; j++) coeff(mu,k,j)=ladd(gcoeff(mu,k,j),gmul(r,gcoeff(mu,k-1,j)));
      mu1=(GEN)(coeff(mu,k,k-1)=ladd(mu1,r));
    }
    if (DEBUGLEVEL>9) fprintferr("bits d'erreur d'arrondi dans lllgram: %ld",e);
    q=gmul((GEN)B[k-1],gsub(cst,mu2=gsqr(mu1)));
    if (gcmp(q,(GEN)B[k])>0)
    {
      if ((DEBUGLEVEL>=4)&&(k==kmax))
	{ fprintferr(" (%ld)",gexpo(q)-gexpo((GEN)B[k])); flusherr(); }
      BB=gadd((GEN)B[k],gmul((GEN)B[k-1],mu2));
      coeff(mu,k,k-1)=ldiv(gmul(mu1,(GEN)B[k-1]),BB);
      B[k]=lmul((GEN)B[k-1],BK=gdiv((GEN)B[k],BB)); B[k-1]=(long)BB;
      temp=u[k]; u[k]=u[k-1]; u[k-1]=temp;
      temp=x[k]; x[k]=x[k-1]; x[k-1]=temp;
      for (j=1; j<=n; j++)
      { temp=coeff(x,k,j); coeff(x,k,j)=coeff(x,k-1,j); coeff(x,k-1,j)=temp; }
      for (j=1; j<=k-2; j++)
      { temp=coeff(mu,k,j); coeff(mu,k,j)=coeff(mu,k-1,j); coeff(mu,k-1,j)=temp; }
      for (i=k+1; i<=kmax; i++)
      {
	p=gcoeff(mu,i,k); coeff(mu,i,k)=lsub(gcoeff(mu,i,k-1),gmul(mu1,p));
	coeff(mu,i,k-1)=ladd(gmul(BK,p),gmul(gcoeff(mu,k,k-1),gcoeff(mu,i,k-1)));
      }
      if (k>2) k--;
      if (DEBUGLEVEL>5) { fprintferr(" %ld",k); flusherr(); }
      if (DEBUGLEVEL>9 && k==2)
      {
	fprintferr("\n");
	for (i=1; i<lx; i++)
	{
	  fprintferr("%ld : ",i); outerr(qfeval(x,(GEN)u[i]));
	}
      }
    }
    else
    {
      for (l=k-2; l; l--)
      {
	if (DEBUGLEVEL>9)
	  fprintferr(" %ld",gexpo(gcoeff(mu,k,l))-((lg(gcoeff(mu,k,l))-2)<<TWOPOTBITS_IN_LONG));
	if (!gcmp0(r=grndtoi(gcoeff(mu,k,l),&e)))
	{
	  r=negi(r);
	  for (j=1; j<=kmax; j++)
	    coeff(u,j,k)=laddii(gcoeff(u,j,k),mulii(r,gcoeff(u,j,l)));
	  for (j=1; j<=n; j++)
	    coeff(x,j,k)=ladd(gcoeff(x,j,k),gmul(r,gcoeff(x,j,l)));
	  for (j=1; j<=n; j++)
	    coeff(x,k,j)=ladd(gcoeff(x,k,j),gmul(r,gcoeff(x,l,j)));
	  for (j=1; j<l; j++)
	    coeff(mu,k,j)=ladd(gcoeff(mu,k,j),gmul(r,gcoeff(mu,l,j)));
	  coeff(mu,k,l)=ladd(gcoeff(mu,k,l),r);
	 }
	if (DEBUGLEVEL>9)
	  fprintferr("bits d'erreur d'arrondi dans lllgram: %ld",e);
      }
      k++; if (DEBUGLEVEL>5) { fprintferr(" %ld",k); flusherr(); }
      if (DEBUGLEVEL>9)
      {
	fprintferr("\n k = %ld\n",k);
	for (i=1; i<lx; i++)
	{
	  fprintferr("%ld : ",i); outerr(qfeval(x,(GEN)u[i]));
	}
      }
    }
    if (avma<lim)
    {
      gptr[0]=&B; gptr[1]=&u; gptr[2]=&mu; gptr[3]=&A;
      gptr[4]=&cst; gptr[5]=&x; gerepilemany(av,gptr,6);
    }
  }
  while (k<=n);
  if (DEBUGLEVEL>=3) { fprintferr("\n"); flusherr(); }
  tetpil=avma; return gerepile(av0,tetpil,gcopy(u));
}

/* x est ici la matrice de GRAM des bi.  */
GEN
lllgram_new(GEN x, long prec)
/* A DEBBUGER ...  A DEBBUGER ...  A DEBBUGER ...  A DEBBUGER ...  */
{
  GEN mu,u,B,BB,BK,p,q,r,cst,unreel,mu1,mu2,p1,A,s,p2,eps;
  GEN *gptr[6];
  long av0,av,tetpil,lim,l,i,j,k,k1,lx=lg(x),n,temp,e,kmax,l1,l2,cmpt;

  if (typ(x) != t_MAT) err(lllger1);
  if (lg((GEN)x[1])!=lx) err(lllger2); n=lx-1;
  if (n<=1) return idmat(n);
  av0=avma; cmpt=0; u=idmat(n); eps=dbltor(0.00000001);

LABLLLGRAM:
  av=avma; cst=gdivgs(stoi(99),100); /* LLL proposent 0.75 */
  lim=(avma+bot)>>1;
  l1=2;
  for (j=1; j<lx; j++)
  {
    p2=(GEN)x[j];
    for (i=1; i<lx; i++)
      if (typ(p2[i])==t_REAL) { l2=lg((GEN)p2[i]); if (l2>l1) l1=l2; }
  }
  if (l1>2)
  {
    if (prec>l1) l1=prec;
    l2=(long)((l1-2)*K); x=gprec(x,l2); cst=gprec(cst,l2);
  }
  else
  {
    if (prec)
    {
      affsr(1, unreel=cgetr(prec+1)); x=gmul(x,unreel); cst=gmul(cst,unreel);
    }
    else { avma=av0; return lllgramint(x); }
  }
  mu=cgetg(lx,t_MAT); B=cgetg(lx,t_COL); A=cgetg(lx,t_VEC);
  for (j=1; j<lx; j++)
  {
    p1=cgetg(lx,t_COL); mu[j]=(long)p1; for (i=1; i<lx; i++) p1[i]=zero;
    B[j]=zero; A[j]=zero;
  }
  k=2; kmax=1; B[1]=coeff(x,1,1);
  if (DEBUGLEVEL>5) { fprintferr("k = %ld",k); flusherr(); }
  do
  {
    if (k>kmax)
    {
      kmax=k;
      for (j=1; j<k; j++)
      {
	s=gzero; for (i=1; i<j; i++) s=gadd(s,gmul(gcoeff(mu,j,i),(GEN)A[i]));
	s=gsub(gcoeff(x,k,j),s); A[j]=(long)s; coeff(mu,k,j)=ldiv(s,(GEN)B[j]);
      }
      s=gzero; for (i=1; i<k; i++) s=gadd(s,gmul(gcoeff(mu,k,i),(GEN)A[i]));
      s=gsub(gcoeff(x,k,k),s); B[k]=(long)s;
      if (gcmp(gabs(s,prec),eps)<0)
      {
	cmpt++;
	if (cmpt<3)
	{
	  tetpil=avma; x=gmul(x,u); u=gcopy(u);
	  gptr[0]=&x; gptr[1]=&u; gerepilemanysp(av,tetpil,gptr,2);
	  prec=(prec<<1)-2; goto LABLLLGRAM;
	}
	else { if (DEBUGLEVEL) outerr(x); avma=av0; err(lllger3); }
      }
    }
    if (DEBUGLEVEL>9)
      fprintferr(" %ld",gexpo(gcoeff(mu,k,k-1))-((lg(gcoeff(mu,k,k-1))-2)<<TWOPOTBITS_IN_LONG));
    mu1=gcoeff(mu,k,k-1);
    if ((!gcmp0(mu1))&&((2*gexpo(mu1))>((lg(mu1)-2)<<TWOPOTBITS_IN_LONG)))
    {
      if (DEBUGLEVEL>9)
      {
	fprintferr("\nRecalcul de Gram Schmidt dans lllgram pour kmax = %ld\n Valeurs de B avant :\n",kmax); outerr(B);
      }
      for (k1=1; k1<=kmax; k1++)
      {
	for (j=1; j<k1; j++)
	{
	  s=gzero;
	  for (i=1; i<j; i++)
	    s=gadd(s,gmul(gcoeff(mu,j,i),(GEN)A[i]));
	  s=gsub(gcoeff(x,k1,j),s); A[j]=(long)s;
	  coeff(mu,k1,j)=ldiv(s,(GEN)B[j]);
	 }
	s=gzero;
	for (i=1; i<k1; i++)
	  s=gadd(s,gmul(gcoeff(mu,k1,i),(GEN)A[i]));
	s=gsub(gcoeff(x,k1,k1),s); B[k1]=(long)s;
	if (gcmp(gabs(s,prec),eps)<0)
	{
	  cmpt++;
	  if (cmpt<3)
	  {
	    tetpil=avma; x=gmul(x,u); u=gcopy(u);
	    gptr[0]=&x; gptr[1]=&u; gerepilemanysp(av,tetpil,gptr,2);
	    prec=(prec<<1)-2; goto LABLLLGRAM;
	  }
	  else{ if (DEBUGLEVEL) outerr(x); avma=av0; err(lllger3); }
	 }
      }
      if (avma<lim)
      {
	gptr[0]=&B; gptr[1]=&u; gptr[2]=&mu; gptr[3]=&A;
	gerepilemany(av,gptr,4);
      }
      mu1=gcoeff(mu,k,k-1);
      if (DEBUGLEVEL>9){ fprintferr("Valeurs de B apres :\n"); outerr(B); }
    }
    if (!gcmp0(r=grndtoi(mu1,&e)))
    {
      r=negi(r);
      for (j=1; j<=kmax; j++)
	coeff(u,j,k)=laddii(gcoeff(u,j,k),mulii(r,gcoeff(u,j,k-1)));
      for (j=1; j<=n; j++)
	coeff(x,j,k)=ladd(gcoeff(x,j,k),gmul(r,gcoeff(x,j,k-1)));
      for (j=1; j<=n; j++)
	coeff(x,k,j)=ladd(gcoeff(x,k,j),gmul(r,gcoeff(x,k-1,j)));
      for (j=1; j<k-1; j++)
	coeff(mu,k,j)=ladd(gcoeff(mu,k,j),gmul(r,gcoeff(mu,k-1,j)));
      mu1=(GEN)(coeff(mu,k,k-1)=ladd(mu1,r));
    }
    if (DEBUGLEVEL>9) fprintferr("bits d'erreur d'arrondi dans lllgram: %ld",e);
    q=gmul((GEN)B[k-1],gsub(cst,mu2=gsqr(mu1)));
    if (gcmp(q,(GEN)B[k])>0)
    {
      if (DEBUGLEVEL>4 && k==n)
      { fprintferr("ex = %ld",gexpo(gsub(q,(GEN)B[k]))); flusherr(); }
      BB=gadd((GEN)B[k],gmul((GEN)B[k-1],mu2));
      coeff(mu,k,k-1)=ldiv(gmul(mu1,(GEN)B[k-1]),BB);
      B[k]=lmul((GEN)B[k-1],BK=gdiv((GEN)B[k],BB)); B[k-1]=(long)BB;
      temp=u[k]; u[k]=u[k-1]; u[k-1]=temp;
      temp=x[k]; x[k]=x[k-1]; x[k-1]=temp;
      for (j=1; j<=n; j++)
      { 
	temp=coeff(x,k,j); coeff(x,k,j)=coeff(x,k-1,j);
	coeff(x,k-1,j)=temp;
      }
      for (j=1; j<=k-2; j++)
      {
	temp=coeff(mu,k,j); coeff(mu,k,j)=coeff(mu,k-1,j);
	coeff(mu,k-1,j)=temp;
      }
      for (i=k+1; i<=kmax; i++)
      {
	p=gcoeff(mu,i,k); coeff(mu,i,k)=lsub(gcoeff(mu,i,k-1),gmul(mu1,p));
	coeff(mu,i,k-1)=ladd(gmul(BK,p),gmul(gcoeff(mu,k,k-1),gcoeff(mu,i,k-1)));
      }
      if (k>2) k--;
      if (DEBUGLEVEL>5) { fprintferr(" %ld",k); flusherr(); }
      if (DEBUGLEVEL>9 && k==2)
      {
	fprintferr("\n");
	for (i=1; i<lx; i++)
	{
	  fprintferr("%ld : ",i); outerr(qfeval(x,(GEN)u[i]));
	}
      }
    }
    else
    {
      for (l=k-2; l; l--)
      {
	if (DEBUGLEVEL>9)
	  fprintferr(" %ld",gexpo(gcoeff(mu,k,l))-((lg(gcoeff(mu,k,l))-2)<<TWOPOTBITS_IN_LONG));
	if (!gcmp0(r=grndtoi(gcoeff(mu,k,l),&e)))
	{
	  r=negi(r);
	  for (j=1; j<=kmax; j++)
	    coeff(u,j,k)=laddii(gcoeff(u,j,k),mulii(r,gcoeff(u,j,l)));
	  for (j=1; j<=n; j++)
	    coeff(x,j,k)=ladd(gcoeff(x,j,k),gmul(r,gcoeff(x,j,l)));
	  for (j=1; j<=n; j++)
	    coeff(x,k,j)=ladd(gcoeff(x,k,j),gmul(r,gcoeff(x,l,j)));
	  for (j=1; j<l; j++)
	    coeff(mu,k,j)=ladd(gcoeff(mu,k,j),gmul(r,gcoeff(mu,l,j)));
	  coeff(mu,k,l)=ladd(gcoeff(mu,k,l),r);
	 }
	if (DEBUGLEVEL>9) fprintferr("bits d'erreur d'arrondi dans lllgram: %ld",e);
      }
      k++;
      if (DEBUGLEVEL>5) { fprintferr(" %ld",k); flusherr(); }
      if (DEBUGLEVEL>9)
      {
	fprintferr("\n k = %ld\n",k);
	for (i=1; i<lx; i++)
	{ 
	  fprintferr("%ld : ",i); outerr(qfeval(x,(GEN)u[i]));
	}
      }
    }
    if (avma<lim)
    {
      gptr[0]=&B; gptr[1]=&u; gptr[2]=&mu; gptr[3]=&A;
      gptr[4]=&cst; gptr[5]=&x; gerepilemany(av,gptr,6);
    }
  }
  while (k<=n);
  if (DEBUGLEVEL>=3) { fprintferr("\n"); flusherr(); }
  tetpil=avma; return gerepile(av0,tetpil,gcopy(u));
}

/* x est ici la matrice de GRAM des bi.  */
GEN
lllgram1(GEN x, long prec)
{

  GEN  mu,u,B,BB,BK,p,q,r,cst,unreel,sv,mu1,mu2;
  long av,tetpil,lim,l,i,j,k,lx=lg(x),n,temp,e;

  if (typ(x) != t_MAT) err(lllger1);
  if (lg((GEN)x[1])!=lx) err(lllger2); n=lx-1;
  if (n<=1) return idmat(n);
  cst=gdivgs(stoi(99),100); /* LLL proposent 0.75 */
  lim=(avma+bot)>>1;
  if (prec)
  {
    affsr(1,unreel=cgetr(prec+1));
    x=gmul(x,unreel);
    cst=gmul(cst,unreel);
  }
  av=avma;
  mu=gtrans(sqred(x)); B=cgetg(lx,t_COL);
  for (i=1,l=0; i<=n; i++)
  {
    if (gsigne((GEN)(B[i]=coeff(mu,i,i)))>0) l++;
    coeff(mu,i,i)=un;
  }
  if (l<n) err(lllger3);

  u=idmat(n);
  k=2;
  do
  {
    if (!gcmp0(r=grndtoi(gcoeff(mu,k,k-1),&e)))
    {
      u[k]=lsub((GEN)u[k],gmul(r,(GEN)u[k-1]));
      for (j=1; j<k-1; j++)
	coeff(mu,k,j)=lsub(gcoeff(mu,k,j),gmul(r,gcoeff(mu,k-1,j)));
      mu1=(GEN)(coeff(mu,k,k-1)=lsub(gcoeff(mu,k,k-1),r));
    }
    else mu1=gcoeff(mu,k,k-1);
    q=gmul((GEN)B[k-1],gsub(cst,mu2=gsqr(mu1)));
    if (gcmp(q,(GEN)B[k])>0)
    {
      BB=gadd((GEN)B[k],gmul((GEN)B[k-1],mu2));
      coeff(mu,k,k-1)=ldiv(gmul(mu1,(GEN)B[k-1]),BB);
      B[k]=lmul((GEN)B[k-1],BK=gdiv((GEN)B[k],BB));
      B[k-1]=(long)BB;
      temp=u[k]; u[k]=u[k-1]; u[k-1]=temp;
      for (j=1; j<=k-2; j++)
      {
	temp=(long)coeff(mu,k,j); coeff(mu,k,j)=coeff(mu,k-1,j);
	coeff(mu,k-1,j)=temp;
      }
      for (i=k+1; i<=n; i++)
      {
	p=gcoeff(mu,i,k);
	coeff(mu,i,k)=lsub(gcoeff(mu,i,k-1),gmul(mu1,p));
	coeff(mu,i,k-1)=ladd(gmul(BK,p),gmul(gcoeff(mu,k,k-1),gcoeff(mu,i,k-1)));
      }
      if (k>2) k--;
    }
    else
    {
      for (l=k-2; l; l--)
      {
	if (!gcmp0(r=grndtoi(gcoeff(mu,k,l),&e)))
	{
	  u[k]=lsub((GEN)u[k],gmul(r,(GEN)u[l]));
	  for (j=1; j<l; j++)
	    coeff(mu,k,j)=lsub(gcoeff(mu,k,j),gmul(r,gcoeff(mu,l,j)));
	  coeff(mu,k,l)=lsub(gcoeff(mu,k,l),r);
	 }
      }
      k++;
    }
    if (avma<lim)
    {
      tetpil=avma;
      sv=cgetg(4,t_VEC);
      sv[1]=lcopy(B); sv[2]=lcopy(u); sv[3]=lcopy(mu);
      sv=gerepile(av,tetpil,sv);
      B=(GEN)sv[1]; u=(GEN)sv[2]; mu=(GEN)sv[3];
    }
  }
  while (k<=n);
  tetpil=avma; return gerepile(av,tetpil,gcopy(u));
}

GEN
lllgramintindep(GEN x)
{
  long av=avma,tetpil,lx=lg(x),i,j,k,l,p1,n,lim;
  GEN u,B,lam,q,r,h,la,bb,sv;

  if (typ(x) != t_MAT) err(lllger1);
  n=lx-1; if (n<=1) return idmat(n);
  if (lg((GEN)x[1])!=lx) err(lllger2);
  av=avma; lim=(avma+bot)>>1;
  B=cgetg(lx+1,t_COL);
  B[1]=un; lam=cgetg(lx,t_MAT);
  for (j=1; j<=n; j++) lam[j]=lgetg(lx,t_COL);
  for (i=1; i<=n; i++)
    for (j=1; j<=i; j++)
    {
      u=gcoeff(x,i,j);
      if (typ(u)!=t_INT) err(lllger4);
      for (k=1; k<j; k++)
	u=divii(subii(mulii((GEN)B[k+1],u),mulii(gcoeff(lam,i,k),gcoeff(lam,j,k))),(GEN)B[k]);
      if (j<i) coeff(lam,i,j)=(long)u;
      else
      {
	if (signe(u)) B[i+1]=(long)u;
	else err(lllger5);
      }
      coeff(lam,j,i)=zero;
    }
  k=2; h=idmat(n);
  for(;;)
  {
    if (cmpii(absi(u=shifti(gcoeff(lam,k,k-1),1)),(GEN)B[k])>0)
    {
      q=dvmdii(addii(u,(GEN)B[k]),shifti((GEN)B[k],1),&r);
      if (signe(r)<0) q=addsi(-1,q);
      h[k]=lsub((GEN)h[k],gmul(q,(GEN)h[k-1]));
      coeff(lam,k,k-1)=lsubii(gcoeff(lam,k,k-1),mulii(q,(GEN)B[k]));
      for (i=1; i<k-1; i++) coeff(lam,k,i)=lsubii(gcoeff(lam,k,i),mulii(q,gcoeff(lam,k-1,i)));
    }
    if (cmpii(mulsi(99,sqri((GEN)B[k])),mulsi(100,addii(mulii((GEN)B[k-1],(GEN)B[k+1]),mulii(gcoeff(lam,k,k-1),gcoeff(lam,k,k-1)))))>0)
    {
      la=gcoeff(lam,k,k-1); p1=h[k-1]; h[k-1]=h[k]; h[k]=p1;
      for (j=1; j<=k-2; j++)
      {
	p1=(long)coeff(lam,k-1,j); coeff(lam,k-1,j)=coeff(lam,k,j);
	coeff(lam,k,j)=p1;
      }
      for (i=k+1; i<=n; i++)
      {
	bb=gcoeff(lam,i,k);
	coeff(lam,i,k)=ldivii(subii(mulii((GEN)B[k+1],gcoeff(lam,i,k-1)),mulii(la,bb)),(GEN)B[k]);
	coeff(lam,i,k-1)=ldivii(addii(mulii(la,gcoeff(lam,i,k-1)),mulii((GEN)B[k-1],bb)),(GEN)B[k]);
      }
      B[k]=ldivii(addii(mulii((GEN)B[k-1],(GEN)B[k+1]),mulii(la,la)),(GEN)B[k]);
      if (k>2) k--;
    }
    else
    {
      for (l=k-2; l>=1; l--)
      {
	if (cmpii(absi(u=shifti(gcoeff(lam,k,l),1)),(GEN)B[l+1])>0)
	{
	  q=dvmdii(addii(u,(GEN)B[l+1]),shifti((GEN)B[l+1],1),&r);
	  if (signe(r)<0) q=addsi(-1,q);
	  h[k]=lsub((GEN)h[k],gmul(q,(GEN)h[l]));
	  coeff(lam,k,l)=lsubii(gcoeff(lam,k,l),mulii(q,(GEN)B[l+1]));
	  for (i=1; i<l; i++) coeff(lam,k,i)=lsubii(gcoeff(lam,k,i),mulii(q,gcoeff(lam,l,i)));
	 }
      }
      k++;
      if (k>n) { tetpil=avma; return gerepile(av,tetpil,gcopy(h)); }
    }
    if (avma<lim)
    {
      tetpil=avma;
      sv=cgetg(4,t_VEC);
      sv[1]=lcopy(B); sv[2]=lcopy(h); sv[3]=lcopy(lam);
      sv=gerepile(av,tetpil,sv);
      B=(GEN)sv[1]; h=(GEN)sv[2]; lam=(GEN)sv[3];
    }
  }
}

static GEN
lllgramall(GEN x, long all)
{
  long av=avma,tetpil,lx=lg(x),i,j,k,l,n,lim,kmax;
  GEN u,B,lam,q,r,h,la,bb,p1,p2,p3,p4,y,fl;
  GEN *gptr[8];

  if (typ(x) != t_MAT) err(lllger1);
  n=lx-1;
  if (!n)
  {
    if (all) return cgetg(1,t_MAT);

    y=cgetg(3,t_VEC); y[1]=lgetg(1,t_MAT); y[2]=lgetg(1,t_MAT);
    return y;
  }

  if (n==1)
  {
    if (gcmp0(gcoeff(x,1,1)))
    {
      if (!all)
      { y=cgetg(3,t_VEC); y[1]=(long)idmat(1); y[2]=lgetg(1,t_MAT); return y; }
      return (all==1)?idmat(1):cgetg(1,t_MAT);
    }
    else
    {
      if (!all)
      { y=cgetg(3,t_VEC); y[1]=lgetg(1,t_MAT); y[2]=(long)idmat(1); return y; }
      return (all==1)?cgetg(1,t_MAT):idmat(1);
    }
  }

  if (lg((GEN)x[1])!=lx) err(lllger2);
  av=avma; lim=(avma+bot)>>1; x=gcopy(x);
  B=cgetg(lx+1,t_COL); B[1]=un;
  fl=cgetg(lx,t_VEC); for (i=1; i<lx; i++) { B[i+1]=zero; fl[i]=zero; }
  lam=cgetg(lx,t_MAT);
  for (j=1; j<lx; j++)
  { p2=cgetg(lx,t_COL); lam[j]=(long)p2; for (i=1; i<lx; i++) p2[i]=zero; }
  k=2; h=idmat(n); kmax=1; 
  u=gcoeff(x,1,1); if (typ(u)!=t_INT) err(lllger4);
  if (signe(u)) { B[2]=(long)u; coeff(lam,1,1)=fl[1]=un; }
  else { B[2]=un; fl[1]=zero; }
  if (DEBUGLEVEL>5) { fprintferr("k = %ld",k); flusherr(); }
  for(;;)
  {
    if (k>kmax)
    {
      kmax=k;
      for (j=1; j<=k; j++)
      {
	if ((j==k)||signe(fl[j]))
	{
	  u=gcoeff(x,k,j); if (typ(u)!=t_INT) err(lllger4);
	  for (i=1; i<j; i++)
	    if (signe(fl[i]))
	      u=divii(subii(mulii((GEN)B[i+1],u),mulii(gcoeff(lam,k,i),gcoeff(lam,j,i))),(GEN)B[i]);
	  if (j<k) coeff(lam,k,j)=(long)u;
	  else
	  {
	    if (signe(u)) { B[k+1]=(long)u; coeff(lam,k,k)=fl[k]=un; }
	    else { B[k+1]=B[k]; fl[k]=zero; }
	  }
	}
      }
      if (avma<lim)
      {
	gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
	gptr[4]=&x; gerepilemany(av,gptr,5);
      }
    }
    if (cmpii(absi(u=shifti(gcoeff(lam,k,k-1),1)),(GEN)B[k])>0)
    {
      q=dvmdii(addii(u,(GEN)B[k]),shifti((GEN)B[k],1),&r);
      if (signe(r)<0) q=addsi(-1,q);
      r=negi(q);
      h[k]=ladd((GEN)h[k],gmul(r,(GEN)h[k-1]));
      for (j=1; j<=n; j++) coeff(x,j,k)=ladd(gcoeff(x,j,k),gmul(r,gcoeff(x,j,k-1)));
      for (j=1; j<=n; j++) coeff(x,k,j)=ladd(gcoeff(x,k,j),gmul(r,gcoeff(x,k-1,j)));
      coeff(lam,k,k-1)=lsubii(gcoeff(lam,k,k-1),mulii(q,(GEN)B[k]));
      for (i=1; i<k-1; i++)
	coeff(lam,k,i)=lsubii(gcoeff(lam,k,i),mulii(q,gcoeff(lam,k-1,i)));
    }
    if (avma<lim)
    {
      gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
      gptr[4]=&x; gerepilemany(av,gptr,5);
    }
    if (signe(fl[k-1])&&((cmpii(p1=mulsi(99,sqri((GEN)B[k])),p2=mulsi(100,addii(p3=mulii((GEN)B[k-1],(GEN)B[k+1]),p4=mulii(la=gcoeff(lam,k,k-1),gcoeff(lam,k,k-1)))))>0)||(!signe(fl[k]))))
    {
      if (DEBUGLEVEL>=4 && k==n)
	{ fprintferr(" (%ld)", expi(p1)-expi(p2)); flusherr(); }
      p1=(GEN)h[k-1]; h[k-1]=h[k]; h[k]=(long)p1;
      p1=(GEN)x[k-1]; x[k-1]=x[k]; x[k]=(long)p1;
      for (j=1; j<=n; j++)
      { 
	p1=gcoeff(x,k-1,j); coeff(x,k-1,j)=coeff(x,k,j);
	coeff(x,k,j)=(long)p1;
      }
      for (j=1; j<=k-2; j++)
      { 
	p1=gcoeff(lam,k-1,j); coeff(lam,k-1,j)=coeff(lam,k,j);
	coeff(lam,k,j)=(long)p1;
      }
      if (signe(fl[k]))
      {
	for (i=k+1; i<=kmax; i++)
	{
	  bb=gcoeff(lam,i,k);
	  coeff(lam,i,k)=ldivii(subii(mulii((GEN)B[k+1],gcoeff(lam,i,k-1)),mulii(la,bb)),(GEN)B[k]);
	  coeff(lam,i,k-1)=ldivii(addii(mulii(la,gcoeff(lam,i,k-1)),mulii((GEN)B[k-1],bb)),(GEN)B[k]);
	  if (avma<lim)
	  {
	    gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
	    gptr[4]=&x; gptr[5]=&la; gptr[6]=&p3; gptr[7]=&p4;
	    gerepilemany(av,gptr,8);
	  }
	}
	B[k]=ldivii(addii(p3,p4),(GEN)B[k]);
      }
      else
      {
	if (signe(la))
	{
	  p2=(GEN)B[k]; p1=divii(p4,p2);
	  for (i=k+1; i<=kmax; i++)
	    coeff(lam,i,k-1)=ldivii(mulii(la,gcoeff(lam,i,k-1)),p2);
	  for (j=k+1; j<kmax; j++)
	  {
	    for (i=j+1; i<=kmax; i++)
	      coeff(lam,i,j)=ldivii(mulii(p1,gcoeff(lam,i,j)),p2);
	    if (avma<lim)
	    {
	      gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
	      gptr[4]=&x; gptr[5]=&p1; gptr[6]=&p2;
	      gerepilemany(av,gptr,7);
	    }
	  }
	  B[k+1]=B[k]=(long)p1;
	  for (i=k+2; i<=lx; i++)
	    B[i]=ldivii(mulii(p1,(GEN)B[i]),p2);
	}
	else
	{
	  coeff(lam,k,k-1)=zero;
	  for (i=k+1; i<=kmax; i++)
	  { coeff(lam,i,k)=coeff(lam,i,k-1); coeff(lam,i,k-1)=zero; }
	  B[k]=B[k-1]; fl[k]=un; fl[k-1]=zero;
	}

	if (avma<lim)
	{
	  gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
	  gptr[4]=&x; gerepilemany(av,gptr,5);
	}
      }
      if (k>2) k--;
      if (DEBUGLEVEL>5) { fprintferr(" %ld",k); flusherr(); }
    }
    else
    {
      for (l=k-2; l>=1; l--)
      {
	if (cmpii(absi(u=shifti(gcoeff(lam,k,l),1)),(GEN)B[l+1])>0)
	{
	  q=dvmdii(addii(u,(GEN)B[l+1]),shifti((GEN)B[l+1],1),&r);
	  if (signe(r)<0) q=addsi(-1,q);
	  r=negi(q);
	  h[k]=ladd((GEN)h[k],gmul(r,(GEN)h[l]));
	  for (j=1; j<=n; j++)
	    coeff(x,j,k)=ladd(gcoeff(x,j,k),gmul(r,gcoeff(x,j,l)));
	  for (j=1; j<=n; j++)
	    coeff(x,k,j)=ladd(gcoeff(x,k,j),gmul(r,gcoeff(x,l,j)));
	  coeff(lam,k,l)=lsubii(gcoeff(lam,k,l),mulii(q,(GEN)B[l+1]));
	  for (i=1; i<l; i++)
	    coeff(lam,k,i)=lsubii(gcoeff(lam,k,i),mulii(q,gcoeff(lam,l,i)));
	}
	if (avma<lim)
	{
	  gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
	  gptr[4]=&x; gerepilemany(av,gptr,5);
	}
      }
      k++; if (DEBUGLEVEL>5) { fprintferr(" %ld",k); flusherr(); }
      if (k>n)
      {
	k=1; while (k<=n && !signe(fl[k])) k++;
	tetpil=avma;
	if (!all)
	{
	  y=cgetg(3,t_VEC); p2=cgetg(k,t_MAT);
	  for (i=1; i<k; i++) p2[i]=lcopy((GEN)h[i]);
	  y[1]=(long)p2; p2=cgetg(n-k+2,t_MAT); y[2]=(long)p2;
	  for (i=k; i<=n; i++) p2[i-k+1]=lcopy((GEN)h[i]);
	}
	else
	{
	  if (all==1)
	  { 
	    y=cgetg(k,t_MAT);
	    for (i=1; i<k; i++) y[i]=lcopy((GEN)h[i]);
	  }
	  else
	  { 
	    y=cgetg(n-k+2,t_MAT);
	    for (i=k; i<=n; i++) y[i-k+1]=lcopy((GEN)h[i]); 
	  }
	}
	if (DEBUGLEVEL>=4) { fprintferr("\n"); flusherr(); }
	return gerepile(av,tetpil,y);
      }
    }
    if (avma<lim)
    {
      gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
      gptr[4]=&x; gerepilemany(av,gptr,5);
    }
  }
}

/* la vraie matrice de gram a comme coefficient (i,j)
   x[i,j]*veccon[i]*veccon[j] */
static GEN
lllgramintwithcontent(GEN x, GEN veccon, long all)
{
  long av=avma,tetpil,lx=lg(x),i,j,k,l,n,lim,kmax;
  GEN u,u2,B,lam,q,r,h,la,bb,p1,p2,p3,p4,y,fl,corr,corr2,newcon;
  GEN *gptr[9];

  if (typ(x) != t_MAT) err(lllger1);
  n=lx-1;
  if (!n)
  {
    if (all) return cgetg(1,t_MAT);
    else
    { y=cgetg(3,t_VEC); y[1]=lgetg(1,t_MAT); y[2]=lgetg(1,t_MAT); return y; }
  }
  if (n==1)
  {
    if (gcmp0(gcoeff(x,1,1)))
    {
      if (!all)
      { y=cgetg(3,t_VEC); y[1]=(long)idmat(1); y[2]=lgetg(1,t_MAT); return y; }
      return (all==1)?idmat(1):cgetg(1,t_MAT);
    }
    else
    {
      if (!all)
      { y=cgetg(3,t_VEC); y[1]=lgetg(1,t_MAT); y[2]=(long)idmat(1); return y; }
      return (all==1)?cgetg(1,t_MAT):idmat(1);
    }
  }
  if (lg((GEN)x[1])!=lx) err(lllger2);
  av=avma; lim=(avma+bot)>>1;
  x=gcopy(x); veccon=gcopy(veccon);
  B=cgetg(lx+1,t_COL); B[1]=un;
/* B[i+1]=B_i; le vrai
   B_i est B_i*prod(1,j=1,i,veccon[j]^2) */
  fl=cgetg(lx,t_VEC); for (i=1; i<lx; i++){ B[i+1]=zero; fl[i]=zero; }
  lam=cgetg(lx,t_MAT);
  for (j=1; j<lx; j++)
  { p2=cgetg(lx,t_COL); lam[j]=(long)p2; for (i=1; i<lx; i++) p2[i]=zero; }
/* le vrai lam[i,j] est
   lam[i,j]*veccon[i]*veccon[j]*prod(1,l=1,j-1,veccon[l]^2) */
  k=2; h=idmat(n); kmax=1; 
  u=gcoeff(x,1,1); if (typ(u)!=t_INT) err(lllger4);
  if (signe(u)) { B[2]=(long)u; coeff(lam,1,1)=fl[1]=un; }
  else { B[2]=un; fl[1]=zero; }
  if (DEBUGLEVEL>5) { fprintferr("k = %ld",k); flusherr(); }
  for(;;)
  {
    if (k>kmax)
    {
      kmax=k;
      for (j=1; j<=k; j++)
      {
	if ((j==k)||signe(fl[j]))
	{
	  u=gcoeff(x,k,j); if (typ(u)!=t_INT) err(lllger4);
	  for (i=1; i<j; i++)
	    if (signe(fl[i]))
	      u=divii(subii(mulii((GEN)B[i+1],u),mulii(gcoeff(lam,k,i),gcoeff(lam,j,i))),(GEN)B[i]);
	  if (j<k) coeff(lam,k,j)=(long)u;
	  else
	  {
	    if (signe(u)) { B[k+1]=(long)u; coeff(lam,k,k)=fl[k]=un; }
	    else { B[k+1]=B[k]; fl[k]=zero; }
	  }
	}
      }
      if (avma<lim)
      {
	gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
	gptr[4]=&x; gptr[5]=&veccon; gerepilemany(av,gptr,6);
      }
    }
    u=shifti(mulii(gcoeff(lam,k,k-1),(GEN)veccon[k]),1);
    u2=mulii((GEN)B[k],(GEN)veccon[k-1]);
    if (cmpii(absi(u),u2)>0)
    {
      q=dvmdii(addii(u,u2),shifti(u2,1),&r);
      if (signe(r)<0) q=addsi(-1,q);
      h[k]=lsub((GEN)h[k],gmul(q,(GEN)h[k-1]));
      newcon=mppgcd((GEN)veccon[k],(GEN)veccon[k-1]);
      corr=divii((GEN)veccon[k],newcon); veccon[k]=(long)newcon;
      if(!gcmp1(corr))
      {
	corr2=sqri(corr);
	for (j=1; j<=n; j++)
	  coeff(x,j,k)=coeff(x,k,j)=lmulii(corr,gcoeff(x,k,j));
	coeff(x,k,k)=lmulii(corr,gcoeff(x,k,k));
	for (j=k; j<=kmax; j++) B[j+1]=lmulii(corr2,(GEN)B[j+1]);
	for (i=1; i<k; i++) coeff(lam,k,i)=lmulii(corr,gcoeff(lam,k,i));
	for (i=k+1; i<=kmax; i++)
	{
	  coeff(lam,i,k)=lmulii(corr,gcoeff(lam,i,k));
	  for (j=k+1; j<i; j++)
	    coeff(lam,i,j)=lmulii(corr2,gcoeff(lam,i,j));
	}
      }
      r=negi(mulii(q,divii((GEN)veccon[k-1],(GEN)veccon[k])));
      p1=gcoeff(x,k,k-1);
      for (j=1; j<=n; j++)
	coeff(x,j,k)=coeff(x,k,j)=laddii(gcoeff(x,j,k),mulii(r,(j==k)?p1:gcoeff(x,j,k-1)));
      coeff(x,k,k)=laddii(gcoeff(x,k,k),mulii(r,gcoeff(x,k-1,k)));
      coeff(lam,k,k-1)=laddii(gcoeff(lam,k,k-1),mulii(r,(GEN)B[k]));
      for (i=1; i<k-1; i++)
	coeff(lam,k,i)=laddii(gcoeff(lam,k,i),mulii(r,gcoeff(lam,k-1,i)));
    }
    if (avma<lim)
    {
      gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
      gptr[4]=&x; gptr[5]=&veccon; gerepilemany(av,gptr,6);
    }
    p3=mulii((GEN)B[k-1],(GEN)B[k+1]);la=gcoeff(lam,k,k-1);p4=mulii(la,la);
    p2=mulsi(100,mulii(mulii((GEN)veccon[k],(GEN)veccon[k]),addii(p3,p4)));
    p1=mulii((GEN)veccon[k-1],(GEN)B[k]);p1=mulsi(99,mulii(p1,p1));
    if (signe(fl[k-1])&&(cmpii(p1,p2)>0 || !signe(fl[k])))
    {
      if (DEBUGLEVEL>=4 && k==n)
	{ fprintferr(" (%ld)", expi(p1)-expi(p2)); flusherr(); }
      p1=(GEN)h[k-1]; h[k-1]=h[k]; h[k]=(long)p1;
      p1=(GEN)x[k-1]; x[k-1]=x[k]; x[k]=(long)p1;
      for (j=1; j<=n; j++)
      { p1=gcoeff(x,k-1,j); coeff(x,k-1,j)=coeff(x,k,j); coeff(x,k,j)=(long)p1; }
      p1=(GEN)veccon[k-1]; veccon[k-1]=veccon[k]; veccon[k]=(long)p1;
      for (j=1; j<=k-2; j++)
      { p1=gcoeff(lam,k-1,j); coeff(lam,k-1,j)=coeff(lam,k,j); coeff(lam,k,j)=(long)p1; }
      if (signe(fl[k]))
      {
	for (i=k+1; i<=kmax; i++)
	{
	  bb=gcoeff(lam,i,k);
	  coeff(lam,i,k)=ldivii(subii(mulii((GEN)B[k+1],gcoeff(lam,i,k-1)),mulii(la,bb)),(GEN)B[k]);
	  coeff(lam,i,k-1)=ldivii(addii(mulii(la,gcoeff(lam,i,k-1)),mulii((GEN)B[k-1],bb)),(GEN)B[k]);
	  if (avma<lim)
	  {
	    gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
	    gptr[4]=&x; gptr[5]=&la; gptr[6]=&veccon; gptr[7]=&p3;
	    gptr[8]=&p4; gerepilemany(av,gptr,9);
	  }
	}
	B[k]=ldivii(addii(p3,p4),(GEN)B[k]);
      }
      else
      {
	if (signe(la))
	{
	  p2=(GEN)B[k]; p1=divii(p4,p2);
	  for (i=k+1; i<=kmax; i++)
	    coeff(lam,i,k-1)=ldivii(mulii(la,gcoeff(lam,i,k-1)),p2);
	  for (j=k+1; j<kmax; j++)
	  {
	    for (i=j+1; i<=kmax; i++)
	      coeff(lam,i,j)=ldivii(mulii(p1,gcoeff(lam,i,j)),p2);
	    if (avma<lim)
	    {
	      gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
	      gptr[4]=&x; gptr[5]=&la; gptr[6]=&veccon; gptr[7]=&p1;
	      gptr[8]=&p2; gerepilemany(av,gptr,9);
	    }
	  }
	  B[k+1]=B[k]=(long)p1;
	  for (i=k+2; i<=lx; i++)
	    B[i]=ldivii(mulii(p1,(GEN)B[i]),p2);
	}
	else
	{
	  coeff(lam,k,k-1)=zero;
	  for (i=k+1; i<=kmax; i++)
	  { 
	    coeff(lam,i,k)=coeff(lam,i,k-1);
	    coeff(lam,i,k-1)=zero;
	  }
	  B[k]=B[k-1]; fl[k]=un; fl[k-1]=zero;
	}

	if (avma<lim)
	{
	  gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
	  gptr[4]=&x; gptr[5]=&veccon;
	  gerepilemany(av,gptr,6);
	}
      }
      if (k>2) k--;
      if (DEBUGLEVEL>5) { fprintferr(" %ld",k); flusherr(); }
    }
    else
    {
      for (l=k-2; l>=1; l--)
      {
	u=shifti(mulii(gcoeff(lam,k,l),(GEN)veccon[k]),1);
	u2=mulii((GEN)B[l+1],(GEN)veccon[l]);
	if (cmpii(absi(u),u2)>0)
	{
	  q=dvmdii(addii(u,u2),shifti(u2,1),&r);
	  if (signe(r)<0) q=addsi(-1,q);
	  h[k]=lsub((GEN)h[k],gmul(q,(GEN)h[l]));
	  newcon=mppgcd((GEN)veccon[k],(GEN)veccon[l]);
	  corr=divii((GEN)veccon[k],newcon); veccon[k]=(long)newcon;
	  if(!gcmp1(corr))
	  {
	    corr2=sqri(corr);
	    for (j=1; j<=n; j++)
	      coeff(x,j,k)=coeff(x,k,j)=lmulii(corr,gcoeff(x,k,j));
	    coeff(x,k,k)=lmulii(corr,gcoeff(x,k,k));
	    for (j=k; j<=kmax; j++) B[j+1]=lmulii(corr2,(GEN)B[j+1]);
	    for (i=1; i<k; i++) coeff(lam,k,i)=lmulii(corr,gcoeff(lam,k,i));
	    for (i=k+1; i<=kmax; i++)
	    {
	      coeff(lam,i,k)=lmulii(corr,gcoeff(lam,i,k));
	      for (j=k+1; j<i; j++)
		coeff(lam,i,j)=lmulii(corr2,gcoeff(lam,i,j));
	    }
	  }
	  r=negi(mulii(q,divii((GEN)veccon[l],(GEN)veccon[k])));
	  p1=gcoeff(x,k,l);
	  for (j=1; j<=n; j++)
	    coeff(x,j,k)=coeff(x,k,j)=laddii(gcoeff(x,j,k),mulii(r,(j==k)?p1:gcoeff(x,j,l)));
	  coeff(x,k,k)=laddii(gcoeff(x,k,k),mulii(r,gcoeff(x,l,k)));
	  coeff(lam,k,l)=laddii(gcoeff(lam,k,l),mulii(r,(GEN)B[l+1]));
	  for (i=1; i<l; i++)
	    coeff(lam,k,i)=laddii(gcoeff(lam,k,i),mulii(r,gcoeff(lam,l,i)));
	}
	if (avma<lim)
	{
	  gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
	  gptr[4]=&x; gptr[5]=&veccon; gerepilemany(av,gptr,6);
	}
      }
      k++; if (DEBUGLEVEL>5) { fprintferr(" %ld",k); flusherr(); }
      if (k>n)
      {
	k=1; while (k<=n && !signe(fl[k])) k++;
	tetpil=avma;
	if (!all)
	{
	  y=cgetg(3,t_VEC); p2=cgetg(k,t_MAT);
	  for (i=1; i<k; i++) p2[i]=lcopy((GEN)h[i]);
	  y[1]=(long)p2; p2=cgetg(n-k+2,t_MAT); y[2]=(long)p2;
	  for (i=k; i<=n; i++) p2[i-k+1]=lcopy((GEN)h[i]);
	}
	else
	{
	  if (all==1)
	  { 
	    y=cgetg(k,t_MAT); for (i=1; i<k; i++) y[i]=lcopy((GEN)h[i]);
	  }
	  else
	  { 
	    y=cgetg(n-k+2,t_MAT);
	    for (i=k; i<=n; i++) y[i-k+1]=lcopy((GEN)h[i]); 
	  }
	}
	if (DEBUGLEVEL>=4) { fprintferr("\n"); flusherr(); }
	return gerepile(av,tetpil,y);
      }
    }
    if (avma<lim)
    {
      gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
      gptr[4]=&x; gptr[5]=&veccon; gerepilemany(av,gptr,6);
    }
  }
}

static GEN
lllall0(GEN x, long all)
{
  long av=avma,tetpil,lx=lg(x),i,j,k,l,p1,n,lim,kmax;
  GEN u,B,lam,q,r,h,la,p2,p3,p4,y,fl;
  GEN *gptr[5];

  if (typ(x) != t_MAT) err(lllger1);
  n=lx-1;
  if (!n)
  {
    if (all) return cgetg(1,t_MAT);

    y=cgetg(3,t_VEC); y[1]=lgetg(1,t_MAT); y[2]=lgetg(1,t_MAT);
    return y;
  }
  if (n==1)
  {
    if (gcmp0((GEN)x[1]))
    {
      if (!all)
      { 
	y=cgetg(3,t_VEC); y[1]=(long)idmat(1);
	y[2]=lgetg(1,t_MAT); return y;
      }
      return (all==1)?idmat(1):cgetg(1,t_MAT);
    }
    else
    {
      if (!all)
      { 
	y=cgetg(3,t_VEC); y[1]=lgetg(1,t_MAT);
	y[2]=lcopy(x); return y;
      }
      return (all==1)?cgetg(1,t_MAT):gcopy(x);
    }
  }
  av=avma; x=gcopy(x); lim=(avma+bot)>>1;
  B=cgetg(lx+1,t_COL); for (i=1; i<=lx; i++) B[i]=zero;
  fl=cgetg(lx,t_VEC); for (i=1; i<lx; i++) fl[i]=zero;
  B[1]=un; lam=idmat(n);
  u=gscal((GEN)x[1],(GEN)x[1]);
  if (signe(u)) { B[2]=(long)u; coeff(lam,1,1)=fl[1]=un; }
  else { B[2]=un; coeff(lam,1,1)=fl[1]=zero; }
  k=2; h=idmat(n); kmax=1;
  for(;;)
  {
    if (k>kmax)
    {
      kmax=k;
      for (j=1; j<=k; j++)
      {
	if ((j<k)&&(!signe(fl[j]))) coeff(lam,k,j)=zero;
	else
	{
	  u=gscal((GEN)x[k],(GEN)x[j]);
	  if (typ(u)!=t_INT) err(lllger4);
	  for (i=1; i<j; i++)
	    if (signe(fl[i])) u=divii(subii(mulii((GEN)B[i+1],u),mulii(gcoeff(lam,k,i),gcoeff(lam,j,i))),(GEN)B[i]);
	  if (j<k) coeff(lam,k,j)=(long)u;
	  else
	  {
	    if (signe(u)) { B[k+1]=(long)u; coeff(lam,k,k)=fl[k]=un; }
	    else { B[k+1]=B[k]; coeff(lam,k,k)=fl[k]=zero; }
	  }
	 }
      }
    }
    if (signe(fl[k-1])&&(!signe(fl[k])))
    {
      if (cmpii(absi(u=shifti(gcoeff(lam,k,k-1),1)),(GEN)B[k])>0)
      {
	q=dvmdii(addii(u,(GEN)B[k]),shifti((GEN)B[k],1),&r);
	if (signe(r)<0) q=addsi(-1,q);
	h[k]=lsub((GEN)h[k],gmul(q,(GEN)h[k-1]));
	x[k]=lsub((GEN)x[k],gmul(q,(GEN)x[k-1]));
	coeff(lam,k,k-1)=lsubii(gcoeff(lam,k,k-1),mulii(q,(GEN)B[k]));
	for (i=1; i<k-1; i++) coeff(lam,k,i)=lsubii(gcoeff(lam,k,i),mulii(q,gcoeff(lam,k-1,i)));
      }
      la=gcoeff(lam,k,k-1);
      p3=mulii((GEN)B[k-1],(GEN)B[k+1]); p4=mulii(la,la);
      p1=h[k-1]; h[k-1]=h[k]; h[k]=p1;
      p1=x[k-1]; x[k-1]=x[k]; x[k]=p1;
      for (j=1; j<=k-2; j++)
      {
	p1=(long)coeff(lam,k-1,j); coeff(lam,k-1,j)=coeff(lam,k,j);
	coeff(lam,k,j)=p1;
      }
      if (signe(la))
      {
	p2=(GEN)B[k]; p1=ldivii(p4,p2);
	for (i=k+1; i<=kmax; i++)
	  coeff(lam,i,k-1)=ldivii(mulii(la,gcoeff(lam,i,k-1)),p2);
	for (j=k+1; j<kmax; j++)
	  for (i=j+1; i<=kmax; i++)
	    coeff(lam,i,j)=ldivii(mulii((GEN)p1,gcoeff(lam,i,j)),p2);
	B[k+1]=B[k]=p1;
	for (i=k+2; i<=kmax+1; i++)
	  B[i]=ldivii(mulii((GEN)p1,(GEN)B[i]),p2);
      }
      else
      {
	for (i=k+1; i<=kmax; i++)
	{ coeff(lam,i,k)=coeff(lam,i,k-1); coeff(lam,i,k-1)=zero; }
	B[k]=B[k-1]; fl[k]=un; fl[k-1]=zero;
      }
      if (k>2) k--;
    }
    else
    {
      for (l=k-1; l>=1; l--)
      {
	if (cmpii(absi(u=shifti(gcoeff(lam,k,l),1)),(GEN)B[l+1])>0)
	{
	  q=dvmdii(addii(u,(GEN)B[l+1]),shifti((GEN)B[l+1],1),&r);
	  if (signe(r)<0) q=addsi(-1,q);
	  h[k]=lsub((GEN)h[k],gmul(q,(GEN)h[l]));
	  x[k]=lsub((GEN)x[k],gmul(q,(GEN)x[l]));
	  coeff(lam,k,l)=lsubii(gcoeff(lam,k,l),mulii(q,(GEN)B[l+1]));
	  for (i=1; i<l; i++) coeff(lam,k,i)=lsubii(gcoeff(lam,k,i),mulii(q,gcoeff(lam,l,i)));
	 }
      }
      k++;
      if (k>n)
      {
	k=1; while (k<=n && !signe(fl[k])) k++;
	tetpil=avma;
	if (!all)
	{
	  y=cgetg(3,t_VEC); p2=cgetg(k,t_MAT);
	  for (i=1; i<k; i++) p2[i]=lcopy((GEN)h[i]);
	  y[1]=(long)p2; p2=cgetg(n-k+2,t_MAT); y[2]=(long)p2;
	  for (i=k; i<=n; i++) p2[i-k+1]=lcopy((GEN)h[i]);
	 }
	else
	{
	  if (all==1)
	  {
	    y=cgetg(k,t_MAT); for (i=1; i<k; i++) y[i]=lcopy((GEN)h[i]);
	  }
	  else
	  {
	    y=cgetg(n-k+2,t_MAT);
	    for (i=k; i<=n; i++) y[i-k+1]=lcopy((GEN)h[i]);
	  }
	 }
	return gerepile(av,tetpil,y);
      }
    }
    if (avma<lim)
    {
      gptr[0]=&B; gptr[1]=&lam; gptr[2]=&fl; gptr[3]=&h;
      gptr[4]=&x; gerepilemany(av,gptr,5);
    }
  }
}

GEN
lllgramint(GEN x)
{
  return lllgramall(x,2);
}

GEN
lllgramkerim(GEN x)
{
  return lllgramall(x,0);
}

static GEN
lllintpartialall(GEN mat, long all)
/*
 *  This routine is functionally similar to lllint when all = 0.
 *
 *    The input is an integer matrix mat (not necessarily square) whose
 *  columns are linearly independent.  It outputs another matrix T such that
 *  mat * T is partially reduced.  If all = 0, the size of mat * T is the
 *  same as the size of mat.  If all = 1 the number of columns of mat * T
 *  is at most equal to its number of rows.  A matrix M is said to be
 *  -partially reduced- if | m1 +- m2 | >= |m1| for any two distinct
 *  columns m1, m2, in M.
 *
 *    This routine is designed to quickly reduce lattices in which one row
 *  is huge compared to the other rows.  For example, when searching for a
 *  polynomial of degree 3 with root a mod p, the four input vectors might
 *  be the coefficients of
 *      X^3 - (a^3 mod p), X^2 - (a^2 mod p), X - (a mod p), p.
 *  All four constant coefficients are O(p) and the rest are O(1). By the
 *  pigeon-hole principle, the coefficients of the smallest vector in the
 *  lattice are O(p^(1/4)), Hence significant reduction of vector lengths
 *  can be anticipated.
 *
 *  		Peter Montgomery
 *              July, 1994
 */
{
  const long ncol = lg(mat)-1;
  const long ltop1 = avma;
  long ncolnz;
  GEN tmat1, tmat2, matmid, diag_prod;

  if (typ(mat) != t_MAT) err(lller1);

  if (ncol <= 1) return idmat(ncol);
  else
  {
    GEN dot11 = gscal((GEN)mat[1], (GEN)mat[1]);
    GEN dot12 = gscal((GEN)mat[1], (GEN)mat[2]);
    GEN dot22 = gscal((GEN)mat[2], (GEN)mat[2]);
    GEN tmat  = idmat(2); /* For first two columns only */

    int progress = 0;
    long npass2 = 0;

/*
 * Try to row reduce the first two columns of mat.
 * Our best result so far is (first two columns of mat)*tmat.
 *
 * Initially tmat = 2 x 2 identity matrix.
 * The inner products of the reduced matrix are in
 * dot11, dot12, dot22.
 */
    while (npass2 < 2 || progress)
    {
      GEN gtemp;
      GEN q = gdivround(dot12, dot22);

      npass2++;
      progress = !gcmp0(q);
      if (progress)
      {
      /*
       * Conceptually replace (v1, v2) by (v1 - q*v2, v2),
       * where v1 and v2 represent the reduced basis for the
       * first two columns of the matrix.
       *
       * We do this by updating tmat and the inner products.
       *
       * An improved algorithm would look only at the leading
       * digits of dot11, dot12, dot22.  It would use
       * single-precision calculations as much as possible.
       */
        GEN dot12new = subii(dot12, mulii(q, dot22));

        dot11 = subii(dot11, mulii(q, addii(dot12, dot12new)));
        dot12 = dot12new;
        tmat[1] = lsub((GEN)tmat[1], gmul(q, (GEN)tmat[2]));
      }

      /* Interchange the output vectors v1 and v2.  */
      gtemp = dot11; dot11 = dot22; dot22 = gtemp;
      gtemp = (GEN)tmat[1]; tmat[1] = tmat[2]; tmat[2] = (long)gtemp;

      /* Occasionally (including final pass) do garbage collection.  */
      if (npass2 % 8 == 0 || !progress)
      {
        GEN* gptrs[4];

        gptrs[0] = &dot11; gptrs[1] = &dot12;
        gptrs[2] = &dot22; gptrs[3] = &tmat;
        gerepilemany(ltop1, gptrs, 4);
      }
    } /* while npass2 < 2 || progress */

    {
      long icol;
      GEN det12 = gsub(gmul(dot11, dot22), gmul(dot12, dot12));

      tmat1 = idmat(ncol);
      matmid = cgetg(ncol+1, t_MAT);
      for (icol = 1; icol <= 2; icol++)
      {
        coeff(tmat1, 1, icol) = coeff(tmat, 1, icol);
	coeff(tmat1, 2, icol) = coeff(tmat, 2, icol);
        matmid[icol] = ladd(gmul((GEN)mat[1], gcoeff(tmat, 1, icol)),
                            gmul((GEN)mat[2], gcoeff(tmat, 2, icol)) );
      }
      for (icol = 3; icol <= ncol; icol++)
      {
        GEN curcol = (GEN)mat[icol];
	GEN dot1i = gscal((GEN)matmid[1], curcol);
        GEN dot2i = gscal((GEN)matmid[2], curcol);
      /*
       *  Try to solve
       *
       * ( dot11  dot12 ) (q1)    ( dot1i )
       * ( dot12  dot22 ) (q2)  = ( dot2i )
       *
       * Round -q1 and -q2 to the nearest integer.
       * Then compute curcol - q1*matmid[1] - q2*matmid[2].
       * This will be approximately orthogonal to the
       * first two vectors in the new basis.
       */
	GEN q1neg = gsub(gmul(dot12, dot2i), gmul(dot22, dot1i));
        GEN q2neg = gsub(gmul(dot12, dot1i), gmul(dot11, dot2i));

        q1neg = gdivround(q1neg, det12);
        q2neg = gdivround(q2neg, det12);
        coeff(tmat1, 1, icol) = ladd(gmul(q1neg, gcoeff(tmat, 1, 1)),
				     gmul(q2neg, gcoeff(tmat, 1, 2)));
        coeff(tmat1, 2, icol) = ladd(gmul(q1neg, gcoeff(tmat, 2, 1)),
				     gmul(q2neg, gcoeff(tmat, 2, 2)));
        matmid[icol] = ladd(curcol, gadd(gmul(q1neg, (GEN)matmid[1]),
                                         gmul(q2neg, (GEN)matmid[2])));
      } /* for icol */
    } /* local block */

  } /* else  ncol > 1*/
  if (DEBUGLEVEL>4)
  {
    fprintferr("tmat1 = "); outbeauterr(tmat1); flusherr();
    fprintferr("matmid = "); outbeauterr(matmid); flusherr();
  }
  {
     GEN* gptrs[2];
     gptrs[0] = &tmat1; gptrs[1] = &matmid;
	        /* Only tmat1 and matmid need to be saved for what follows. */
     gerepilemany(ltop1, gptrs, 2);
  }

  {
  /*
   * For each pair of column vectors v and w in matmid * tmat2,
   * try to replace (v, w) by (v, v - q*w) for some q.
   * We compute all inner products and check them repeatedly.
   */
    long npass = 0;
    const long ltop3 = avma; /* Excludes region with tmat1 and matmid */
    GEN dotprd = cgetg(ncol+1, t_MAT);
    long icol;
    long reductions = 0;
    long reductions_since_gerepile = 0;

    tmat2 = idmat(ncol);
    for (icol = 1; icol <= ncol; icol++)
    {
      long jcol;

      dotprd[icol] = lgetg(ncol+1,t_COL);
      for (jcol = 1; jcol <= icol; jcol++)
      {
	coeff(dotprd, icol, jcol) = (long)gscal((GEN)matmid[icol],
			                        (GEN)matmid[jcol]);
	coeff(dotprd, jcol, icol) = coeff(dotprd, icol, jcol);
      }	/* for jcol */
    }  /* for icol */

    while (reductions > 0 || npass == 0)
    {
      if (DEBUGLEVEL>4)
      {
	diag_prod = dbltor(1.0);

	for (icol = 1; icol <= ncol; icol++)
	  diag_prod = gmul(diag_prod, gcoeff(dotprd, icol, icol));
      }
      npass++;
      if (DEBUGLEVEL>4)
      {
	fprintferr("npass = %ld, reductions last time = %ld, diag_prod = ",
	            npass, reductions);
	outbeauterr(diag_prod); fprintferr("\n"); flusherr();
      }
      reductions = 0;

      for (icol = 1; icol <= ncol; icol++)
      {
        long ijdif;

        for (ijdif = 1; ijdif < ncol; ijdif++)
	{
          const long jcol = 1 + ((icol + ijdif - 1) % ncol);
          const long previous_avma = avma;
          const long k1 = (gcmp(gcoeff(dotprd, icol, icol),
		                gcoeff(dotprd, jcol, jcol) ) > 0)
				? icol : jcol; 	/* index of larger column */
	  const long k2 = icol + jcol - k1; 	/* index of smaller column */
	  GEN q;
	  GEN codi = gcoeff(dotprd, k2, k2);

	  q = gcmp0(codi) ? gzero : gdivround(gcoeff(dotprd, k1, k2), codi);

	  /* Try to subtract a multiple of column k2 from column k1.  */
	  if (gcmp0(q)) avma = previous_avma;
          else
	  {
	    long dcol;

	    reductions++;
            if (gcmp1(q))
	    {   /* Case q = +1 */
	      tmat2[k1] = lsub((GEN)tmat2[k1], (GEN)tmat2[k2]);
	      dotprd[k1] = lsub((GEN)dotprd[k1], (GEN)dotprd[k2]);
	      coeff(dotprd, k1, k1) = lsub(gcoeff(dotprd, k1, k1),
				           gcoeff(dotprd, k2, k1));
	    }
	    else if (gcmp_1(q))
	    { 	/* Case q = -1 */
	      tmat2[k1] = ladd((GEN)tmat2[k1], (GEN)tmat2[k2]);
	      dotprd[k1] = ladd((GEN)dotprd[k1], (GEN)dotprd[k2]);
	      coeff(dotprd, k1, k1) = ladd(gcoeff(dotprd, k1, k1),
				           gcoeff(dotprd, k2, k1));
            }
	    else
	    { 	/* General q   */
	      tmat2[k1] = lsub((GEN)tmat2[k1], gmul(q, (GEN)tmat2[k2]));
	      dotprd[k1] = lsub((GEN)dotprd[k1], gmul(q, (GEN)dotprd[k2]));
	      coeff(dotprd, k1, k1) = lsub(gcoeff(dotprd, k1, k1),
				         gmul(q, gcoeff(dotprd, k2, k1)));
            }

	    for (dcol = 1; dcol <= ncol; dcol++)
	      coeff(dotprd, k1, dcol) = coeff(dotprd, dcol, k1);
	    /* for dcol */
	  }	/* if q != 0 */
        } /* for ijdif */
        if (reductions_since_gerepile> ncol)
	{
	  GEN* gptrs[2];

	  reductions_since_gerepile = 0;
	  gptrs[0] = &dotprd; gptrs[1] = &tmat2;
	  gerepilemany(ltop3, gptrs, 2);
        }
      } /* for icol */
    } /* while */

   /*
    * Sort columns so smallest comes first in mat * tmat1 * tmat2.
    * Use insertion sort.
    */
    for (icol = 1; icol < ncol; icol++)
    {
      long jcol, smallest = icol;

      for (jcol = icol+1; jcol <= ncol; jcol++)
      {
	if (gcmp(gcoeff(dotprd,smallest,smallest),gcoeff(dotprd,jcol,jcol)) > 0)
	  smallest = jcol;
      }
      if (icol != smallest)
      { /* Exchange with proper column */
        /* Only diagonal of dotprd is updated */
	long ltemp;

        ltemp = tmat2[icol];
        tmat2[icol] = tmat2[smallest];
        tmat2[smallest] = ltemp;

        ltemp = coeff(dotprd, icol, icol);
        coeff(dotprd, icol, icol) = coeff(dotprd, smallest, smallest);
        coeff(dotprd, smallest, smallest) = ltemp;
      }
    } /* for icol */
    icol=1;
    while (icol <= ncol && gcmp0(gcoeff(dotprd, icol, icol))) icol++;
    ncolnz = ncol - icol + 1;

  } /* local block */

  {
    long lbot = avma;
    if (DEBUGLEVEL>4)
    {
      fprintferr("lllintpartial output = ");
      outbeauterr(gmul(matmid, tmat2));
      fprintferr("tmat1 = "); outbeauterr(tmat1);
      fprintferr("tmat2 = "); outbeauterr(tmat2);
    }
    if (all) return gerepile(ltop1, lbot, gmul(tmat1, tmat2));
    if (ncolnz == lg((GEN)mat[1])-1)
    {
      long icol;
      GEN tmat = cgetg(ncolnz+1,t_MAT);

      for (icol = 1; icol <= ncolnz; icol++)
	tmat[icol] = tmat2[icol + ncol - ncolnz];
      lbot = avma;
      return gerepile(ltop1, lbot, gmul(tmat1, tmat));
    }
    else
    {
      GEN tmat = gmul(mat, gmul(tmat1, tmat2));
      lbot = avma;
      return gerepile(ltop1, lbot, lllint(tmat));
    }
  }
}

GEN
lllintpartial(GEN mat)
{
  return lllintpartialall(mat,0);
}

/********************************************************************/
/**                                                                **/
/**                   LINEAR & ALGEBRAIC DEPENDANCE                **/
/**                                                                **/
/********************************************************************/
GEN
lindep0(GEN x,long bit,long prec)
{
  if (!bit) return lindep(x,prec);
  if (bit>0) return lindep2(x,bit);
  return deplin(x);
}

GEN
lindep2(GEN x, long bit)
{
  long tx=typ(x),lx=lg(x),ly,i,j,flag,av,tetpil,e;
  GEN y,p1,p2,p3,p4,p5;

  if (! is_vec_t(tx)) err(linder1);
  av=avma; p1=greal(x); p2=gimag(x); flag = !gcmp0(p2);
  ly = (flag) ? lx+2 : lx+1;
  p4=cgetg(lx,t_MAT); bit = (long) (bit/L2SL10);
  for (i=1; i<lx; i++)
  {
    p5=cgetg(ly,t_COL); p4[i]=(long)p5;
    for (j=1; j<lx; j++) p5[j]=(i==j) ? un : zero;
    p5[lx]=lcvtoi(gshift((GEN)p1[i],bit),&e);
    if (flag) p5[lx+1]=lcvtoi(gshift((GEN)p2[i],bit),&e);
  }
  p5=gmul(p4,lllint(p4)); p3=(GEN)p5[1];
  tetpil=avma; y=cgetg(lx,t_VEC);
  for (i=1; i<lx; i++) y[i]=lcopy((GEN)p3[i]);
  return gerepile(av,tetpil,y);
}

#define quazero(x) ( gcmp0(x) || ( typ(x)==t_REAL && expo(x) < EXP ) )

GEN
lindep(GEN x, long prec)
{
  GEN b[50],be[50],bn[50],m[50][50],c1,c2,c3,px,py,pxy;
  GEN p1,p2,p3,p4,p5,p6,p7,r,f,em;
  long qzer[50];
  long av1,tetpil,i,j,fl,i1;
  long av = avma, lx = lg(x), tx = typ(x), n = lx-1;
  const long EXP = - ((prec-2)<<TWOPOTBITS_IN_LONG) + 2*n;

  if (! is_vec_t(tx)) err(linder1);
  if (lx>=50) err(linder2);
  p1=greal(x); p2=gimag(x);
  px=gscal(p1,p1); py=gscal(p2,p2); pxy=gscal(p1,p2);
  p3=mpsub(mpmul(px,py),mpmul(pxy,pxy));
  for (i=1; i<=n; i++)
  {
    be[i]=cgetg(lx,t_COL);
    for (j=1; j<=n; j++) be[i][j]=lgetr(prec+1);
  }
  for (j=1; j<=n; j++) bn[j]=cgetr(prec+1);
  for (i=1; i<=n; i++)
  {
    for (j=1; j<i; j++)
      m[i][j]=cgetr(prec+1);
  }
  if (quazero(p1)) { p1=p2; px=py; fl=1; } else fl=quazero(p3);
  for (i=1; i<=n; i++)
  {
    b[i]=cgetg(lx,t_COL);
    for (j=1; j<=n; j++)
      b[i][j]=(i==j) ? un : zero;
  }
  av1=avma;
  for (i=1; i<=n; i++)
  {
    if (fl) p4=gmul(gdiv(gscal(b[i],p1),px),p1);
    else
    {
      p4=gscal(b[i],p1); p5=gscal(b[i],p2);
      p6=gdiv(mpsub(mpmul(py,p4),mpmul(pxy,p5)),p3);
      p7=gdiv(mpsub(mpmul(px,p5),mpmul(pxy,p4)),p3);
      p4=gadd(gmul(p6,p1),gmul(p7,p2));
    }
    if (tx==t_COL) p4=gsub(b[i],p4);
    else p4=gsub(b[i],gtrans(p4));
    for (j=1; j<i; j++)
      if (!qzer[j])
      {
        gdivz(gscal(b[i],be[j]),bn[j],m[i][j]);
        p4=gsub(p4,gmul(m[i][j],be[j]));
      }
      else affrr(bn[j],m[i][j]);
    gaffect(p4,be[i]);
    affrr(gscal(be[i],be[i]),bn[i]);
    qzer[i]=quazero(bn[i]); avma=av1;
  }
  while (qzer[n])
  {
    em=bn[1]; j=1; av1=avma;
    for (i=2; i<n; i++)
    {
      p3=shiftr(bn[i],i);
      if (cmprr(p3,em)>0)
      {
        em=p3; j=i;
      }
    }
    avma=av1;
    if (DEBUGLEVEL>9)
    {
      printf("\n");
      for (i1=1; i1<=n; i1++) 
	for (i=1; i<i1; i++) output(m[i1][i]);
    }
    i=j; i1=i+1;
    r=ground(m[i1][i]);
    f=subri(m[i1][i],r);
    p3=gsub(b[i1],gmul(r,b[i]));
    b[i1]=b[i]; b[i]=p3;
    for (j=1; j<i; j++)
    {
      if (!qzer[j])
      {
        p3=mpsub(m[i1][j],mulir(r,m[i][j]));
        affrr(m[i][j],m[i1][j]); mpaff(p3,m[i][j]);
      }
    }
    c1=addrr(bn[i1],mulrr(mulrr(f,f),bn[i]));
    fl=quazero(c1);
    if (!fl)
    {
      c2=divrr(mulrr(bn[i],f),c1); affrr(c2,m[i1][i]);
      c3=divrr(bn[i1],c1); mulrrz(c3,bn[i],bn[i1]);
      affrr(c1,bn[i]); qzer[i1]=quazero(bn[i1]); qzer[i]=0;
      for (j=i+2; j<=n; j++)
      {
        p3=addrr(mulrr(m[j][i1],c3),mulrr(m[j][i],c2));
        subrrz(m[j][i],mulrr(f,m[j][i1]),m[j][i1]);
        affrr(p3,m[j][i]);
      }
    }
    else
    {
      qzer[i1]=qzer[i]; affrr(bn[i],bn[i1]);
      affrr(c1,bn[i]); qzer[i]=1;
      for (j=i+2; j<=n; j++) affrr(m[j][i],m[j][i1]);
    }
  }
  p3=cgetg(lx,t_COL);
  for (i=1; i<=n; i++) p3[i]=(i==n) ? un : zero;
  p3[n]=un; p5=cgetg(lx,t_MAT);
  for (i=1; i<=n; i++) p5[i]=lcopy(b[i]);
  p4=gauss(gtrans(p5),p3); tetpil=avma;
  return gerepile(av,tetpil,gtrans(p4));
}

GEN
algdep0(GEN x, long n, long bit, long prec)
{
  long tx=typ(x),av,tetpil,i,j,k;
  GEN y,p1;

  if (! is_scalar_t(tx)) err(algder1);
  if (tx==t_POLMOD) { y=gcopy((GEN)x[1]); setvarn(y,0); return y; }
  if (gcmp0(x)) return gzero;

  av=avma; p1=cgetg(n+2,t_COL); p1[1]=un;
  for (i=2; i<=n+1; i++) p1[i]=lmul((GEN)p1[i-1],x);
  p1= bit ? lindep2(p1,bit) : lindep(p1,prec); ;
  tetpil=avma; y=cgetg(n+3,t_POL); setsigne(y,1); setvarn(y,0);
  j=0; k=1; while (gcmp0((GEN)p1[k])) k++;
  for (i=0; i<=n+1-k; i++)
  {
    y[i+2]=lcopy((GEN)p1[k+i]);
    if (!gcmp0((GEN)p1[k+i])) j=i;
  }
  setlgef(y,j+3);
  if (gsigne((GEN)y[j+2])>0) return gerepile(av,tetpil,y);

  tetpil=avma; return gerepile(av,tetpil,gneg(y));
}

GEN
algdep2(GEN x, long n, long bit)
{
  return algdep0(x,n,bit,0);
}

GEN
algdep(GEN x, long n, long prec)
{
  return algdep0(x,n,0,prec);
}

/********************************************************************/
/**                                                                **/
/**                       CHANGE OF VARIABLES                      **/
/**                                                                **/
/********************************************************************/
static int var_not_changed; /* can only be altered in reorder() */

/* Substitution globale des composantes du vecteur y aux variables de x */
GEN
changevar(GEN x, GEN y)
{
  long tx,ty,lx,vx,vy,i,av,tetpil;
  GEN  p1,p2,p3,z;

  if (var_not_changed && y==polvar) return x;
  tx=typ(x); ty=typ(y);
  if (! is_vec_t(ty)) err(changer1);
  if (is_scalar_t(tx))
  {
    if (tx==t_POLMOD)
    {
      av=avma; p1=changevar((GEN)x[1],y); p2=changevar((GEN)x[2],y);
      tetpil=avma;
      if (isonstack((GEN)x[1]))
        return gerepile(av,tetpil,gmodulcp(p1,p2));
      return gerepile(av,tetpil,gmodulo(p1,p2));
    }
    return gcopy(x);
  }

  if (is_rfrac_t(tx))
  {
    av=avma; p1=changevar((GEN)x[1],y); p2=changevar((GEN)x[2],y);
    tetpil=avma; return gerepile(av,tetpil,gdiv(p1,p2));
  }
  lx=lg(x);

  if (tx == t_POL || tx == t_SER)
  {
    vx=varn(x)+1; if (vx>=lg(y)) return gcopy(x);
    if (!signe(x))
    {
      vy=gvar((GEN)y[vx]); if (vy>MAXVARN) err(changer1);
      z=gcopy(x); setvarn(z,vy); return z;
    }
    av=avma; p1=(GEN)y[vx];

    if (tx==t_POL)
    {
      tetpil=avma; lx=lgef(x);
      p2=changevar((GEN)x[lx-1],y);
      for (i=lx-2; i>=2; i--)
      {
	p2=gmul(p2,p1); p3=changevar((GEN)x[i],y);
	tetpil=avma; p2=gadd(p2,p3);
      }
      return gerepile(av,tetpil,p2);
    }

    /* tx == t_SER */
    p2=changevar((GEN)x[lx-1],y);
    for (i=lx-2; i>=2; i--)
    {
      p2=gmul(p2,p1); p3=changevar((GEN)x[i],y);
      p2=gadd(p2,p3);
    }
    p3=ggrando(p1,lx-2); tetpil=avma;
    z=gadd(p2,p3);
    if (valp(x))
    {
      p2=gpuigs(p1,valp(x)); tetpil=avma;
      return gerepile(av,tetpil,gmul(p2,z));
    }
    return gerepile(av,tetpil,z);
  }
  z=cgetg(lx,tx);
  for (i=1; i<lx; i++) z[i]=lchangevar((GEN)x[i],y);
  return z;
}

static
int compare_int(int *a,int *b)
{
  return *a-*b;
}

GEN
reorder(GEN x)
{
  long tx,lx,i,n, nvar = manage_var(3,NULL);
  char t1[MAXVAR];
  int *var,*varsort;

  if (x == NULL) { var_not_changed=1; return gzero; }
  tx=typ(x); lx=lg(x)-1;

  var = (int *) xmalloc(lx*sizeof(int));
  varsort = (int *) xmalloc(lx*sizeof(int));

  if (! is_vec_t(tx)) err(reorder1);
  for (i=0; i<nvar; i++) t1[i] = 0;

  /* check if x is a permutation */
  for (n=0; n<lx; n++)
  {
    varsort[n] = var[n] = i = gvar((GEN)x[n+1]);
    if (i >= nvar) err(reorder2);
    if (t1[i]) err(reorder3);
    t1[i] = 1;
  }

  if (lx > 1)
  {
    qsort(varsort,lx,sizeof(int),(int(*)(ANYARG))compare_int);
    for (n=0; n<lx; n++)
    {
    /*
     * Recall that variables are numbered 0,1 etc...
     * while polvar starts at 1.
     */
      polvar[var[n]+1]=lpolx[varsort[n]];
      ordvar[varsort[n]]=var[n];
    }
    var_not_changed=1;
    for (i=0; i<nvar; i++)
      if (ordvar[i]!=i) { var_not_changed=0; break; }
  }
  free(var); free(varsort);
  return polvar;
}

/********************************************************************/
/**                                                                **/
/**                           HEAPSORT                             **/
/**                                                                **/
/********************************************************************/

GEN
vecsort0(GEN x, GEN k)
{
  return gcmp0(k) ? sort(x) : vecsort(x,k);
}

/* x= vector of elts. cmp = comparison function.
 * flag = 0 : usual sorting
 * flag = 1 : indirect sorting: return permutation that would sort x
 * flag = 2 : as 1, but return permutation as vector of longs (not GEN)
 */
static GEN
gen_sort(GEN x, int cmp(GEN,GEN), int flag)
{
  long av,tetpil,n,i,j,indxt,q,ir,l,tx=typ(x),lx=lg(x);
  GEN y,indx;

  if (!is_vec_t(tx)) err(sorter1);
  av=avma; n=lx-1;
  if (n<=1)
  {
    switch(flag)
    {
      case 0: y=gcopy(x); break;
      case 1: y=cgetg(lx,tx); y[1]=un; break;
      case 2: y=cgetg(lx,tx); y[1]=1; break;
    }
    return y;
  }
  indx=cgeti(lx); for (j=1; j<=n; j++) indx[j]=j;
  l=(n>>1)+1; ir=n;
  for(;;)
  {
    if (l>1) q=x[(indxt=indx[--l])];
    else
    {
      indxt = indx[ir]; q=x[indxt]; indx[ir]=indx[1];
      if (--ir == 1)
      {
        indx[1]=indxt; tetpil=avma; y=cgetg(lx,tx);
	switch(flag)
	{
	  case 0: for (i=1; i<=n; i++) y[i]=lcopy((GEN)x[indx[i]]); break;
	  case 1: for (i=1; i<=n; i++) y[i]=lstoi(indx[i]); break;
	  case 2: for (i=1; i<=n; i++) y[i]=indx[i]; break;
	}
        return gerepile(av,tetpil,y);
      }
    }
    i=l; j=l<<1;
    while (j<=ir)
    {
      if (j<ir && cmp((GEN)x[indx[j]],(GEN)x[indx[j+1]])<0) j++;
      if (cmp((GEN)q,(GEN)x[indx[j]]) < 0) { indx[i]=indx[j]; j+=(i=j); }
      else j=ir+1;
    }
    indx[i]=indxt;
  }
}

GEN
indexlexsort(GEN x)
{
  return gen_sort(x,lexcmp,1);
}

GEN
sindexsort(GEN x)
{
  return gen_sort(x,gcmp,2);
}

GEN
indexsort(GEN x)
{
  return gen_sort(x,gcmp,1);
}

GEN
sort(GEN x)
{
  return gen_sort(x,gcmp,0);
}

GEN
lexsort(GEN x)
{
  return gen_sort(x,lexcmp,0);
}

static long
gveccmp(GEN x,GEN y,GEN k)
{
  long lk=lg(k), i=1;

  while (i<lk)
  {
    long sgn = gcmp((GEN) x[k[i]], (GEN) y[k[i]]);
    if (sgn) return sgn;
    i++;
  }
  return 0;
}

GEN
vecsort(GEN x, GEN k)
{
  long av,tetpil,n,i,j,indxt,q,ir,l,t,tx=typ(x),lx=lg(x),tk=typ(k),lk;
  GEN y,indx,kk;

  if (! is_matvec_t(tx)) err(sorter1);
  av=avma; n=lx-1; if (n<=1) return gcopy(x);

  if (tk==t_INT) { y=cgetg(2,t_VEC); y[1]=(long)k; k=y; }
  else if (! is_vec_t(tk)) err(vecsorter1);
  q=0; lk=lg(k); kk=cgeti(lk);
  for (i=1; i<lk; i++)
  {
    l=itos((GEN)k[i]); if (l<=0) err(vecsorter2);
    q=max(q,l); kk[i]=l;
  }
  indx=cgeti(lx);
  for (j=1; j<=n; j++)
  {
    indx[j]=j; t=typ(x[j]);
    if (! is_vec_t(t)) err(sorter1);
    if (lg((GEN)x[j]) <= q) err(vecsorter2);
  }
  l=(n>>1)+1; ir=n;
  for(;;)
  {
    if (l>1) q=x[(indxt=indx[--l])];
    else
    {
      q=x[(indxt=indx[ir])]; indx[ir]=indx[1];
      if (--ir ==1)
      {
        indx[1]=indxt; tetpil=avma; y=cgetg(lx,tx);
        for (i=1; i<=n; i++) y[i]=lcopy((GEN)x[indx[i]]);
        return gerepile(av,tetpil,y);
      }
    }
    i=l; j=l<<1;
    while (j<=ir)
    {
      if (j<ir && gveccmp((GEN)x[indx[j]],(GEN)x[indx[j+1]],kk)<0) j++;
      if (gveccmp((GEN)q,(GEN)x[indx[j]],kk)<0)
      {
	indx[i]=indx[j]; j+=(i=j);
      }
      else j=ir+1;
    }
    indx[i]=indxt;
  }
}

/********************************************************************/
/**                                                                **/
/**                  POLYNOMIAL INTERPOLATION                      **/
/**                                                                **/
/********************************************************************/

GEN
polint(GEN xa, GEN ya, GEN x, GEN *dy)
{
  long av,tetpil,i,m,n;
  long ns=1, tx=typ(xa), ty=typ(ya), lx=lg(xa), ly=lg(ya);
  GEN den,dif,dift,ho,hp,w,y,c,d;
  GEN *gptr[2];

  if (! is_vec_t(tx) || ! is_vec_t(ty) || lx!=ly) err(polinter1);
  n=lx-1; if (n<=1) { y=gcopy((GEN)ya[1]); *dy=gcopy(y); return y; }
  av=avma; c=cgetg(ly,ty); d=cgetg(ly,ty);
  dif=gabs(gsub(x,(GEN)xa[1]),MEDDEFAULTPREC);
  for (i=1; i<=n; i++)
  {
    dift = gabs(gsub(x,(GEN)xa[i]), MEDDEFAULTPREC);
    if (gcmp(dift,dif)<0) { ns=i; dif=dift; }
    c[i]=ya[i]; d[i]=ya[i];
  }
  y=(GEN)ya[ns--];
  tetpil=avma;
  for (m=1; m<n; m++)
  {
    for (i=1; i<=n-m; i++)
    {
      ho=gsub((GEN)xa[i],x); hp=gsub((GEN)xa[i+m],x);
      w=gsub((GEN)c[i+1],(GEN)d[i]); den=gsub(ho,hp);
      if (gcmp0(den)) err(polinter2);
      den=gdiv(w,den); d[i]=lmul(hp,den); c[i]=lmul(ho,den);
    }
    *dy= (2*ns < (n-m))? (GEN)c[ns+1]: (GEN)d[ns--];
    tetpil=avma; y=gadd(y,*dy);
  }
  *dy=gcopy(*dy); gptr[0]=&y; gptr[1]=dy;
  gerepilemanysp(av,tetpil,gptr,2);
  return y;
}

/********************************************************************/
/**                                                                **/
/**                        POLRED & CO.                            **/
/**                                                                **/
/********************************************************************/

GEN
polsym(GEN x, long n)
{
  long av1,av2,dx=lgef(x)-3,i,k;
  GEN s,y;

  if (n<0) err(impl,"polsym of a negative n");
  if (typ(x) != t_POL || !signe(x)) err(poltyper);
  y=cgetg(n+2,t_COL); y[1]=lstoi(dx);
  for (k=1; k<=n; k++)
  {
    av1=avma; s = (dx>=k)? gmulsg(k,(GEN)x[dx+2-k]): gzero;
    for (i=1; i<k && i<=dx; i++)
      s=gadd(s,gmul((GEN)y[k-i+1],(GEN)x[dx+2-i]));
    if (!gcmp1((GEN)x[dx+2])) s=gdiv(s,(GEN)x[dx+2]);
    av2=avma; y[k+1]=lpile(av1,av2,gneg(s));
  }
  return y;
}

static GEN
pols_for_polred(GEN x, GEN base, GEN p1, GEN *pta)
{
  long i,j, v = varn(x), n = lgef(x)-3;
  GEN p2,p3,p4, y, a = cgetg(n+1,t_COL);

  for (i=1; i<=n; i++)
    a[i]=lmul(base,(GEN)p1[i]);

  y=cgetg(n+1,t_COL); /* no garbage is created beyond this point */
  for (i=1; i<=n; i++)
  {
    long av = avma, tetpil;

    if (DEBUGLEVEL>=3) { fprintferr("i = %ld\n",i); flusherr(); }
    p1=gmodulcp((GEN)a[i],x); p4=content((GEN)p1[2]);
    if (gcmp1(p4)) p1=caract(p1,v);
    else
    {
      p1=caract(gdiv(p1,p4),v);
      p1=gmul(gpuigs(p4,lgef(p1)-3),gsubst(p1,v,gdiv(polx[v],p4)));
    }
    if (DEBUGLEVEL>=4) { outerr(p1); flusherr(); }

    p2=ggcd(deriv(p1,v),p1);
    p3=(GEN)p2[lgef(p2)-1]; if (!gcmp1(p3)) p2=gdiv(p2,p3);

    tetpil=avma; p1=gerepile(av,tetpil,gdiv(p1,p2));
    y[i]=(long)p1; j=lgef(p1)-2;
    while (j>=2 && !signe(p1[j])) j-=2;
    if (j>=2 && signe(p1[j]) > 0)
    {
      while (j>=2) { setsigne(p1[j],-signe(p1[j])); j-=2; }
      if (pta) gnegz((GEN)a[i],(GEN)a[i]);
    }
  }
  if (pta) *pta=a;
  return y;
}

static GEN
allpolred(GEN x, GEN *pta, long code, long prec)
{
  GEN y,p1,p2,p3,p4,base,a;
  long n = lgef(x)-3, i,j,tetpil, av = avma;

  if (typ(x) != t_POL) err(poltyper);
  if (!signe(x)) return gcopy(x);
  if (!gcmp1((GEN)x[n+2])) err(poltyper);

  base=allbase4(x,code,&p2,NULL); /* p2 is junk */

  if (sturm(x)<n)
  {
    p2=roots(x,prec); p3=cgetg(n+1,t_MAT);
    for (i=1; i<=n; i++)
    {
      p1=cgetg(n+1,t_COL); p3[i]=(long)p1;
      for (j=1; j<=n; j++)
	p1[j] = (long) poleval((GEN)base[i],(GEN)p2[j]);
    }
    p2=greal(gmul(gconj(gtrans(p3)),p3));
    p1=lllgram(p2,prec);
  }
  else /* totally real */
  {
    GEN ptrace=cgetg(n+2,t_VEC);
    long k;
    
    ptrace[2]=lstoi(n);
    for (k=2; k<=n; k++)
    {
      long l = n-k+1;

      p1=gmulsg(k-1,(GEN)x[l+2]);
      for (i=k; i>2; i--)
	p1=gadd(p1,gmul((GEN)x[l+i],(GEN)ptrace[i]));
      ptrace[k+1]=lneg(p1);
    }
    p2=cgetg(n+1,t_MAT);
    for (i=1; i<=n; i++)
    {
      p1=cgetg(n+1,t_COL); p2[i]=(long)p1;
      for (j=1; j<i ; j++) p1[j]=lcopy(gcoeff(p2,i,j));
      for (   ; j<=n; j++)
      {
	p4=gres(gmul((GEN)base[i],(GEN)base[j]),x); p3=gzero;
	for (k=2; k<lgef(p4); k++)
	  p3=gadd(p3,gmul((GEN)p4[k],(GEN)ptrace[k]));
	p1[j]=(long)p3;
      }
    }
    p1=lllgramint(p2);
  }

  tetpil=avma;
  y = pols_for_polred(x,base,p1,pta);

  if (pta)
  {
    GEN *gptr[2];

    gptr[0]=pta; gptr[1]=&y;
    gerepilemanysp(av,tetpil,gptr,2);
    return y;
  }
  return gerepileupto(av,y);
}

GEN
polred0(GEN x,long flag, GEN p, long prec)
{
  GEN y;
  long small = (flag & 1)? 1:0; /* small */

  if (!gcmp0(p)) small=(long) p; /* factored polred */
  if (flag & 2) /* polred2 */
  {
    y=cgetg(3,t_MAT);
    y[2]=(long)allpolred(x,(GEN*)(y+1),small,prec);
    return y;
  }
  return allpolred(x,NULL,small,prec);
}

GEN
ordred(GEN x, long prec)
{
  GEN p1,p2,p3,base;
  long n=lgef(x)-3,i,j,av=avma,v = varn(x),tetpil;

  if (typ(x) != t_POL) err(poltyper);
  if (!signe(x)) return gcopy(x);
  if (!gcmp1((GEN)x[n+2])) err(poltyper);

  p2=roots(x,prec); p3=cgetg(n+1,t_MAT);

  base=cgetg(n+1,t_VEC); base[1]=(long)polun[v]; /* power basis */
  for (i=2; i<=n; i++)
    base[i]=lmul(polx[v],(GEN)base[i-1]);

  for (i=1; i<=n; i++)
  {
    p1=cgetg(n+1,t_COL); p3[i]=(long)p1;
    for (j=1; j<=n; j++)
      p1[j]=lpuigs((GEN)p2[j],i-1);
  }
  p2=greal(gmul(gconj(gtrans(p3)),p3));
  p1=lllgram(p2,prec);
  
  tetpil=avma;
  return gerepile(av,tetpil, pols_for_polred(x,base,p1,NULL));
}

static int
checkgenerator(GEN nf, GEN x, GEN *ptp)
{
  long v = varn((GEN)nf[1]);
  GEN p = caradj(gmodulcp(gmul((GEN)nf[7],x),(GEN)nf[1]), v, 0);

  if (lgef(ggcd(p,deriv(p,v))) == 3) { *ptp = p; return 1; }
  return 0;
}

GEN
polredabsfast(GEN x, long prec)
{
  long av=avma,tetpil,j,nv;
  GEN nf,t2,v,mres,w,per,charpol = NULL;

  if (typ(x) != t_POL) err(talker,"not a polynomial in polredabs");
  if (lgef(x)==4) return gcopy(polx[varn(x)]);

  nf=initalgall0(x,nf_REDUCE|nf_SMALL,prec);

  t2=gmael(nf,5,3); nv=min(2,lg(t2)-1);
  v = minim(t2, gceil(gcoeff(t2,nv,nv)), stoi(10000));
  mres=(GEN)v[3]; nv=lg(mres);
  if (nv>=10000)
  { 
    fprintferr("warning: not enough storage in minimgenerator\n");
    flusherr();
  }
  w=cgetg(nv,t_VEC);
  for (j=1; j<nv; j++) 
    w[j] = (long) qfeval(t2,(GEN)mres[j]);
  per=sindexsort(w);

  for (j=1; j<nv; j++)
    if (checkgenerator(nf,(GEN)mres[per[j]],&charpol)) break;

  x = gcopy(charpol? charpol: (GEN)nf[1]);
  tetpil=avma; return gerepile(av,tetpil,x);
}

/* A VIRER. Seules differences avec allporedabs:
 *  - utilise minim au lieu de fincke-pohst.
 *  - n'utilise pas findmindisc.
 */
GEN
polredabsall(GEN x, long flag, long prec)
{
  long av=avma,tetpil,j,nv,c,i;
  GEN p1,p2,nf,t2,v,mres,w,per,charpol,eps,wlim,y,ypro,alphapro,phimax;

  switch(flag)
  {
    case 0:
      nf=initalgall0(x,nf_SMALL|nf_REDUCE|nf_PARTIAL,prec);
      phimax=NULL; break;
    case 1:
      p1=initalgall0(x,nf_SMALL|nf_REDUCE|nf_PARTIAL|nf_ORIG,prec);
      nf=(GEN)p1[1]; phimax=(GEN)p1[2]; break;
    case 2:
      nf=initalgall0(x,nf_SMALL|nf_REGULAR,prec); phimax=NULL; break;
    case 3:
      nf=initalgall0(x,nf_SMALL|nf_REGULAR,prec); phimax=gcopy(x); break;
    default: err(talker,"incorrect flag in polredabsall");
  }

  t2=gmael(nf,5,3); nv=min(2,lg(t2)-1);
  v=minim(t2,gceil(gcoeff(t2,nv,nv)),stoi(10000));
  mres=(GEN)v[3]; nv=lg(mres);
  if (nv>=10000)
  {
    fprintferr("warning: not enough storage in minimgenerator\n");
    flusherr();
  }
  w=cgetg(nv,t_VEC);
  for (j=1; j<nv; j++)
    w[j] = (long) qfeval(t2,(GEN)mres[j]);
  per=sindexsort(w);

  alphapro=cgetg(nv,t_VEC); ypro=cgetg(nv,t_VEC); eps=dbltor(0.000001);
  c=0;
  for (j=1; j<nv; j++)
  {
    long jj=per[j];

    if (checkgenerator(nf,(GEN)mres[jj],&charpol))
    {
      if (!c)
	wlim=gadd((GEN)w[jj],eps);
      else
	if (gcmp((GEN)w[jj],wlim)>=0)
	{
	  tetpil=avma;
	  if (phimax)
	  {
	    y=cgetg(3,t_MAT);
	    p1=cgetg(c+1,t_COL); y[1]=(long)p1;
	    p2=cgetg(c+1,t_COL); y[2]=(long)p2;
	    for (i=1; i<=c; i++)
	    { 
	      p1[i]=lcopy((GEN)ypro[i]);
	      p2[i]=lcopy((GEN)alphapro[i]);
	    }
	  }
	  else
	  {
	    y=cgetg(c+1,t_VEC);
	    for (i=1; i<=c; i++) y[i]=lcopy((GEN)ypro[i]);
	  }
	  return gerepile(av,tetpil,y);
	}
      c++; ypro[c]=(long)charpol; alphapro[c]=mres[jj];
    }
  }
  tetpil=avma;
  if (c)
  {
    y=cgetg(c+1,t_VEC);
    for (i=1; i<=c; i++) y[i]=lcopy((GEN)ypro[i]);
  }
  else y=gcopy((GEN)nf[1]);
  return gerepile(av,tetpil,y);
}

/* no garbage collecting. It will be done in allpolredabs */
static GEN
findmindisc(GEN nf, GEN y, GEN alpha, long c, GEN phimax)
{
  long i,ind;
  GEN dmin,z,beta, st = cgetg(c+1,t_VEC);
  
  dmin=gabs(discsr((GEN)y[1]),DEFAULTPREC);
  st[1]=(long)dmin; ind=1;
  for (i=2; i<=c; i++)
  {
    z=gabs(discsr((GEN)y[i]),DEFAULTPREC); st[i]=(long)z;
    if (gcmp(z,dmin) < 0) { dmin=z; ind=i; }
  }
  z=(GEN)y[ind];
  for (i=ind+1; i<=c; i++)
    if (gegal((GEN)st[i],dmin) && gpolcomp((GEN)y[i],z) < 0)
      { z=(GEN)y[i]; ind=i; }

  i=lgef(z)-2; while (i>=2 && !signe(z[i])) i-=2;
  beta=(GEN)alpha[ind];
  if (i>=2 && signe(z[i])>0)
  {
    for (; i>=2; i-=2) setsigne(z[i],-signe(z[i]));
    if (phimax) beta=gneg(beta);
  }

  if (!phimax) y=gcopy(z);
  else
  {
    GEN p1;

    if (typ(phimax)==t_POLMOD) phimax=(GEN)phimax[2];
    p1=polymodrecip(gmodulcp(gmul((GEN)nf[7],beta), (GEN)nf[1]));

    y=cgetg(3,t_VEC); y[1]=lcopy(z);
    y[2]=lsubst(phimax,varn(phimax),p1);
  }
  return y;
}

GEN
allpolredabs(GEN x, long flun, long prec)
{
  long av=avma,j,nv,c;
  GEN p1,nf,t2,v,mres,w,per, charpol = NULL, eps,wlim,ypro,alphapro,phimax;

  switch(flun)
  {
    case 0:
      nf=initalgall0(x,nf_REDUCE|nf_SMALL|nf_PARTIAL,prec);
      phimax=NULL; break;
    case 1:
      p1=initalgall0(x,nf_REDUCE|nf_SMALL|nf_PARTIAL|nf_ORIG,prec);
      nf=(GEN)p1[1]; phimax=(GEN)p1[2]; break;
    case 2:
      nf=initalgall0(x,nf_SMALL|nf_REGULAR,prec); phimax=NULL; break;
    case 3:
      nf=initalgall0(x,nf_SMALL|nf_REGULAR,prec); phimax=gcopy(x); break;
    default: err(talker,"incorrect flag in allpolredabs");
  }

  t2=gmael(nf,5,3); nv=min(lg(t2)-1,2);
  if (DEBUGLEVEL)
  { 
    fprintferr("recherche T2-norme <= ");
    outerr(gcoeff(t2,nv,nv)); flusherr();
  }
  v=fincke_pohst(t2,gceil(gcoeff(t2,nv,nv)),stoi(10000),prec);
  mres=(GEN)v[3]; nv=lg(mres);
  if (nv>=10000)
  { 
    fprintferr("warning: not enough storage in minimgenerator\n");
    flusherr();
  }
  else if (DEBUGLEVEL) 
  { 
    fprintferr("%ld minimaux trouves. j = :\n",nv);
    flusherr();
  }
  w=cgetg(nv,t_VEC);
  for (j=1; j<nv; j++)
    w[j] = (long) qfeval(t2,(GEN)mres[j]);;
  per=sindexsort(w);

  alphapro=cgetg(nv,t_VEC); ypro=cgetg(nv,t_VEC); eps=dbltor(0.000001);
  c=0;
  for (j=1; j<nv; j++)
  {
    long jj=per[j];

    if (DEBUGLEVEL) { fprintferr(" %ld",j); flusherr(); }
    if (checkgenerator(nf,(GEN)mres[jj],&charpol))
    {
      if (!c)
	wlim=gadd((GEN)w[jj],eps);
      else
      {
	if (gcmp((GEN)w[jj],wlim) >= 0)
	  return gerepileupto(av,findmindisc(nf,ypro,alphapro,c,phimax));
      }
      c++; ypro[c]=(long)charpol; alphapro[c]=mres[jj];
    }
  }
  if (DEBUGLEVEL) fprintferr("\n");
  if (!c)
  {
    c++; x=(GEN)nf[1]; ypro[c]=lcopy(x);
    alphapro[c]=(long)polx[varn(x)];
  }
  return gerepileupto(av, findmindisc(nf,ypro,alphapro,c,phimax));
}

GEN
polredabsall0(GEN x, long flag, long prec)
{
  if (flag<4) return allpolredabs(x,flag,prec);
  if (flag !=8) return polredabsall(x,flag-4,prec);
  return polredabsfast(x,prec);
}

GEN
polredabs(GEN x, long prec)
{
  return allpolredabs(x,0,prec);
}

GEN
polredabs2(GEN x, long prec)
{
  return allpolredabs(x,1,prec);
}

GEN
polredabsnored(GEN x, long prec)
{
  return allpolredabs(x,2,prec);
}

GEN
polred(GEN x, long prec)
{
  return allpolred(x,(GEN*)0,0,prec);
}

GEN
smallpolred(GEN x, long prec)
{
  return allpolred(x,(GEN*)0,1,prec);
}

GEN
factoredpolred(GEN x, GEN p, long prec)
{
  return allpolred(x,(GEN*)0,(long)p,prec);
}

GEN
polred2(GEN x, long prec)
{
  GEN y=cgetg(3,t_MAT);

  y[2]= (long) allpolred(x,(GEN*)(y+1),0,prec);
  return y;
}

GEN
smallpolred2(GEN x, long prec)
{
  GEN y=cgetg(3,t_MAT);

  y[2]= (long) allpolred(x,(GEN*)(y+1),1,prec);
  return y;
}

GEN
factoredpolred2(GEN x, GEN p, long prec)
{
  GEN y=cgetg(3,t_MAT);

  y[2]= (long) allpolred(x,(GEN*)(y+1),(long)p,prec);
  return y;
}

/********************************************************************/
/**                                                                **/
/**                              MINIM                             **/
/**                                                                **/
/********************************************************************/

/* Minimal vectors for the integral definite quadratic form: a.
 * Result u:
 *   u[1]= Number of vectors of square norm <= BORNE
 *   u[2]= maximum norm found
 *   u[3]= list of vectors found (at most STOCKMAX)
 *
 *  If BORNE = gzero: Minimal non-zero vectors.
 *  status = 0, as above
 *  status = 1, exits when first suitable vector is found.
 *  status = 2, only compute rank of the family of v.v~ (v min.) incrementally.
 */
static GEN
minim00(GEN a, GEN BORNE, GEN STOCKMAX, long status)
{
  GEN u,r,unr,S,S1,borne,norme1,normax1;
  GEN *gptr[2];
  long tetpil,av1,av=avma,look_for_minvec,maxrank,found=0;
  long i,j,k,s,stockmax,*x, n = lg(a);
  double p,b,c,borne1;
  double **q,*v,*y,*z;
  const double eps=0.000001;

  x = (long*) xmalloc(n*sizeof(long));
  y = (double*) xmalloc(n*sizeof(double));
  z = (double*) xmalloc(n*sizeof(double));
  v = (double*) xmalloc(n*sizeof(double));
  q = (double**) xmalloc(n*sizeof(double*));
  for (j=1; j<n; j++)
    q[j] = (double*) xmalloc(n*sizeof(double));

  n--;

  u=lllgram(a,BIGDEFAULTPREC);
  a=gmul(gtrans(u),gmul(a,u));
  affsr(1,unr=cgetr(10));
  a=gmul(a,unr); r=sqred1(a);
  if (DEBUGLEVEL>1) { fprintferr("minim: r = "); outerr(r); flusherr(); }
  for (j=1; j<=n; j++)
  {
    v[j]=rtodbl(gcoeff(r,j,j));
    for (i=1; i<j; i++)
      q[i][j]=rtodbl(gcoeff(r,i,j));
  }

  look_for_minvec = gcmp0(BORNE);
  if (look_for_minvec)
  {
    b=rtodbl(gcoeff(a,1,1));
    for (i=2; i<=n; i++)
      if ((c=rtodbl(gcoeff(a,i,i))) < b) b=c;
    borne=ground(dbltor(b+eps)); borne1=b+eps;
  } else {
    borne=BORNE; normax1=gzero;
    borne1=gtodouble(BORNE)+eps;
  }
  if (status == 0)
  { 
    stockmax=itos(STOCKMAX);
    S=cgetg(stockmax+1,t_MAT);
  }
  else if (status == 2)
  {
    maxrank = (n*(n+1))>>1;
    r = cgetg(1+(maxrank<<1),t_MAT);
  }

  s=0; av1=avma;
  k=n; y[n]=z[n]=0;
  x[n]=(long) sqrt(borne1/v[n]+eps);
  for(;;)
  {
    do
    {
      if (k>1)
      {
	k--; z[k]=0;
	for (j=k+1; j<=n; j++) z[k] += q[k][j]*x[j];
	p=x[k+1]+z[k+1];
	y[k]=y[k+1] + p*p*v[k+1];
	x[k]=(long) floor(sqrt((borne1-y[k]+eps)/v[k])-z[k]);
      }
      for(;;)
      {
	p=x[k]+z[k];
	if (y[k] + p*p*v[k] <= borne1+eps) break;
	k++; x[k]--;
      }
    }
    while (k>1);
    if (! x[1] && y[1]<=eps) break;

    p=x[1]+z[1]; tetpil=avma; norme1=ground(dbltor(y[1]+ p*p*v[1]));
    if (look_for_minvec) 
    {
      if (gcmp(norme1,borne) < 0)
      {
	borne1=gtodouble(norme1); s=0;
	borne=gerepile(av1,tetpil,norme1);
      }
    } else if (gcmp(norme1,normax1) > 0) normax1=norme1;

    if (status == 0)
    {
      s++;
      if (s<=stockmax)
      {
	S[s]=lgetg(n+1,t_COL);
	for (i=1; i<=n; i++)
	  coeff(S,i,s)=lstoi(x[i]);
      }
    }
    else if (status == 1)
    {
      if (!look_for_minvec && gcmp(norme1,borne)<=0)
      {
	r=cgetg(n+1,t_COL); for (k=1; k<=n; k++) r[k]=lstoi(x[k]);
	tetpil=avma; r=gerepile(av,tetpil,gmul(u,r));
	u=cgetg(3,t_VEC); u[1]=(long)norme1; u[2]=(long)r;
	found=1;
	break;
      }
    }
    else
    {
      GEN g = cgetg(1+maxrank,t_VEC);
      long I=1;

      s++;
      for (i=1; i<=n; i++)
	for (j=i; j<=n; j++,I++)
	  g[I]=lmulss(x[i],x[j]);
      r[s]=(long)g;
      
      if (s >= maxrank<<1)
      {
	g=image(r); s=lg(g)-1;
	if (s==maxrank) break;
	s++;
	for (i=1;i<=s;i++) r[i]=g[i];
      }
    }

    x[k]--;
  }

  for (j=1; j<n; j++) free(q[j]);
  free(x); free(y); free(z); free(v); free(q);
  switch(status)
  {
    case 1:
      if (!found) return cgetg(1,t_VEC);
      return u;
    case 2: 
      if (s!=maxrank) { setlg(r,s+1); s=rank(r); }
      avma=av; return stoi(s);
  }

  if (stockmax)
  {
    k=min(s,stockmax); S1=cgetg(k+1,t_MAT);
    for (j=1; j<=k; j++)
      S1[j]=lmul(u,(GEN)S[j]);
    tetpil=avma; S=gcopy(S1);
  }
  else
  {
    tetpil=avma; S=cgetg(1,t_MAT);
  }
  
  r = look_for_minvec? norme1: normax1;
  r=gcopy(r); gptr[0]=&r; gptr[1]=&S;
  gerepilemanysp(av,tetpil,gptr,2);
  u=cgetg(4,t_VEC);
  u[1]=lstoi(s<<1);
  u[2]=(long)r;
  u[3]=(long)S;
  return u;
}

GEN
minim0(GEN a, GEN borne, GEN stockmax, long flag, long prec)
{
  switch(flag)
  {
    case 0: return minim00(a,borne,stockmax,0);
    case 1: return minim00(a,borne,gzero,1);
    case 2: return fincke_pohst(a,borne,stockmax,prec);
    default: err(flagerr);
  }
  return NULL; /* not reached */
}

GEN
minim(GEN a, GEN borne, GEN stockmax)
{
  return minim00(a,borne,stockmax,0);
}

GEN
minim2(GEN a, GEN borne, GEN stockmax)
{
  return minim00(a,borne,stockmax,1);
}

GEN
perf(GEN a)
{
  return minim00(a,gzero,gzero,2);
}

/* programme general pour les formes quadratiques definies positives
 * quelconques. On doit avoir BORNE != 0; on n'entre dans cette fonction qu'a
 * partir de fincke_pohst (la reduction LLL n'est donc pas faite ici)
 */
static GEN
smallvectors(GEN a, GEN BORNE, GEN STOCKMAX,long prec)
{
  long av,tetpil,av2,lim,n1,n,i,j,k,s,stockmax,pr,fl,epsbit;
  GEN unr,u,r,S,S1,x,y,z,v,q,norme1,normax1,p,borne1,eps,p1;

  if (gcmp0(BORNE)) err(talker,"bound equal 0 in smallvectors");
  n1=lg(a); n=n1-1; stockmax=itos(STOCKMAX);

  av=avma; lim=(bot+avma)>>1; r=sqred1(a);
  if (DEBUGLEVEL>5) { fprintferr("r = "); outerr(r); flusherr(); }
  pr=gprecision(r);
  if (!pr) /* ne devrait pas arriver */
  { 
    unr=cgetr(prec); affsr(1,unr); r=gmul(r,unr);
  } 
  else prec=pr;
  epsbit=((prec-2)<<(TWOPOTBITS_IN_LONG-1));
  eps=cgetr(prec); affsr(1,eps); eps=gshift(eps,-epsbit);

  v=cgetg(n1,t_VEC);
  q=cgetg(n1,t_MAT);
  for (j=1; j<=n; j++)
  { 
    v[j]=coeff(r,j,j);
    q[j]=lgetg(n1,t_COL);
    for (i=1; i<j; i++)  coeff(q,i,j)=coeff(r,i,j);
    for (   ; i<=n; i++) coeff(q,i,j)=zero;
  }
  borne1=gadd(BORNE,eps); normax1=gzero;

  av2=avma;
  x=cgetg(n1,t_VEC);
  y=cgetg(n1,t_VEC);
  z=cgetg(n1,t_VEC);
  if (stockmax)
  { 
    S=cgetg(stockmax+1,t_MAT);
    for (i=1; i<=stockmax; i++) S[i]=zero;
  }
  s=0; k=n;
  for (i=1; i<=n; i++) y[i]=z[i]=zero;
  for (i=1; i<n; i++) x[i]=zero;
  x[n]=lfloor(gadd(eps,gsqrt(gdiv(BORNE,(GEN)v[n]),prec)));
  if (DEBUGLEVEL>3)
  { 
    fprintferr("\nx[%ld] = ",n); outerr((GEN)x[n]); flusherr();
  }

  for(;;)
  {
    do
    {
      fl=0;
      if (k>1)
      {
	k--; z[k]=zero;
	for (j=k+1; j<=n; j++)
	  z[k] = ladd((GEN)z[k], gmul(gcoeff(q,k,j),(GEN)x[j]));
	p=gadd((GEN)x[k+1],(GEN)z[k+1]);
	y[k] = ladd((GEN)y[k+1],gmul(gsqr(p),(GEN)v[k+1]));
	p1=gsub(BORNE,(GEN)y[k]);
	if (gsigne(p1)<0)
	{
	  if (gexpo(p1)<-epsbit) p1=gzero; else fl=1;
	}
	if (!fl)
	  x[k]=lfloor(gadd(eps,
			   gsub(gsqrt(gdiv(p1,(GEN)v[k]),prec),
			        (GEN)z[k])));
      }
      for(;;)
      { 
	if (!fl)
	{
	  p1=gmul((GEN)v[k],gsqr(gadd((GEN)x[k],(GEN)z[k])));
	  if (gcmp(gsub(p1,gadd(gsub(BORNE,(GEN)y[k]),eps)),
	           gmul(eps,gabs(p1,prec))) <= 0) break;
	}
	k++; x[k]=laddis((GEN)x[k],-1); fl=0;
      }
      if (DEBUGLEVEL>3)
      {
	if (DEBUGLEVEL>5) fprintferr("%ld ",k);
	if (k==n) { fprintferr("\nx[%ld] = ",n); outerr((GEN)x[n]); }
	flusherr();
      }
      if (avma<lim)
      {
	GEN *gptr[5];

	if (DEBUGLEVEL>2)
	  { fprintferr("   gestion de pile...\n"); flusherr(); }
	gptr[0]=&x; gptr[1]=&y; gptr[2]=&z; gptr[3]=&normax1;
	if (stockmax) gptr[4]=&S;
	gerepilemany(av2,gptr,stockmax?5:4);
      }
    }
    while (k>1);
    if (gcmp0((GEN)x[1]) && gcmp((GEN)y[1],eps) <= 0) break;

    p1=gsqr(gadd((GEN)x[1],(GEN)z[1]));
    norme1=gadd((GEN)y[1],gmul((GEN)v[1],p1));
    if(gcmp(norme1,borne1)<=0) /* ce n'est pas bien, mais quoi faire ? */
    {
      if (gcmp(norme1,normax1)>0) normax1=norme1;
      s++;
      if (s<=stockmax)
      { 
	S[s]=lgetg(n+1,t_COL);
	for (i=1; i<=n; i++) coeff(S,i,s)=x[i];
      }
    }
    x[k]=laddis((GEN)x[k],-1);
  }

  if (stockmax)
  {
    k=min(s,stockmax); S1=cgetg(k+1,t_MAT);
    for (j=1; j<=k; j++) S1[j]=S[j];
    tetpil=avma; S=gcopy(S1);
  }
  else
  {
    tetpil=avma; S=cgetg(1,t_MAT);
  }
  u=cgetg(4,t_VEC);
  u[1]=lstoi(s<<1);
  u[2]=lcopy(normax1);
  u[3]=(long)S;
  tetpil=avma; return gerepile(av,tetpil,gcopy(u));
}

/* resolution de x~.a.x <= borne, a definie positive */
GEN
fincke_pohst(GEN a,GEN borne,GEN stockmax,long prec)
{
  long pr,av=avma,tetpil,i,j,n,m;
  GEN r,rinv,rinvtrans,v,vtrans,u,s,sinvtrans,a1,res,x,y,matx,maty,z;
  GEN unr,vnorm,sperm,perm;

  if (DEBUGLEVEL>3)
    { fprintferr("\nEntree dans fincke_pohst\n"); flusherr(); }
  pr=gprecision(a);
  if (!pr) { unr=cgetr(prec); affsr(1,unr); a=gmul(a,unr); }
  else prec=pr;
  r=sqred1(a); n=lg(a)-1;
  for (i=1; i<=n; i++)
  {
    coeff(r,i,i)=lsqrt(gcoeff(r,i,i),prec);
    for (j=i+1; j<=n; j++)
      coeff(r,i,j)=lmul(gcoeff(r,i,i),gcoeff(r,i,j));
  }
  rinv=invmatreel(r); rinvtrans=gtrans(rinv);
  if (DEBUGLEVEL>8)
  {
    fprintferr("prec = %ld, precision de rinvtrans = %ld\n",
                prec,gprecision(rinvtrans));
    fprintferr("rinvtrans = "); outerr(rinvtrans);
    flusherr();
  }
  if (DEBUGLEVEL>3) { fprintferr("\nEntree dans LLL\n"); flusherr(); }
  v=lll(rinvtrans,2*prec-2);
  if (DEBUGLEVEL>8) { fprintferr("v = "); outerr(v); flusherr(); }

  vtrans=gtrans(v); u=invmat(vtrans);
  s=gmul(r,u); sinvtrans=gmul(rinvtrans,v);
  vnorm=cgetg(n+1,t_VEC);
  for (j=1; j<=n; j++)
    vnorm[j]=lnorml2((GEN)sinvtrans[j]);
  perm=sindexsort(vnorm); sperm=cgetg(n+1,t_MAT);
  for (j=1; j<=n; j++)
    sperm[n+1-j]=s[perm[j]];
  a1=gmul(gtrans(sperm),sperm);
  if (DEBUGLEVEL>3)
    { fprintferr("\nEntree dans smallvectors\n"); flusherr(); }
  res=smallvectors(a1,borne,stockmax,prec);
  maty=(GEN)res[3]; m=lg(maty)-1; matx=cgetg(m+1,t_MAT);
  for (j=1; j<=m; j++)
  {
    y=(GEN)maty[j]; x=cgetg(n+1,t_COL);
    for (i=1; i<=n; i++)
      x[perm[i]] = y[n+1-i];
    matx[j]=lmul(u,x);
  }
  if (DEBUGLEVEL>3){ fprintferr("\nSortie de fincke_pohst\n"); flusherr(); }
  tetpil=avma;
  z=cgetg(4,t_VEC);
  z[1]=lcopy((GEN)res[1]);
  z[2]=pr ? lcopy((GEN)res[2]) : lround((GEN)res[2]);
  z[3]=lcopy(matx); return gerepile(av,tetpil,z);
}
